<?php

/**
 * AsalaeCore\Command\AdminCommand
 */

namespace AsalaeCore\Command;

use AsalaeCore\Form\AdminShellForm;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Core\Configure;
use Cake\Utility\Hash;
use Exception;
use stdClass;

/**
 * Command d'install d'admin fonctionnel
 *
 * @category    Command
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2019, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class AdminCommand extends Command
{
    use PasswordTrait;

    public const string ACTION_ADD = 'a';
    public const string ACTION_EDIT = 'm';
    public const string ACTION_DELETE = 's';
    public const string ACTION_QUIT = 'q';

    /**
     * @var Arguments
     */
    public $args;
    /**
     * @var ConsoleIo
     */
    public $io;

    /**
     * @var array Champs affichés dans la liste d'admins ['champ' => 'traduction']
     */
    public $fields = [];

    /**
     * Chemin vers le fichier json qui conserve les données
     * @var string
     */
    protected $adminFileUri = '';

    /**
     * @var stdClass liste les administrateurs configurés
     */
    protected $admins = [];

    /**
     * @var AdminShellForm Classe de formulaire
     */
    protected $form;

    /**
     * Constructs this Shell instance.
     *
     * @link https://book.cakephp.org/3.0/en/console-and-shells.html#Shell
     */
    public function __construct()
    {
        $this->fields = [
            'username' => [
                'label' => __("Identifiant de connexion"),
                'primary' => true,
                'required' => true,
            ],
            'name' => [
                'label' => __("Nom de l'utilisateur"),
                'required' => true,
            ],
            'email' => [
                'label' => __("Courriel"),
                'email' => true,
                'required' => false,
            ],
            'password' => [
                'label' => __("Mot de passe"),
                'password' => true,
                'confirm' => 'confirm-password',
                'display' => false,
                'required' => true,
            ],
        ];
    }

    /**
     * Run the command.
     *
     * @param array     $argv Arguments from the CLI environment.
     * @param ConsoleIo $io   The console io
     * @return int|null Exit code or null for success.
     */
    public function run(array $argv, ConsoleIo $io): ?int
    {
        $this->io = $io; // rend io accéssible dans initialize()
        return parent::run($argv, $io);
    }

    /**
     * Initialisation du shell
     */
    public function initialize(): void
    {
        parent::initialize();

        $this->adminFileUri = Configure::read(
            'App.paths.administrators_json',
            Configure::read('App.paths.data') . DS . 'administrateurs.json'
        );
        if (!is_writeable(dirname($this->adminFileUri))) {
            $this->io->abort(
                __(
                    "Le repertoire de configuration {0} n'est pas accessible en écriture",
                    $this->adminFileUri
                )
            );
        }
        $this->form = AdminShellForm::buildForm();
    }

    /**
     * liste des actions disponibles
     * @return array ['action' => ['label' => ..., 'short' => ..., 'enabled' => ...]]
     */
    protected function getActions(): array
    {
        return [
            'edit' => [
                'label' => __("Modifier"),
                'short' => self::ACTION_EDIT,
                'enabled' => !empty($this->admins),
            ],
            'add' => [
                'label' => __("Ajouter"),
                'short' => self::ACTION_ADD,
                'enabled' => true,
            ],
            'delete' => [
                'label' => __("Supprimer"),
                'short' => self::ACTION_DELETE,
                'enabled' => !empty($this->admins),
            ],
            'quit' => [
                'label' => __("Quitter"),
                'short' => self::ACTION_QUIT,
                'enabled' => true,
            ],
        ];
    }

    /**
     * Permet de choisir parmis une liste d'actions possibles
     */
    protected function selectAction()
    {
        $actions = $this->getActions();
        $message = [];
        $enabled = [];
        foreach ($actions as $params) {
            if (!$params['enabled']) {
                continue;
            }
            $message[] = $params['label'] . ' (' . $params['short'] . ')';
            $enabled[] = $params['short'];
        }
        return $this->io->askChoice(implode(', ', $message), $enabled, 'a');
    }

    /**
     * méthode principale
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return void
     * @throws Exception
     */
    public function execute(Arguments $args, ConsoleIo $io)
    {
        $this->args = $args;
        $this->io = $io;
        $this->admins = is_file($this->adminFileUri)
            ? json_decode(file_get_contents($this->adminFileUri), true)
            : [];
        $this->renderAdminsList();

        switch ($this->selectAction()) {
            case self::ACTION_ADD:
                $this->actionAdd();
                break;
            case self::ACTION_EDIT:
                $this->actionEdit();
                break;
            case self::ACTION_DELETE:
                $this->actionDelete();
                break;
            case self::ACTION_QUIT:
                $this->actionQuit();
                return;
        }
        $this->execute($args, $io);
    }

    /**
     * Affiche la liste des administrateurs sous forme de tableau
     */
    protected function renderAdminsList()
    {
        $tds = [];
        $fields = [];
        foreach ($this->fields as $field => $params) {
            if (isset($params['display']) && !$params['display']) {
                continue;
            }
            $fields[] = $field;
            $tds[] = $params['label'];
        }
        $tds[] = __("Notifications");
        $trs = [$tds];

        foreach ($this->admins as $admin) {
            $tds = [];
            foreach ($fields as $field) {
                $tds[] = $admin[$field] ?? null;
            }
            $tds[] = ($admin['notify'] ?? '1') ? __("Oui") : '';
            $trs[] = $tds;
        }
        $this->io->helper('Table')->output($trs);
    }

    /**
     * Renvoi la clé primaire dans $this->fields()
     * @return string
     * @throws Exception
     */
    protected function getPrimaryKey(): string
    {
        foreach ($this->fields as $field => $params) {
            if (!empty($params['primary'])) {
                return $field;
            }
        }
        throw new Exception("Missing primary key");
    }

    /**
     * Effectue une validation avec affichage d'erreurs
     * @param array $data
     * @return bool
     */
    protected function validateField(array $data): bool
    {
        $success = $this->form->validate($data);
        if (!$success) {
            foreach (Hash::flatten($this->form->getErrors()) as $error) {
                $this->io->err($error);
            }
        }
        return $success;
    }

    /**
     * Modification d'un administrateur existant
     */
    protected function actionEdit()
    {
        $this->io->out(__('Modification d\'un administrateur (CTRL+C pour quitter)'));
        $this->io->hr();
        $options = [];
        $primary = $this->getPrimaryKey();
        foreach ($this->admins as $admin) {
            $options[] = $admin[$primary];
        }
        $username = $this->io->askChoice(
            __('Identifiant de l\'administrateur a modifier'),
            $options,
            $options[0]
        );
        $data = null;
        foreach ($this->admins as $admin) {
            if (strcasecmp($admin[$primary], $username) === 0) {
                $data = $admin;
            }
        }
        $data['username'] = $username;
        $data['action'] = 'edit';
        $initial = $data;
        foreach ($this->fields as $field => $params) {
            do {
                $success = $this->askField($field, $data, 'edit', $initial);
            } while (!$success);
        }
        $notify = $this->io->askChoice(
            __("Recevoir les notifications d'exploitation ?"),
            ['o', 'n'],
            'o'
        );
        $data['notify'] = $notify === 'o' ? '1' : '0';
        if (!$this->form->execute($data)) {
            $this->io->abort(current(Hash::flatten($this->form->getErrors())));
        }
    }

    /**
     * Ajout d'un administrateur
     * @throws Exception
     */
    protected function actionAdd()
    {
        $this->io->out(__("Ajout d'un administrateur (CTRL+C pour quitter)"));
        $this->io->hr();
        $data = ['action' => 'add'];
        foreach (array_keys($this->fields) as $field) {
            do {
                $success = $this->askField($field, $data, 'add');
            } while (!$success);
        }
        $notify = $this->io->askChoice(
            __("Recevoir les notifications d'exploitation ?"),
            ['o', 'n'],
            'o'
        );
        $data['notify'] = $notify === 'o' ? '1' : '0';
        if (!$this->form->execute($data)) {
            $this->io->abort(current(Hash::flatten($this->form->getErrors())));
        }
    }

    /**
     * Demande un champ de formulaire
     * @param string $field
     * @param array  $data
     * @param string $action
     * @param array  $initial
     * @return bool
     * @throws Exception
     */
    private function askField(string $field, array &$data, string $action, array $initial = []): bool
    {
        $params = $this->fields[$field];
        if (!empty($params['primary']) && $action === 'edit') {
            $success = true;
        } elseif (!empty($params['password'])) {
            $askPasswordMessage = $action === 'edit'
                ? __('{0} (laisser vide pour ne pas modidier la valeur du mot de passe)', $params['label'])
                : $params['label'];
            $data[$field] = $this->enterPassword($askPasswordMessage);
            if ($data[$field]) {
                $data[$params['confirm']] = $this->enterPassword(__("Confirmez le mot de passe") . ' ');
            }
            if (empty($data[$field]) && $action === 'edit') {
                $data[$field] = $initial[$field];
                $data[$params['confirm']] = $data[$field];
                $data['encrypted'] = true;
                $success = true;
            } else {
                $success = $this->validateField(
                    [
                        $field => $data[$field],
                        $params['confirm'] => $data[$params['confirm']],
                        'action' => $action,
                    ]
                );
            }
            if (!$success) {
                $choice = $this->io->askChoice(__("Voulez-vous malgré tout l'utiliser ?"), ['o', 'n'], 'o');
                if (strcasecmp($choice, 'o') === 0) {
                    $success = true;
                    $data['ignorePasswordComplexity'] = true;
                }
            }
        } else {
            $data[$field] = $this->io->ask($params['label'], $initial[$field] ?? null);
            $success = $this->validateField([$field => $data[$field], 'action' => $action]);
        }
        return $success;
    }

    /**
     * Suppression d'un administrateur
     * @throws Exception
     */
    protected function actionDelete()
    {
        $this->io->out(__('Suppression d\'un administrateur (CTRL+C pour quitter)'));
        $this->io->hr();
        $options = [];
        $primary = $this->getPrimaryKey();
        foreach ($this->admins as $admin) {
            $options[] = $admin[$primary];
        }
        $username = $this->io->askChoice(
            __('Identifiant de l\'administrateur a supprimer'),
            $options,
            end($options)
        );
        $confirm = $this->io->askChoice(
            __("Êtes-vous certain de vouloir supprimer {0} ?", $username),
            ['y', 'n']
        );
        if ($confirm === 'y') {
            $form = new AdminShellForm();
            $form->execute(['action' => 'delete', 'username' => $username]);
        }
    }

    /**
     * Arret du shell
     */
    protected function actionQuit()
    {
        $this->abort(self::CODE_SUCCESS);
    }
}
