<?php

/**
 * AsalaeCore\Command\ControllersCommand
 */

namespace AsalaeCore\Command;

use AsalaeCore\Model\Table\AcosTable;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Console\ConsoleOptionParser as CakeConsoleOptionParser;
use Cake\Core\App;
use Cake\Core\Configure;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Utility\Inflector;
use Exception;
use ReflectionClass;
use ReflectionException;

/**
 * Shell de modification du controllers.json et des Acos
 *
 * @category    Command
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2019, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 * @property AcosTable $Acos
 */
class ControllersSetCommand extends Command
{
    use CreateAcosTrait;

    /**
     * @var Arguments
     */
    public $args;
    /**
     * @var ConsoleIo
     */
    public $io;
    private AcosTable|Table $Acos;

    /**
     * Get the command name.
     * @return string
     */
    public static function defaultName(): string
    {
        return 'controllers set';
    }

    /**
     * Gets the option parser instance and configures it.
     *
     * By overriding this method you can configure the ConsoleOptionParser before returning it.
     *
     * @return CakeConsoleOptionParser
     * @link https://book.cakephp.org/3.0/en/console-and-shells.html#configuring-options-and-generating-help
     */
    public function getOptionParser(): CakeConsoleOptionParser
    {
        $parser = new ConsoleOptionParser();
        $parser->addArgument(
            'target',
            [
                'help' => __("Action sous la forme Controller/action"),
            ]
        );
        $parser->addOption(
            'invisible',
            [
                'short' => 'i',
                'boolean' => true,
                'help' => __("Affiche l'action dans l'édition des permissions"),
            ]
        );
        $parser->addOption(
            'accesParDefaut',
            [
                'short' => 'a',
                'boolean' => true,
                'help' => __(
                    "Donne un accès à tous les aros déjà créés,"
                    . " et ajoutera un accès lors de la création d'un nouvel aro"
                ),
            ]
        );
        $parser->addOption(
            'commeDroits',
            [
                'short' => 'c',
                'help' => __(
                    "L'aco créé sera déclaré enfant de la cible défini ici. Syntaxe: NomDuController::nomDeLaction"
                ),
            ]
        );
        $parser->setDescription(
            [
                __(
                    "Permet d'ajouter une entrée dans controllers.json"
                    . " suite à l'ajout d'une action (évite de passer par /devs/index)"
                ),
            ]
        );
        $parser->addOption(
            'interactive',
            [
                'help' => __("Demandera toutes les options si à Yes"),
                'default' => 'y',
                'choice' => ['y', 'n'],
            ]
        );
        $parser->addOption(
            'force',
            [
                'help' => __("Ne contrôle pas les choix possibles"),
                'boolean' => true,
            ]
        );
        $parser->addOption(
            'plugin',
            [
                'help' => __("Cible un plugin"),
            ]
        );

        return $parser;
    }

    /**
     * Action principale
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return void
     * @throws Exception
     */
    public function execute(Arguments $args, ConsoleIo $io)
    {
        $this->args = $args;
        $this->io = $io;
        $Acos = $this->fetchTable('Acos');
        $root = $Acos->find()
            ->select(['id', 'lft', 'rght'])
            ->where(['model' => 'root', 'alias' => 'controllers'])
            ->first();
        if (!$root) {
            $this->io->abort(__("Les permissions n'ont pas été initialisés"));
        }
        $controllers = $this->getControllersActions();
        $pathToControllersJson = Configure::read(
            'App.paths.controllers_rules',
            RESOURCES . 'controllers.json'
        );
        $data = json_decode(file_get_contents($pathToControllersJson), true);
        $args = $this->args->getArgument('target');
        if ($args === null) {
            $args = $this->askControllerAction($data, $controllers);
        }
        $params = $this->args->getOption('interactive') === 'y'
            ? $this->askParams($data, $controllers, $args)
            : [
                'invisible' => (string)(int)$this->args->getOption('invisible'),
                'accesParDefaut' => (string)(int)$this->args->getOption('accesParDefaut'),
                'commeDroits' => $this->args->getOption('commeDroits') ?: '',
            ];

        if ($this->args->getOption('commeDroits')) {
            [$controller, $action] = explode('::', $this->args->getOption('commeDroits'));
            if (!$this->checkExistance($controller, $action)) {
                $this->io->abort(__("l'action du commeDroits n'a pas été trouvée"));
            }
        }
        while (
            empty($args)
            || !preg_match('/^([A-Za-z\d.-]+)\/([A-Za-z\d-]+)$/', $args, $match)
            || !$this->checkExistance($match[1], $match[2])
        ) {
            $this->io->out("La cible doit être indiquée sous la forme Controllers/actionEnCamelCase");
            $args = $this->io->ask(__("Veuillez saisir une cible"));
        }
        $pathToControllersJson = Configure::read(
            'App.paths.controllers_rules',
            RESOURCES . 'controllers.json'
        );
        $data = json_decode(file_get_contents($pathToControllersJson), true);
        [, $controller, $action] = $match;
        $controller = Inflector::camelize($controller, '-');
        $action = Inflector::variable(Inflector::camelize($action, '-'));
        if (!isset($data[$controller])) {
            $data[$controller] = [];
        }
        $data[$controller][$action] = $params;
        ksort($data[$controller]);
        ksort($data);

        file_put_contents($pathToControllersJson, json_encode($data, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES));

        if (!$this->args->getOption('commeDroits')) {
            $parent = $Acos->findOrCreate(
                [
                    'model' => 'controllers',
                    'alias' => $controller,
                    'parent_id' => $root->get('id'),
                ]
            );
        } else {
            [$cdController, $cdAction] = explode('::', $this->args->getOption('commeDroits'));
            $parent = $Acos->find()
                ->select(['id'])
                ->where(
                    [
                        'model' => $cdController,
                        'alias' => $cdAction,
                        'lft >=' => $root->get('lft'),
                        'rght <=' => $root->get('rght'),
                    ]
                )
                ->firstOrFail();
        }
        $aco = $Acos->findOrCreate(
            [
                'model' => $controller,
                'alias' => $action,
                'parent_id' => $parent->get('id'),
            ]
        );

        if ($this->args->getOption('accesParDefaut')) {
            $this->allowForAll($aco);
        }

        $this->io->out('done');
    }

    /**
     * Donne un array avec le nom du controlleur en clé et le nom de l'action en valeur
     * @return array
     * @throws ReflectionException
     */
    protected function getControllersActions()
    {
        $controllers = [];
        $plugin = $this->args->getOption('plugin');
        $namespace = Configure::read('App.namespace');
        $classname = $namespace . "\\Controller\\AppController";
        if (class_exists($classname)) {
            $appControllerMethods = get_class_methods($classname);
        } else {
            $appControllerMethods = get_class_methods(
                $namespace . "\\Controller\\Controller"
            );
        }
        foreach (App::classPath('Controller', $plugin) as $dir) {
            foreach (glob($dir . '*Controller.php') as $filename) {
                $controller = basename($filename, 'Controller.php');
                $classname = App::className(
                    ($plugin ? $plugin . '.' : '') . $controller,
                    'Controller',
                    'Controller'
                );
                if (!$classname) {
                    continue;
                }
                $refl = new ReflectionClass($classname);
                if ($refl->isAbstract() || $refl->isTrait() || $refl->isInterface()) {
                    continue;
                }
                $methods = [];
                foreach ($refl->getMethods() as $method) {
                    if ($method->isPublic() && !$method->isStatic() && $method->class === $classname) {
                        $methods[] = $method->getName();
                    }
                }
                $controllers[$controller] = array_diff($methods, $appControllerMethods);
            }
        }
        return $controllers;
    }

    /**
     * Mode interactif
     * @param array $data
     * @param array $controllers
     * @return string
     */
    private function askControllerAction(array $data, array $controllers): string
    {
        $missings = [];
        foreach ($controllers as $controller => $actions) {
            foreach ($actions as $action) {
                if (!Hash::get($data, $controller . '.' . $action)) {
                    $missings[$controller][$action] = true;
                }
            }
        }
        if (!$missings) {
            $this->io->warning(__("Toutes les actions ont été définies"));
            $defaultController = null;
        } else {
            $defaultController = current(array_keys($missings));
        }
        if ($this->args->getOption('force')) {
            $controller = $this->io->ask(
                __("Sélectionner un Controller"),
                $defaultController
            );
        } else {
            $controller = $this->io->askChoice(
                __("Sélectionner un Controller"),
                array_keys($controllers),
                $defaultController
            );
        }
        if (empty($missings[$controller])) {
            $this->io->warning(__("Toutes les actions ont été définies"));
            $defaultAction = null;
        } else {
            $defaultAction = current(array_keys($missings[$controller]));
        }
        if ($this->args->getOption('force')) {
            $action = $this->io->ask(
                __("Sélectionner une action"),
                $defaultAction
            );
        } else {
            $action = $this->io->askChoice(
                __("Sélectionner une action"),
                $controllers[$controller],
                $defaultAction
            );
        }
        return $controller . '/' . $action;
    }

    /**
     * Paramètres de l'action (commeDroit, invisible, etc..)
     * @param array  $data
     * @param array  $controllers
     * @param string $args
     * @return array
     */
    private function askParams(
        array $data,
        array $controllers,
        string $args
    ): array {
        $params = [
            'invisible' => (string)(int)$this->args->getOption('invisible'),
            'accesParDefaut' => (string)(int)$this->args->getOption('accesParDefaut'),
            'commeDroits' => $this->args->getOption('commeDroits') ?: '',
        ];
        if (!$params['commeDroits']) {
            $params['commeDroits'] = $this->io->askChoice(
                __("L'action hérite des permissions d'accès d'une autre action ?"),
                ['y', 'n'],
                'n'
            );
            if ($params['commeDroits'] === 'y') {
                $controller = $this->io->askChoice(
                    __("Sélectionner un Controller"),
                    $this->args->getOption('force') ? null : array_keys($controllers),
                    explode('/', $args)[0]
                );
                if ($this->args->getOption('force')) {
                    $action = $this->io->askChoice(
                        __("Sélectionner une action"),
                        $controllers[$controller]
                    );
                } else {
                    $action = $this->io->ask(__("Sélectionner une action"));
                }
                $params['commeDroits'] = Hash::get($data, $controller . '.' . $action . '.commeDroits')
                    ?: $controller . '::' . $action;
            } else {
                $params['commeDroits'] = '';
            }
        }
        if (!$params['commeDroits'] && $this->args->getOption('invisible') === false) {
            $params['invisible'] = $this->io->askChoice(
                __("L'action est-elle visible dans les permissions ?"),
                ['y', 'n'],
                'y'
            ) === 'y' ? '0' : '1';
        }
        if (!$params['commeDroits'] && $this->args->getOption('accesParDefaut') === false) {
            $params['accesParDefaut'] = $this->io->askChoice(
                __("L'action est-elle accessible par défaut ?"),
                ['y', 'n'],
                'n'
            ) === 'y' ? '1' : '0';
        }
        if (!$params['commeDroits'] && $params['invisible'] === '0') {
            $pathToControllersGroupJson = Configure::read(
                'App.paths.controllers_group',
                RESOURCES . 'controllers_group.json'
            );
            $groups = json_decode(file_get_contents($pathToControllersGroupJson), true);
            $groups[] = 'empty';
            foreach ($groups as $key => $group) {
                $this->io->info(sprintf('%d - %s', $key, $group));
            }
            $key = $this->io->askChoice(
                __("Indiquer le groupe de la permission"),
                $this->args->getOption('force') ? null : array_keys($groups)
            );
            $params['group'] = $groups[$key] ?? $key;
            if ($params['group'] === 'empty') {
                unset($params['group']);
            }
        }
        return $params;
    }

    /**
     * Vérifi l'existance d'une action
     * @param string $controller
     * @param string $action
     * @return bool
     */
    protected function checkExistance(string $controller, string $action): bool
    {
        $controller = Inflector::camelize($controller, '-');
        $action = Inflector::variable(Inflector::camelize($action, '-'));
        $plugin = $this->args->getOption('plugin');
        $classname = App::className(
            ($plugin ? $plugin . '.' : '') . $controller,
            'Controller',
            'Controller'
        );
        if (!$classname) {
            return false;
        }
        return method_exists($classname, $action);
    }

    /**
     * Vérifi controllers.json
     * @return void
     * @throws ReflectionException
     */
    public function check()
    {
        $controllers = $this->getControllersActions();

        $pathToControllersJson = Configure::read(
            'App.paths.controllers_rules',
            RESOURCES . 'controllers.json'
        );
        $data = json_decode(file_get_contents($pathToControllersJson), true);
        $error = false;
        foreach ($data as $controller => $actions) {
            foreach (array_keys($actions) as $action) {
                if (!isset($controllers[$controller])) {
                    $error = true;
                    $this->io->err(__("Le controller {0} n'a pas été trouvé dans l'application", $controller));
                } elseif (!in_array($action, $controllers[$controller])) {
                    $error = true;
                    $this->io->err(
                        __(
                            "L'action {0} : {1} n'a pas été trouvé dans l'application",
                            $controller,
                            $action
                        )
                    );
                }
            }
        }
        foreach ($controllers as $controller => $actions) {
            foreach ($actions as $action) {
                if (!Hash::get($data, $controller . '.' . $action)) {
                    $this->io->err(
                        __(
                            "L'action {0} : {1} n'est pas définie dans controllers.json",
                            $controller,
                            $action
                        )
                    );
                    $error = true;
                }
            }
        }
        if (!$error) {
            $this->io->success('ok');
        }
    }
}
