<?php

/**
 * AsalaeCore\CronCommand
 */

namespace AsalaeCore\Command;

use AsalaeCore\Factory\Utility;
use AsalaeCore\Model\Table\CronExecutionsTable;
use AsalaeCore\Console\ConsoleIo as CoreConsoleIo;
use AsalaeCore\Controller\Component\ConditionComponent;
use AsalaeCore\Cron\CronInterface;
use AsalaeCore\Utility\Notify;
use AsalaeCore\View\AppView;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Console\ConsoleOutput;
use Cake\Controller\Component;
use Cake\Controller\Controller;
use Cake\Core\Configure;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Response as CakeResponse;
use Cake\Http\ServerRequest;
use Cake\I18n\DateTime as I18nDateTime;
use Cake\Mailer\Mailer;
use Cake\ORM\Table;
use Cake\View\Helper\UrlHelper;
use DateInterval;
use DateTime;
use DateTimeInterface;
use Exception;
use IntlDateFormatter;
use Libriciel\Filesystem\Utility\Filesystem;
use ReflectionException;

/**
 * Permet de lancer un cron (utilisé par crontab)
 *
 * @category    Command
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2018, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 * @property Table $Crons
 * @property Table $CronExecutions
 */
class CronCommand extends Command
{
    /**
     * @var Arguments
     */
    public $args;
    /**
     * @var ConsoleIo
     */
    public $io;

    /**
     * @var EntityInterface|null Cron
     */
    public $cron;

    /**
     * @var EntityInterface|null CronExecution
     */
    public $exec;

    /**
     * CronCommand constructor.
     */
    public function __construct()
    {
        I18nDateTime::$niceFormat = [IntlDateFormatter::SHORT, IntlDateFormatter::MEDIUM];
    }

    /**
     * Getter du stdout
     * @return ConsoleOutput
     */
    public function getOut(): ConsoleOutput
    {
        if ($this->io instanceof CoreConsoleIo) {
            return $this->io->getOut();
        }
        return new ConsoleOutput('php://stdout');
    }

    /**
     * Getter du stderr
     * @return ConsoleOutput
     */
    public function getErr(): ConsoleOutput
    {
        if ($this->io instanceof CoreConsoleIo) {
            return $this->io->getErr();
        }
        return new ConsoleOutput('php://stderr');
    }

    /**
     * Display help for this console.
     *
     * @return ConsoleOptionParser
     */
    public function getOptionParser(): ConsoleOptionParser
    {
        $parser = new ConsoleOptionParser();

        $parser->addOption(
            'no-interaction',
            [
                'help' => __("Désactive la partie interactive"),
                'boolean' => true,
            ]
        );

        return $parser;
    }

    /**
     * Lancement des crons
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return void
     * @throws Exception
     */
    public function execute(Arguments $args, ConsoleIo $io)
    {
        $this->args = $args;
        $this->io = $io;
        register_shutdown_function([$this, 'shutdown']);
        $this->io->out(__("Lancement des crons..."));

        $conditions = [
            'Crons.active' => true,
            'Crons.locked' => false,
            'Crons.next <' => new DateTime(),
        ];
        if ($this->fetchTable('Crons')->exists(['locked' => true, 'active' => true])) {
            $conditions['Crons.parallelisable'] = true;
        }

        $crons = $this->fetchTable('Crons')->find()
            ->where($conditions)
            ->all()
            ->toArray();
        $ids = array_map(
            function (EntityInterface $v) {
                return $v->get('id');
            },
            $crons
        );
        $this->io->out(__n("Il y a {0} cron à lancer", "Il y a {0} crons à lancer", $c = count($ids), $c));
        /** @var EntityInterface $cron */
        foreach ($ids as $id) {
            $this->exec = null;
            $this->cron = null;
            $cron = $this->fetchTable('Crons')->find()
                ->where(['Crons.id' => $id])
                ->andWhere($conditions)
                ->first();
            if (!$cron) {
                continue;
            }
            try {
                $this->runCron($cron);
            } catch (Exception $e) {
                $this->io->err((string)$e);
                $cron->set('locked', false);
                $this->fetchTable('Crons')->save($cron);
                $this->exec?->set(
                    'report',
                    $this->exec->get('report') . " \n"
                    . __("Erreur lors de l'execution du cron ({0})", $e->getMessage())
                );
            }
        }
        $this->io->out(__("Fin."));
    }

    /**
     * Envoi un message d'alerte aux administrateurs
     * @param EntityInterface $cron
     * @param EntityInterface $exec
     * @throws Exception
     */
    private function sendMail(EntityInterface $cron, EntityInterface $exec)
    {
        $lastmail = $cron->get('last_email');
        $interval = date_interval_create_from_date_string(
            Configure::read('Crons.emails.min_interval', '1 day')
        );
        if ($lastmail instanceof DateTimeInterface && $lastmail->add($interval) > new DateTime()) {
            return;
        }
        // lecture des mails des administateurs techniques (les adresses mails xxx@test.fr sont ignorées)
        $mailField = Configure::read('AsalaeCore.CronCommand.email.adminEmailField', 'email');
        $notifyField = Configure::read('AsalaeCore.CronCommand.email.notificationField', 'notify');
        $adminFileUri = Configure::read(
            'App.paths.administrators_json',
            Configure::read('App.paths.data') . DS . 'administrateurs.json'
        );
        $mails = array_map(
            function ($v) use ($mailField, $notifyField) {
                return ($v[$notifyField] ?? '') === '0' || !isset($v[$mailField]) ? false : $v[$mailField];
            },
            json_decode(file_get_contents($adminFileUri), true)
        );
        $mails = array_filter(
            $mails,
            function ($v) {
                return $v && substr($v, strpos($v, '@') + 1) !== 'test.fr';
            }
        );
        if (empty($mails)) {
            return;
        }

        // initialisation et envoi du mail
        /** @var Mailer $email */
        $email = clone Utility::get(Mailer::class);
        $email->viewBuilder()
            ->setTemplate(Configure::read('AsalaeCore.CronCommand.email.template', 'AsalaeCore.cronfailed'))
            ->setLayout(Configure::read('AsalaeCore.CronCommand.email.layout', 'AsalaeCore.default'))
            ->addHelpers(['Html', 'Url']);
        $mailTitlePrefix = Configure::read('AsalaeCore.CronCommand.email.titlePrefix', '');
        try {
            $success = $email->setEmailFormat(Configure::read('AsalaeCore.CronCommand.email.format', 'both'))
                ->setSubject(
                    $mailTitlePrefix . ($mailTitlePrefix ? '  ' : '')
                    . __("Rapport de tâche planifiée ''{0}''", $cron->get('name'))
                )
                ->setViewVars(compact('cron', 'exec'))
                ->setTo(array_values($mails))
                ->send();
            if ($response = $this->debugMail($email)) {
                $body = $response->getBody();
                $body->rewind();
                $this->io->out('debug_mail: ' . $body->getContents());
            }
        } catch (Exception $e) {
            $exec->set(
                'report',
                $exec->get('report') . " \n"
                . __("Erreur lors de l'envoi de la notification d'erreur ({0})", $e->getMessage())
            );
            $this->fetchTable('CronExecutions')->save($exec);
            throw $e;
        }

        if ($success) {
            $cron->set('last_email', new DateTime());
            $this->fetchTable('Crons')->saveOrFail($cron);
            $this->io->out(
                __n(
                    "Un email a été envoyé à l'administrateur: {0}",
                    "Un email a été envoyé aux administrateurs: {0}",
                    count($mails),
                    implode(', ', $mails)
                )
            );
        }
    }

    /**
     * Choisi un cron en bdd à partir d'une chaine de caractère
     * @param string $cron
     * @return EntityInterface
     * @throws ReflectionException
     */
    private function chooseCron(string $cron): EntityInterface
    {
        $request = new ServerRequest();
        $controller = new Controller($request);
        $registry = $controller->components();
        $Conditions = new ConditionComponent($registry);
        do {
            $crons = $this->fetchTable('Crons')->find()
                ->where(
                    [
                        'OR' => [
                            $Conditions->ilike('name', '%' . $cron . '%'),
                            $Conditions->ilike('classname', '%' . $cron . '%'),
                        ],
                    ]
                )
                ->all();
            if ($crons->count() === 1) {
                $classname = $crons->first()->get('classname');
                $classname = substr($classname, strrpos($classname, '\\') + 1);
                $this->io->out(__("Trouvé: {0}", $crons->first()->get('name')));
                $response = $this->io->askChoice(__("Lancer le cron {0} ?", $classname), ['y', 'n'], 'y');
                if ($response === 'y') {
                    $cron = $crons->first();
                    break;
                }
            } else {
                if ($crons->count() === 0) {
                    $crons = $this->fetchTable('Crons')->find()->all();
                }
                $this->io->out(__("Choisir parmis :"));
                $classnames = [];
                foreach ($crons as $cronEntity) {
                    $classname = $cronEntity->get('classname');
                    $classnames[] = substr($classname, strrpos($classname, '\\') + 1);
                }
                sort($classnames);
                array_map(fn($v) => $this->io->info($v), $classnames);
            }
            $cron = $this->io->ask(__("Choisir un cron"));
        } while (true);
        return $cron;
    }

    /**
     * Lance un cron
     * @param int|string|EntityInterface $cron
     * @throws Exception
     */
    public function runCron($cron)
    {
        $now = new DateTime();
        if (is_numeric($cron)) {
            try {
                $cron = $this->fetchTable('Crons')->get($cron);
            } catch (RecordNotFoundException) {
                $this->io->abort(__("Le cron id={0} n'existe pas", $cron));
            }
        } elseif ($this->args->getOption('no-interaction') === false && is_string($cron)) {
            $cron = $this->chooseCron($cron);
        }
        if (!$cron instanceof EntityInterface) {
            throw new BadRequestException();
        }
        $this->cron = $cron;
        $this->io->hr();
        $this->io->out((new I18nDateTime())->nice() . ' - ' . __("Lancement du cron: {0}", $cron->get('name')));
        $this->io->hr();
        $cron->set('locked', true);
        $this->fetchTable('Crons')->saveOrFail($cron);
        $pid = getmypid();
        $exec = $this->fetchTable('CronExecutions')->newEntity(
            [
                'cron_id' => $cron->get('id'),
                'date_end' => null,
                'state' => CronExecutionsTable::S_RUNNING,
                'report' => '',
                'pid' => file_exists('/proc/' . $pid) ? $pid : null,
            ]
        );
        // NOTE: le setter de cake ne prend pas les microsecondes
        $exec->set('date_begin', (new DateTime())->format('Y-m-d H:i:s.u'), ['setter' => false]);
        $this->fetchTable('CronExecutions')->saveOrFail($exec);
        Notify::emit(
            'crons',
            [
                'cron_id' => $cron->id,
                'exec_id' => $exec->id,
                'state' => CronExecutionsTable::S_RUNNING,
            ]
        );
        $this->exec = $exec;
        $params = (array)json_decode($cron->get('app_meta'));
        $params['context'] = $this->args->getOption('context');
        // FIXME ajouter option "params" ?
        foreach ((array)$this->args->getOption('params') as $p) {
            if (!strpos($p, '=')) {
                $params[$p] = true;
            } else {
                [$key, $value] = explode('=', $p, 2);
                $params[$key] = $value;
            }
        }
        $Class = null;
        try {
            $classname = $cron->get('classname');
            /** @var CronInterface $Class */
            $Class = new $classname($params, $this->getOut(), $this->getErr());
            $state = $Class->work($exec, $cron);
            $output = $Class->getOutput();
        } catch (Exception $e) {
            $conn = $this->fetchTable('Crons')->getConnection();
            while ($conn->inTransaction()) {
                $conn->rollback();
            }
            try {
                $Class->err((string)$e);
                $output = $Class->getOutput();
                $this->exec?->set(
                    'report',
                    $this->exec->get('report') . " \n"
                    . __("Erreur lors de l'execution du cron ({0})", $e->getMessage())
                );
            } catch (Exception) {
                $output = $e->getMessage();
            }
            $state = CronExecutionsTable::S_ERROR;
        }
        $this->fetchTable('CronExecutions')->patchEntity(
            $exec,
            [
                'state' => $state,
                'report' => $output,
            ]
        );
        $exec->set('date_end', (new DateTime())->format('Y-m-d H:i:s.u'), ['setter' => false]);
        $this->fetchTable('CronExecutions')->save($exec);
        Notify::emit('crons', ['cron_id' => $cron->id, 'exec_id' => $exec->id, 'state' => $state]);
        $cron->set('locked', false);
        if ($cron->get('frequency') === 'one_shot') {
            if ($state === CronExecutionsTable::S_SUCCESS) {
                $this->fetchTable('CronExecutions')->updateAll(['cron_id' => null], ['cron_id' => $cron->get('id')]);
                $this->fetchTable('Crons')->delete($cron);
                Notify::emit('crons', ['cron_id' => $cron->id, 'deleted' => true]);
            } else {
                $cron->set('active', false);
                $this->fetchTable('Crons')->saveOrFail($cron);
            }
            $this->logOneShot($cron, $output);
        } else {
            $interval = new DateInterval($cron->get('frequency'));
            /** @var DateTime $next */
            $next = $cron->get('next');
            while ($next < $now) {
                $next = $next->add($interval);
            }
            $cron->set('next', $next);
            $this->fetchTable('Crons')->saveOrFail($cron);
        }
        if ($state === CronExecutionsTable::S_ERROR) {
            try {
                $this->sendMail($cron, $exec);
            } catch (Exception $e) {
                $this->io->err(
                    __(
                        "Erreur lors de l'envoi des mails de notification d'erreur: {0} ({1}:{2})",
                        $e->getMessage(),
                        'ROOT' . substr($e->getFile(), strlen(ROOT)),
                        $e->getLine()
                    )
                );
            }
        }
        $status = '<' . $state . '>' . $state . '</' . $state . '>';
        $this->io->out(
            __("Le cron a terminé (execution n°{0}), son statut est {1}", $exec->get('id'), $status)
        );
    }

    /**
     * S'assure de lever le vérrou
     */
    public function __destruct()
    {
        $this->shutdown();
    }

    /**
     * S'assure de lever le vérrou
     */
    public function shutdown()
    {
        if (!empty($this->cron) && $this->cron->get('locked')) {
            $this->cron->set('locked', false);
            try { // fails on testing 'cause the cron_execution table is allready down (sqlite)
                if ($this->cron->get('frequency') === 'one_shot') {
                    $this->fetchTable('CronExecutions')
                        ->updateAll(['cron_id' => null], ['cron_id' => $this->cron->get('id')]);
                    $this->fetchTable('Crons')->delete($this->cron);
                    unset($this->cron);
                    return;
                }
                $now = new DateTime();
                $interval = new DateInterval($this->cron->get('frequency'));
                /** @var DateTime $next */
                $next = $this->cron->get('next');
                while ($next < $now) {
                    $next = $next->add($interval);
                }
                $this->cron->set('next', $next);
                $this->fetchTable('Crons')->saveOrFail($this->cron);

                $this->exec->set('state', CronExecutionsTable::S_ABORTED);
                $this->exec->set('date_end', (new DateTime())->format('Y-m-d H:i:s.u'), ['setter' => false]);
                $this->fetchTable('CronExecutions')->save($this->exec);
                unset($this->cron);
            } catch (Exception) {
            }
        }
    }

    /**
     * Si le debug est activé, renvoi une url pour consulter le mail
     * @param Mailer $email
     * @return CakeResponse|null
     * @throws Exception
     */
    public function debugMail(Mailer $email): ?CakeResponse
    {
        if (Configure::read('debug_mails')) {
            $basePath = Configure::read('debug_mails_basepath', TMP . 'debug_mails');
            $filename = (new DateTime())->format('Ymd_H-i') . uniqid('_sendTestMail_') . '.html';
            Filesystem::dumpFile(
                $basePath . DS . $filename,
                $email->getMessage()->getBodyHtml()
            );
            $url = new UrlHelper(new AppView());
            $request = new ServerRequest();
            $controller = new Controller($request);
            $registry = $controller->components();
            $component = new Component($registry);
            return $component->getController()
                ->getResponse()
                ->withType('txt')
                ->withHeader('X-Email-Debug', 'true')
                ->withStringBody(
                    $url->build('/devs/debug-mail/' . base64_encode($filename), ['fullBase' => true])
                );
        }
        return null;
    }

    /**
     * Permet de consulter le rapport d'un cron "one_shot" dans les logs
     * @param EntityInterface $cron
     * @param string          $output
     * @return void
     */
    private function logOneShot(EntityInterface $cron, string $output): void
    {
        if (is_writable(dirname($cron->get('log_filename')))) {
            file_put_contents($cron->get('log_filename'), $output . "\n", FILE_APPEND);
        }
    }
}
