<?php

/**
 * AsalaeCore\Command\RolesPermsCommand
 */

namespace AsalaeCore\Command;

use AsalaeCore\Application;
use AsalaeCore\Controller\ApiInterface;
use AsalaeCore\Controller\Controller;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Core\Configure;
use Cake\Database\Expression\IdentifierExpression;
use Cake\Datasource\EntityInterface;
use Cake\Utility\Hash;
use Exception;

/**
 * Permet d'exporter/importer les permissions liés aux rôles globaux
 *
 * @category    Command
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2018, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class RolesPermsExportCommand extends Command
{
    use FileTrait;

    /**
     * @var Arguments
     */
    public $args;
    /**
     * @var ConsoleIo
     */
    public $io;

    /**
     * Get the command name.
     * @return string
     */
    public static function defaultName(): string
    {
        return 'roles_perms export';
    }

    /**
     * Gets the option parser instance and configures it.
     *
     * By overriding this method you can configure the ConsoleOptionParser before returning it.
     *
     * @return ConsoleOptionParser
     * @link https://book.cakephp.org/3.0/en/console-and-shells.html#configuring-options-and-generating-help
     */
    public function getOptionParser(): ConsoleOptionParser
    {
        $parserExport = parent::getOptionParser();

        $parserExport->addArgument(
            'destination',
            [
                'help' => __("Emplacement du fichier de destination"),
            ]
        );
        $parserExport->addOption(
            'pretty-print',
            [
                'boolean' => true,
                'help' => __("effectue l'export en mode pretty-print"),
            ]
        );

        return $parserExport;
    }

    /**
     * Export des roles
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return void
     * @throws Exception
     */
    public function execute(Arguments $args, ConsoleIo $io)
    {
        $this->args = $args;
        $this->io = $io;

        $Roles = $this->fetchTable('Roles');
        $Acos = $this->fetchTable('Acos');
        $query = $Roles->find()
            ->where(['Roles.org_entity_id IS' => null])
            ->contain(['Aros' => ['ArosAcos' => ['Acos']]])
            ->orderBy(['Roles.name']);
        $export = [];
        /** @var EntityInterface $role */
        foreach ($query as $role) {
            $export[$role->get('name')] = [];
            /** @var EntityInterface $perm */
            foreach (Hash::get($role, 'aro.aros_acos', []) as $perm) {
                $access = '';
                foreach (['_create', '_read', '_update', '_delete'] as $crud) {
                    $access .= $perm->get($crud) === '1' ? '1' : '0';
                }
                $q = $Acos->find()
                    ->select(['parents.alias'])
                    ->where(['Acos.id' => $perm->get('aco_id')])
                    ->innerJoin(
                        ['parents' => 'acos'],
                        [
                            'parents.lft <=' => new IdentifierExpression(
                                'Acos.lft'
                            ),
                            'parents.rght >=' => new IdentifierExpression(
                                'Acos.rght'
                            ),
                        ]
                    )
                    ->orderByAsc('parents.lft')
                    ->all()
                    ->map(
                        function (EntityInterface $entity) {
                            return $entity->get('parents')['alias'];
                        }
                    );
                $path = implode('/', $q->toArray());
                $export[$role->get('name')][$path] = $access;
            }
        }
        $controllers = [];
        $apis = [];
        $basePath = Configure::read('App.namespace') . '\\Controller\\';
        $classname = $basePath . "AppController";
        if (class_exists($classname)) {
            $appControllerMethods = get_class_methods($classname);
        } else {
            $appControllerMethods = get_class_methods($basePath . "Controller");
        }
        foreach (Application::getControllers() as $controllerClass) {
            /** @var string|Controller $controllerClass */
            $methods = class_exists($controllerClass) ? get_class_methods($controllerClass) : [];
            $arrayName = explode('\\', $controllerClass);
            $name = end($arrayName);
            $controllerName = substr(
                $name,
                0,
                strlen($name) - strlen('Controller')
            );

            // get_class_methods() envoi les methodes protégés dans ce contexte
            if ($controllerName === 'Devs') {
                $methods = ['index'];
            }

            $controllers[$controllerName] = array_diff(
                $methods,
                $appControllerMethods
            );
            sort($controllers[$controllerName]);
            if (
                class_exists($controllerClass)
                && in_array(ApiInterface::class, class_implements($controllerClass))
            ) {
                $apis[$controllerName] = $controllerClass::getApiActions();
            }
        }
        foreach ($export as $role => $actions) {
            ksort($export[$role]);
            $export[$role] = array_filter(
                $export[$role],
                function ($key) use ($controllers, $apis) {
                    if (
                        preg_match(
                            '#^root/api/([^/]+)(?:/([^/]+))?$#',
                            $key,
                            $m
                        )
                    ) {
                        $controller = $m[1];
                        $action = $m[2] ?? 'default';
                        if ($action === 'default') {
                            return isset($apis[$controller]);
                        }
                        return isset($apis[$controller]) && in_array(
                            $action,
                            $apis[$controller]
                        );
                    }
                    if (
                        preg_match(
                            '#^root/controllers/([^/]+)/([^/]+)$#',
                            $key,
                            $m
                        )
                    ) {
                        return isset($controllers[$m[1]]) && in_array(
                            $m[2],
                            $controllers[$m[1]]
                        );
                    }
                    return true;
                },
                ARRAY_FILTER_USE_KEY
            );
        }
        do {
            $file = $args->getArgument('destination') ?: $this->io->ask(
                __("Sélectionner un fichier de sortie"),
                TMP . 'export_roles.json'
            );
            $writtable = true;
            if (!is_writable(dirname($file))) {
                $writtable = false;
                $this->io->err(__("Impossible d'écrire sur {0}", $file));
            }
        } while (!$writtable);
        $json = json_encode(
            $export,
            ($this->args->getOption('pretty-print') ? JSON_PRETTY_PRINT : 0)
            | JSON_UNESCAPED_SLASHES
        );
        $this->createFileWithDefault($file, $json, 'y');
    }
}
