<?php

/**
 * AsalaeCore\Command\RolesPermsCommand
 */

namespace AsalaeCore\Command;

use AsalaeCore\Model\Table\ArosAcosTable;
use AsalaeCore\Model\Table\ArosTable;
use AsalaeCore\Model\Table\RolesTable;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Datasource\EntityInterface;
use Cake\ORM\Table;
use ErrorException;
use Exception;
use Throwable;

/**
 * Permet d'exporter/importer les permissions liés aux rôles globaux
 *
 * @category    Command
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2018, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class RolesPermsImportCommand extends Command
{
    use FileTrait;

    /**
     * @var Arguments
     */
    public $args;
    /**
     * @var ConsoleIo
     */
    public $io;

    private Table|RolesTable $Roles;
    private Table|ArosAcosTable $Permissions;
    private Table|ArosTable $Aros;

    /**
     * Get the command name.
     * @return string
     */
    public static function defaultName(): string
    {
        return 'roles_perms import';
    }

    /**
     * Gets the option parser instance and configures it.
     *
     * By overriding this method you can configure the ConsoleOptionParser before returning it.
     *
     * @return ConsoleOptionParser
     * @link https://book.cakephp.org/3.0/en/console-and-shells.html#configuring-options-and-generating-help
     */
    public function getOptionParser(): ConsoleOptionParser
    {
        $parserImport = parent::getOptionParser();

        $parserImport->addArgument(
            'filename',
            [
                'help' => __("Nom de fichier d'import"),
            ]
        );
        $parserImport->addOption(
            'delete',
            [
                'boolean' => true,
                'help' => __(
                    "Suppression des permissions par défaut (à faire si les permissions par défaut ont changé)"
                ),
            ]
        );
        $parserImport->addOption(
            'keep',
            [
                'boolean' => true,
                'help' => __(
                    "On ne touche pas aux permissions par défaut existantes"
                ),
            ]
        );
        $parserImport->addOption(
            'datasource',
            [
                'help' => __("Datasource utilisé"),
                'default' => 'default',
            ]
        );

        return $parserImport;
    }

    /**
     * Import des permissions de roles
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return void
     * @throws Exception
     */
    public function execute(Arguments $args, ConsoleIo $io)
    {
        $this->args = $args;
        $this->io = $io;

        $filename = $this->args->getArgument('filename');
        while (empty($filename) || !is_readable($filename)) {
            $filename = $this->io->ask(
                __("Veuillez saisir le chemin vers le fichier d'import")
            );
            if (!is_readable($filename)) {
                $this->io->err(__("Impossible de lire le fichier {0}", $filename));
                $filename = null;
            }
        }
        $opts = ['connectionName' => $this->args->getOption('datasource')];
        $import = json_decode(file_get_contents($filename), true);
        $this->Roles = $this->fetchTable('Roles', $opts);
        $this->Aros = $this->fetchTable('Aros', $opts);
        $this->Permissions = $this->fetchTable('ArosAcos', $opts);
        $this->checkAros();
        $conn = $this->Roles->getConnection();
        $conn->begin();

        $error = false;
        $deleteAll = false;
        if ($this->args->getOption('delete')) {
            $deleteAll = true;
        }
        if ($deleteAll) {
            $ids = $this->Roles->find()
                ->innerJoinWith('Aros')
                ->select(['Aros.id'])
                ->where(
                    [
                        'Roles.name IN' => array_keys($import),
                        'org_entity_id IS' => null,
                    ]
                )
                ->innerJoinWith('Aros')
                ->all()
                ->map(
                    function (EntityInterface $entity) {
                        /** @var EntityInterface $aro */
                        $aro = $entity->get('_matchingData')['Aros'];
                        return $aro->get('id');
                    }
                )
                ->toArray();
            if (!empty($ids)) {
                $this->Permissions->deleteAll(['aro_id IN' => $ids]);
                $this->io->out(__("Les permissions ont été supprimées"));
            }
        }
        foreach ($import as $role => $accesses) {
            $roleEntity = $this->Roles->find()
                ->where(['Roles.name' => $role, 'org_entity_id IS' => null])
                ->contain(['Aros'])
                ->first();

            if (!$roleEntity) {
                continue;
            }

            set_error_handler(
                function ($errno, $errstr, $errfile, $errline) {
                    throw new ErrorException(
                        $errstr,
                        0,
                        $errno,
                        $errfile,
                        $errline
                    );
                }
            );
            foreach ($accesses as $acosPath => $access) {
                /** @var string|array $aro Permissions->allow() la doc indique un string
                 *                         mais la fonction autorise les array
                 */
                $aro = [
                    'foreign_key' => $roleEntity->get('id'),
                    'model' => 'Roles',
                ];
                try {
                    $this->allow($access, $aro, $acosPath);
                } catch (Throwable) {
                    $error = true;
                    $this->io->err(
                        __(
                            "Erreur lors de l'import de {0}:{1} l'aco n'existe pas.",
                            $role,
                            $acosPath,
                            $access
                        )
                    );
                }
                $this->io->verbose(
                    $role . ' : ' . $acosPath . ' : ' . $access
                );
            }
        }
        restore_error_handler();

        if ($error) {
            $conn->rollback();
            $this->io->abort(__("Une erreur a été détectée, rollback..."));
        }

        $conn->commit();
        $this->io->out(__("L'import s'est effectué avec succès"));
    }

    /**
     * Ce shell ne peux pas fonctionner si les aros sont cassés
     */
    private function checkAros()
    {
        $this->Roles = $this->fetchTable('Roles');
        $count = $this->Roles->find()
            ->where(['org_entity_id IS' => null, 'Aros.id IS' => null])
            ->leftJoinWith('Aros')
            ->count();
        if ($count) {
            $this->io->abort(__("Il y a {0} roles sans leurs aros", $count));
        }
    }

    /**
     * Crée la permission
     * @param string $access
     * @param array  $aro
     * @param string $acosPath
     * @return void
     * @throws Exception
     */
    protected function allow(string $access, array $aro, string $acosPath): void
    {
        if ($access === '1111' || $access === '1') {
            $this->Permissions->allow($aro, $acosPath);
        } elseif ($access === '0000' || $access === '-1') {
            $this->Permissions->allow($aro, $acosPath, '*', -1);
        } else {
            $subAccesses = [
                '_create' => (bindec($access) & 8) !== 0,
                '_read' => (bindec($access) & 4) !== 0,
                '_update' => (bindec($access) & 2) !== 0,
                '_delete' => (bindec($access) & 1) !== 0,
            ];
            foreach ($subAccesses as $crud => $subAccess) {
                $this->Permissions->allow(
                    $aro,
                    $acosPath,
                    $crud,
                    $subAccess ? 1 : -1
                );
            }
        }
    }
}
