<?php

/**
 * AsalaeCore\Controller\Component\CsvRenderHelper
 */

namespace AsalaeCore\Controller\Component;

use Cake\Datasource\EntityInterface;
use Cake\I18n\Date as CakeDate;
use Cake\I18n\Number;
use Cake\ORM\Query;
use Cake\ORM\ResultSet;
use Cake\Utility\Hash;
use DateTimeInterface;

/**
 * Simule le CsvHelper en rendu direct pour le IndexComponent
 * @category Controller
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2024 Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class CsvRenderHelper
{
    /**
     * @param int    $limit
     * @param string $separator
     * @param string $escapeChar
     * @param string $newLine
     */
    public function __construct(
        public int $limit = 1000,
        public string $separator = ',',
        public string $escapeChar = '"',
        public string $newLine = "\r\n",
    ) {
    }

    /**
     * Génère un CSV de tableau de résultats
     * @param array                                     $cols
     * @param Query|ResultSet|EntityInterface[]|array[] $results
     * @param callable|null                             $map
     * @return string
     */
    public function resultTable(array $cols, $results, callable $map = null): string
    {
        foreach ($cols as $key => $value) {
            $cols[$key] = str_replace('"', '""', $value);
        }

        // LIGNE 1 : Noms de colonnes
        echo $this->escapeChar
            . implode($this->escapeChar . $this->separator . $this->escapeChar, $cols)
            . $this->escapeChar . $this->newLine;

        // LIGNE 2+ : résultats
        if ($results instanceof Query) {
            $count = $results->count();
            $results->limit($this->limit);
            $offset = 0;
            do {
                $results->offset($offset);
                foreach ($results as $entity) {
                    $this->writeEntity($cols, $entity, $map);
                }
                $offset += $this->limit;
            } while ($offset < $count);
        } else {
            foreach ($results as $entity) {
                $this->writeEntity($cols, $entity, $map);
            }
        }
        return '';
    }

    /**
     * Transforme une entité/un array en ligne du CSV avec "echo"
     * @param array         $cols
     * @param mixed         $entity
     * @param callable|null $map
     * @return void
     */
    private function writeEntity(array $cols, mixed $entity, ?callable $map)
    {
        if ($map) {
            $entity = $map($entity);
        }
        $row = [];
        foreach (array_keys($cols) as $path) {
            $value = Hash::get($entity, $path);
            if ($value instanceof CakeDate || $value instanceof DateTimeInterface) {
                $value = (string)$value;
            } elseif (is_object($value)) {
                $value = $this->escapeChar . str_replace('"', '""', $value) . $this->escapeChar;
            } elseif (is_bool($value)) {
                $value = $value ? __("Oui") : __("Non");
            } elseif (is_string($value) && $value) {
                $value = $this->escapeChar . str_replace('"', '""', $value) . $this->escapeChar;
            } elseif (is_float($value)) {
                $value = $this->escapeChar . Number::format(
                    $value,
                    ['precision' => 13, 'pattern' => '0.#############']
                ) . $this->escapeChar;
            }
            $row[] = $value;
        }
        echo implode($this->separator, $row) . $this->newLine;
    }
}
