<?php

/**
 * AsalaeCore\Controller\Component\IndexComponent
 */

namespace AsalaeCore\Controller\Component;

use AsalaeCore\Exception\GenericException;
use AsalaeCore\Http\CallbackStream;
use AsalaeCore\View\Helper\TranslateHelper;
use Cake\Controller\Component;
use Cake\Core\Configure;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Response;
use Cake\ORM\Query;
use Cake\ORM\Query\SelectQuery;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Utility\Inflector;
use Cake\View\View;
use DateTime;
use Exception;
use ReflectionClass;
use ReflectionException;
use Throwable;

/**
 * Pour générer un affichage de résultats
 * @category Controller
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2019 Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 * @property \AsalaeCore\Controller\Component\AjaxPaginatorComponent AjaxPaginator
 * @property \AsalaeCore\Controller\Component\ConditionComponent     Condition
 * @property \AsalaeCore\Controller\Component\FilterComponent        Filter
 * @property \AsalaeCore\Controller\Component\JstableComponent       Jstable
 */
class IndexComponent extends Component
{
    public const string FILTER_COUNTOPERATOR = 'countoperator';
    public const string FILTER_DATECOMPARENOW = 'datecompare';
    public const string FILTER_DATEOPERATOR = 'dateoperator';
    public const string FILTER_FAVORITES = 'favorites';
    public const string FILTER_ILIKE = 'ilike';
    public const string FILTER_IN = 'in';
    public const string FILTER_IS = 'is';
    public const string FILTER_IS_NOT = 'is-not';
    public const string FILTER_NOT_IN = 'not-in';
    public const string FILTER_EQUAL = '=';
    public const string FILTER_GREATER_OR_EQUAL = '>=';
    public const string FILTER_LESS_OR_EQUAL = '<=';
    public const string FILTER_GREATER = '>';
    public const string FILTER_LESS = '<';
    public const string FILTER_NOT_EQUAL = '!=';
    public const string FILTER_SIZEOPERATOR = 'sizeoperator';

    /**
     * Other Components this component uses.
     *
     * @var array
     */
    public array $components = [
        'AsalaeCore.AjaxPaginator',
        'AsalaeCore.Condition',
        'AsalaeCore.Filter',
        'AsalaeCore.Jstable',
    ];

    /**
     * Remplace le CsvHelper de l'object View
     * @var CsvRenderHelper
     */
    public CsvRenderHelper $Csv;

    /**
     * Default config
     *
     * These are merged with user-provided config when the component is used.
     *
     * @var array<string, mixed>
     */
    protected array $_defaultConfig = [
        'csv_filename' => null, // Traduction de Controller/Action si null
    ];
    /**
     * @var Query $query
     */
    private $query;

    /**
     * Initialise un index
     * @param array $config
     * @throws ReflectionException
     */
    public function initialize(array $config = []): void
    {
        parent::initialize($config);
        $this->Csv = new CsvRenderHelper();
        $controller = $this->getController();
        $request = $controller->getRequest();
        $action = $request->getParam('action');
        $reflection = new ReflectionClass($controller);
        $property = $reflection->getProperty('paginate');
        $paginate = $property->getValue($controller);
        $paginate['limit'] = $paginate['limit']
            ?? Configure::read('Pagination.limit', 20);
        $property->setValue($controller, $paginate);
        $viewBuilder = $controller->viewBuilder();
        $model = $this->getConfig('model', $controller->getName());
        $tableId = sprintf(
            '%s-%s-table',
            Inflector::dasherize($model),
            Inflector::dasherize($action)
        );
        $controller->set('tableId', $tableId);
        $this->Jstable->configurablePagination($tableId);
        $this->Filter->loadAndSave();

        if (
            $request->is('ajax')
            && $viewBuilder->getLayout() === null
            && !isset($config['template'])
            && $this->checkAjaxTemplateExist()
        ) {
            $viewBuilder->setTemplate('ajax_' . $action);
        }
    }

    /**
     * @return void
     * @deprecated déplacé dans initialize()
     */
    public function init()
    {
    }

    /**
     * Set le query de l'index pour le component
     * @param Query $query
     * @return $this
     */
    public function setQuery(Query $query)
    {
        $this->query = $query;
        return $this;
    }

    /**
     * Ajoute, pour l'export CSV, la clause order by comme le fait Paginator
     * @return void
     * @throws ReflectionException
     */
    public function addPaginationSort()
    {
        if (empty($this->query)) {
            return;
        }
        $request = $this->getController()->getRequest();
        $sort = $request->getQuery('sort');
        $controller = $this->getController();
        $reflection = new ReflectionClass($controller);
        $property = $reflection->getProperty('paginate');
        $paginate = $property->getValue($controller);
        $direction = $request->getQuery('direction', 'asc');
        if ($sort === 'favorite') {
            $paginate['order'] = $this->Condition->orderByFavorites($direction);
            $property->setValue($controller, $paginate);
            $this->query->orderBy($paginate['order']);
            return;
        }
        if ($this->getConfig('model')) {
            $paginate = $this->getConfig('paginate') ?? [];
        }

        if (!empty($sort)) {
            if ($sortable = Hash::get($paginate, "sortable.$sort")) {
                $sort = $sortable;
            }
            $order = [$sort => $direction];
        } else {
            $order = Hash::get($paginate, 'order');
        }
        if (!$order) {
            return;
        }
        $nOrder = [];
        foreach ($order as $field => $value) {
            if ($sortable = Hash::get($paginate, "sortableFields.$field")) {
                $field = $sortable;
            }
            if (!str_contains($field, '.')) {
                $field = sprintf('%s.%s', $this->query->getRepository()->getAlias(), $field);
            }
            $nOrder[$field] = $value;
        }
        $this->checkAndApplyOrder($nOrder);
    }

    /**
     * Permet de filtrer simplement les résultats d'un index
     * @param string               $urlParam
     * @param string|null|callable $type     type prédéfini ou
     *                                       callback(datavalue, query, fieldname)
     * @param string|null          $field
     * @return $this
     * @throws Exception
     */
    public function filter(string $urlParam, $type = null, string $field = null)
    {
        $q = $this->query;
        if (!$q instanceof Query) {
            throw new Exception(__("\$this->Index->setQuery() n'a pas été appelé avant \$this->Index->filter()"));
        }
        $controller = $this->getController();
        $model = $this->getConfig('model', $controller->getName());
        if ($field === null) {
            $field = str_contains($urlParam, '.') ? $urlParam : "$model.$urlParam";
        }
        $request = $controller->getRequest();
        $params = $request->getQuery($urlParam, []);
        foreach ((array)$params as $i => $value) {
            if (
                $value === ''
                || (is_array($value)
                && (!isset($value['value']) || $value['value'] === '')
                && (!isset($value[0]) || $value[0] === ''))
            ) {
                continue;
            }
            $this->filterParam($i, $value, $urlParam, $type, $field);
        }
        return $this;
    }

    /**
     * Applique les filtres sur un query param
     * @param int             $i
     * @param mixed           $value
     * @param string          $urlParam
     * @param string|callable $type
     * @param string          $field
     * @return void
     * @throws Exception
     */
    private function filterParam(int $i, $value, string $urlParam, $type, string $field)
    {
        $controller = $this->getController();
        $model = $this->getConfig('model', $controller->getName());
        $request = $controller->getRequest();
        $q = $this->query;
        switch ($type) {
            case null:
            default:
                $q->where([$field => $value]);
                break;
            case self::FILTER_DATECOMPARENOW:
                $this->filterDateCompareNow($field, $value);
                break;
            case self::FILTER_DATEOPERATOR:
                $operator = $request->getQuery("dateoperator_$urlParam.$i");
                $this->filterDateOperator($field, $value, $operator);
                break;
            case self::FILTER_FAVORITES:
                $q->where($this->Condition->inFavorites($model . '.id'));
                break;
            case self::FILTER_ILIKE:
                $q->where($this->Condition->ilike($field, $value));
                break;
            case self::FILTER_IN:
                $q->where($this->Condition->in($field, $value));
                break;
            case self::FILTER_IS:
                $q->where([$field . ' IS' => $value]);
                break;
            case self::FILTER_IS_NOT:
                $q->where([$field . ' IS NOT' => $value]);
                break;
            case self::FILTER_NOT_IN:
                $q->where($this->Condition->notIn($field, $value));
                break;
            case self::FILTER_SIZEOPERATOR:
                $this->filterSizeOperator($field, $value);
                break;
            case self::FILTER_COUNTOPERATOR:
                $this->filterCountOperator($field, $value);
                break;
            case self::FILTER_EQUAL:
            case self::FILTER_GREATER:
            case self::FILTER_GREATER_OR_EQUAL:
            case self::FILTER_LESS:
            case self::FILTER_LESS_OR_EQUAL:
            case self::FILTER_NOT_EQUAL:
                $q->where([$field . ' ' . $type => $value]);
                break;
            case is_callable($type):
                $this->filterCallable($type, $field, $value);
                break;
        }
    }

    /**
     * Condition ["$field $operator" => now]
     * @param string $field
     * @param string $operator
     * @return void
     * @throws Exception
     */
    private function filterDateCompareNow(string $field, string $operator)
    {
        $q = $this->query;
        if (!in_array($operator, ['=', '<=', '>=', '<', '>'])) {
            throw new BadRequestException(__("L'opérateur indiqué n'est pas autorisé"));
        }
        $q->where($this->Condition->dateOperator($field, $operator, new DateTime()));
    }

    /**
     * Compare une quantité
     * @param string $field
     * @param string $value
     * @param string $operator
     * @return void
     * @throws Exception
     */
    private function filterDateOperator(string $field, $value, $operator)
    {
        $q = $this->query;
        if ($value) {
            $q->where($this->Condition->dateOperator($field, $operator, $value));
        }
    }

    /**
     * Compare une taille
     * @param string $field
     * @param string $values
     * @return void
     */
    private function filterSizeOperator(string $field, $values)
    {
        $q = $this->query;
        if (!is_array($values) || !isset($values['value'])) {
            return;
        }
        $v = $values['value'] * $values['mult'];
        $operator = $values['operator'];
        if (!in_array($operator, ['=', '<=', '>=', '<', '>'])) {
            throw new BadRequestException(__("L'opérateur indiqué n'est pas autorisé"));
        }
        $q->where([$field . ' ' . $operator => $v]);
    }

    /**
     * Compare une quantité
     * @param string $field
     * @param string $values
     * @return void
     */
    private function filterCountOperator(string $field, $values)
    {
        $q = $this->query;
        if (!is_array($values) || !isset($values['value'])) {
            return;
        }
        $operator = $values['operator'];
        if (!in_array($operator, ['=', '<=', '>=', '<', '>'])) {
            throw new BadRequestException(__("L'opérateur indiqué n'est pas autorisé"));
        }
        $q->where([$field . ' ' . $operator => $values['value']]);
    }

    /**
     * Compare une quantité
     * @param callable $type
     * @param string   $field
     * @param string   $value
     * @return void
     */
    private function filterCallable(callable $type, string $field, $value)
    {
        $q = $this->query;
        $call = call_user_func($type, $value, $q, $field);
        if (is_array($call)) {
            $q->where($call);
        }
    }

    /**
     * Bouton Télécharger en CSV / rendu CSV
     * Nécéssite un template en templates/<Controller>/csv/<action>.php
     * @return Response|void
     * @throws ReflectionException
     */
    public function exportCsv()
    {
        $controller = $this->getController();
        $request = $controller->getRequest();
        if ($request->getQuery('export_csv')) {
            return $this->renderCsv();
        } else {
            $uri = $controller->getRequest()->getUri();
            $queryParams = $uri->getQuery();
            if ($queryParams) {
                $queryParams .= '&';
            }
            $queryParams .= 'export_csv=true';
            $downloadCsv = [
                'href' => (string)$uri->withQuery($queryParams),
                'download' => $this->getCsvFilename(),
            ];
            $controller->set('downloadCsv', $downloadCsv);
        }
    }

    /**
     * Rendu CSV
     * @param null|Query|array $data
     * @return Response
     * @throws ReflectionException
     * @see IndexComponent::exportCsv()
     */
    public function renderCsv($data = null)
    {
        $controller = $this->getController();
        $viewBuilder = $controller->viewBuilder();
        if (empty($data)) {
            $data = $this->query;
        }
        $controller->set('results', $data);
        $this->addPaginationSort();
        $viewBuilder->setLayout('ajax');

        return $controller->getResponse()
            ->withType('csv')
            ->withHeader(
                'Content-Disposition',
                'attachment; filename="' . $this->getCsvFilename() . '"'
            )
            ->withBody(
                new CallbackStream(
                    function () {
                        ob_end_flush();
                        ob_implicit_flush();
                        set_time_limit(0);
                        $this->readCsv();
                    }
                )
            );
    }

    /**
     * Donne le filename du template qui sera utilisé lors du controller->render()
     * @param string|null $template
     * @return string
     * @throws ReflectionException
     */
    public function getViewTemplateFilename(?string $template = null): string
    {
        $builder = $this->getController()->viewBuilder();
        $request = $this->getController()->getRequest();
        if (!$builder->getTemplatePath()) {
            $templatePath = $this->getController()->getName();
            if ($request->getParam('prefix')) {
                $prefixes = array_map(
                    'Cake\Utility\Inflector::camelize',
                    explode('/', $request->getParam('prefix'))
                );
                $templatePath = implode(DIRECTORY_SEPARATOR, $prefixes)
                    . DIRECTORY_SEPARATOR . $templatePath;
            }
            $builder->setTemplatePath($templatePath);
        }
        if ($template !== null) {
            $builder->setTemplate($template);
        }
        if ($builder->getTemplate() === null) {
            $builder->setTemplate($request->getParam('action'));
        }
        $view = $builder->build($request, $this->getController()->getResponse());
        $reflection = new ReflectionClass($view);
        $method = $reflection->getMethod('_getTemplateFileName');
        return $method->invoke($view, $template);
    }

    /**
     * Donne le filename du csv sous la forme
     * "YYYY-MM-DD_HHMMSS_NomActionFrançisé.csv"
     * @return string
     */
    private function getCsvFilename()
    {
        $request = $this->getController()->getRequest();
        $traductionHelper = new TranslateHelper(new View());
        $trad = Inflector::camelize(
            $traductionHelper->permission(
                $request->getParam('controller'),
                $request->getParam('action') === 'index'
                    ? null
                    : $request->getParam('action')
            )
        );
        return sprintf(
            '%s_%s.csv',
            (new DateTime())->format('Y-m-d_His'),
            $this->getConfig('csv_filename', $trad)
        );
    }

    /**
     * Vrai si un template existe préfixé par "ajax_" (ex: ajax_index)
     * @return bool
     */
    private function checkAjaxTemplateExist(): bool
    {
        $controller = $this->getController();
        $request = $controller->getRequest();
        $action = Inflector::underscore($request->getParam('action'));
        $templatesBasePath = Configure::read('App.paths.templates');
        $template = $controller->getName() . '/ajax_' . $action . '.php';
        foreach ($templatesBasePath as $basePath) {
            if (file_exists($basePath . $template)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Evite les erreurs en cas de tri impossible
     * @param array $nOrder
     * @return void
     * @throws Exception
     */
    private function checkAndApplyOrder(array $nOrder)
    {
        $models = $this->extractModels($this->query);
        $table = $this->query->getRepository();
        $alias = $table->getAlias();
        $schemas = [$alias => $table->getSchema()];
        $loc = TableRegistry::getTableLocator();
        foreach ($nOrder as $modelField => $value) {
            [$modelFieldWithoutOperator] = explode(' ', $modelField);
            [$model, $field] = explode('.', $modelFieldWithoutOperator);
            if (!in_array($model, $models)) {
                throw new GenericException(__("Model manquant dans la query ({0})", $model));
            }
            if (!isset($schemas[$model])) {
                $schemas[$model] = $loc->get($model)->getSchema();
            }
            if (!$schemas[$model]->hasColumn($field)) {
                trigger_error(__("Tri sur champ inconnu : {0}", $modelFieldWithoutOperator));
                unset($nOrder[$modelField]);
            }
        }
        if ($nOrder) {
            $this->query->orderBy($nOrder);
        }
    }

    /**
     * Donne une liste de noms de models
     * @param SelectQuery $query
     * @return string[]
     */
    public static function extractModels(SelectQuery $query): array
    {
        $eagerLoader = $query->getEagerLoader();
        $models = array_merge(
            [$query->getRepository()->getAlias()],
            self::extractNestedModels($eagerLoader->getContain()),
            self::extractNestedModels($eagerLoader->getMatching()),
        );
        return array_unique($models);
    }

    /**
     * Fonction récursive pour récupérer les models dans "matching"
     * @param array $matching
     * @return array
     */
    private static function extractNestedModels(array $matching): array
    {
        $models = [];
        foreach ($matching as $model => $data) {
            $models[] = $model;
            foreach ($data as $key => $value) {
                if (!is_array($value)) {
                    continue;
                }
                $nestedMatching = [$key => $value];
                $models = array_merge($models, self::extractNestedModels($nestedMatching));
            }
        }
        return $models;
    }

    /**
     * Simule un render directement dans la sortie sans passer dans la RAM
     * @return void
     * @throws ReflectionException
     */
    private function readCsv(): void
    {
        $controller = $this->getController();
        $request = $controller->getRequest();
        $action = Inflector::underscore($request->getParam('action'));
        $viewBuilder = $controller->viewBuilder();
        $template = $this->getViewTemplateFilename($viewBuilder->getTemplate() ?: 'csv/' . $action);
        $dir = getcwd();
        chdir(dirname($template));

        extract($viewBuilder->getVars());
        try {
            include $template; // NOSONAR pas de include_once ici
        } catch (Throwable $exception) {
            while (ob_get_level() > $bufferLevel) {
                ob_end_clean();
            }

            throw $exception;
        } finally {
            chdir($dir);
        }
    }
}
