<?php

/**
 * AsalaeCore\Controller\Controller
 */

namespace AsalaeCore\Controller;

use AsalaeCore\Controller\Component\AjaxPaginatorComponent;
use AsalaeCore\Controller\Component\AuthorizationComponent;
use AsalaeCore\Controller\Component\ConditionComponent;
use AsalaeCore\Controller\Component\CookieComponent;
use AsalaeCore\Controller\Component\FilterComponent;
use AsalaeCore\Controller\Component\IndexComponent;
use AsalaeCore\Controller\Component\JstableComponent;
use AsalaeCore\Controller\Component\ModalComponent;
use AsalaeCore\Controller\Component\ReaderComponent;
use AsalaeCore\Controller\Component\RestComponent;
use AsalaeCore\Error\ErrorTrap;
use AsalaeCore\Http\AppServerRequest;
use AsalaeCore\Http\Response as CoreResponse;
use AsalaeCore\Model\Table\FiltersTable;
use AsalaeCore\Model\Table\SavedFiltersTable;
use Authentication\Controller\Component\AuthenticationComponent;
use Cake\Controller\Controller as CakeController;
use Cake\Core\Configure;
use Cake\Database\Query;
use Cake\Datasource\ResultSetInterface;
use Cake\Error\Debugger;
use Cake\Error\PhpError;
use Cake\Event\EventManagerInterface;
use Cake\Http\ServerRequest;
use Cake\Log\Log;
use Cake\ORM\ResultSet;
use Exception;
use Psr\Http\Message\ResponseInterface;

/**
 * CoreController
 *
 * @category Controller
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2020, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 * @property AjaxPaginatorComponent  AjaxPaginator
 * @property AppServerRequest        request
 * @property AuthenticationComponent Authentication
 * @property AuthorizationComponent  Authorization
 * @property ConditionComponent      Condition
 * @property CookieComponent         Cookie
 * @property FilterComponent         Filter
 * @property FiltersTable            Filters
 * @property IndexComponent          Index
 * @property JstableComponent        Jstable
 * @property ModalComponent          Modal
 * @property ReaderComponent         Reader
 * @property RestComponent           Rest
 * @property SavedFiltersTable       SavedFilters
 */
class Controller extends CakeController
{
    /**
     * @var array
     */
    public $cookies = [];
    /**
     * @var array
     */
    public array $paginate = [];
    /**
     * @var ServerRequest
     */
    protected ServerRequest $request;
    /**
     * @var AuthenticationComponent
     */
    protected AuthenticationComponent $Authentication;

    /**
     * Constructor.
     *
     * Sets a number of properties based on conventions if they are empty. To override the
     * conventions CakePHP uses you can define properties in your class declaration.
     *
     * @param ServerRequest|null         $request      Request object for this controller.
     *                                                 but expect that features that use
     *                                                 the request parameters will not
     *                                                 work.
     * @param string|null                $name         Override the name useful in testing when using mocks.
     * @param EventManagerInterface|null $eventManager The event manager. Defaults to a new instance.
     */
    public function __construct(
        ?ServerRequest $request = null,
        ?string $name = null,
        ?EventManagerInterface $eventManager = null,
    ) {
        parent::__construct($request, $name, $eventManager);
        $this->setResponse(new CoreResponse());
    }

    /**
     * Donne les informations sur l'api du controller (actions => accès)
     * @return array
     */
    public static function getApiActions(): array
    {
        return ['default'];
    }

    /**
     * Donne la pagination sous forme de ResultSet (comme avant cake 5)
     * @param mixed $object   query ou modelName
     * @param array $settings
     * @return ResultSetInterface
     * @throws Exception
     */
    public function paginateToResultSet($object = null, array $settings = []): ResultSetInterface
    {
        if ($object instanceof Query && $this->components()->has('Index')) {
            $this->Index->addPaginationSort();
        }
        $paginated = $this->paginate($object, $settings);
        $this->set('paginated', $paginated);
        $data = iterator_to_array($paginated->items());
        return new ResultSet($data);
    }

    /**
     * Perform the various shutdown processes for this controller.
     * Fire the Components and Controller callbacks in the correct order.
     *
     * - triggers the component `shutdown` callback.
     * - calls the Controller's `afterFilter` method.
     *
     * @return \Psr\Http\Message\ResponseInterface|null
     */
    public function shutdownProcess(): ?ResponseInterface
    {
        $request = $this->getRequest();
        if ($request->is('json') && ($errors = ErrorTrap::getErrors())) {
            // supprime le contexte qui peut causer des problèmes de mémoire
            foreach ($errors as $key => $values) {
                $values['context'] = null;
                $errors[$key] = $values;
            }
            ErrorTrap::enableErrorOutput();
            $response = $this->getResponse();
            $body = $response->getBody();
            $body->rewind();
            $content = $body->getContents();
            $json = json_decode($content, true);
            if ($json && is_array($json)) {
                $json['php_error'] = $errors;
                $body->rewind();
                $body->write(json_encode($json));
                return $response;
            } elseif (!$json) {
                $debug = Configure::read('debug');
                $errorTrap = new ErrorTrap();
                $renderer = $errorTrap->renderer();
                foreach ($errors as $error) {
                    $error = new PhpError(
                        $error['code'],
                        $error['description'],
                        $error['file'] ?? null,
                        $error['line'] ?? null,
                        $error['trace'] ?? []
                    );
                    $renderer->write($renderer->render($error, $debug));
                }
            } else {
                Log::error(Debugger::exportVar($errors));
            }
        }
        return parent::shutdownProcess();
    }
}
