<?php

/**
 * AsalaeCore\Controller\FiltersControllerTrait
 */

namespace AsalaeCore\Controller;

use AsalaeCore\Controller\Component\FilterComponent;
use AsalaeCore\Controller\Component\ModalComponent;
use AsalaeCore\Model\Entity\Filter;
use AsalaeCore\Utility\FormatError;
use Cake\Http\Exception\UnauthorizedException;
use Cake\Http\Response;
use Cake\Utility\Hash;
use Cake\View\Helper\UrlHelper;
use Cake\View\View;
use Exception;

/**
 * Filtres de rechercher
 *
 * @category Controller
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
trait FiltersControllerTrait
{
    /**
     * Permet de rediriger vers l'url construite à partir des filtres sauvegardés
     * @param string $id
     * @return Response|null
     * @throws Exception
     */
    public function ajaxRedirectUrl(string $id)
    {
        $this->getRequest()->allowMethod('ajax');
        return $this->redirect($this->getUrlById($id));
    }

    /**
     * Permet de récupérer une url à partir des filtres sauvegardé
     * @param string $id
     * @return string
     * @throws Exception
     */
    private function getUrlById(string $id): string
    {
        $request = $this->getRequest();
        $identity = $request->getAttribute('identity');
        $user = $identity ? $identity->getOriginalData() : null;
        if (!$user) {
            throw new UnauthorizedException();
        }
        $SavedFilters = $this->fetchTable('SavedFilters');
        $save = $SavedFilters->find()
            ->where(
                [
                    'SavedFilters.id' => $id,
                    'user_id' => Hash::get($user, 'id'),
                ]
            )
            ->contain(['Filters'])
            ->first();

        if (!$save) {
            throw new Exception(__("Valeur non trouvée"));
        }

        $UrlHelper = new UrlHelper(new View());
        $params = [
            'controller' => $save->get('controller'),
            'action' => $save->get('action'),
            '?' => ["SaveFilterSelect" => $id],
        ];
        /** @var Filter $filter */
        foreach ($save->get('filters') as $filter) {
            $params['?'][$filter->get('key')] = $filter->get('value');
        }

        return $UrlHelper->build($params, ['escape' => false, 'fullBase' => true]);
    }

    /**
     * Permet de supprimer une sauvegarde de filtre de recherche
     * @param string $id
     * @throws Exception
     */
    public function ajaxDeleteSave(string $id)
    {
        $request = $this->getRequest();
        $identity = $request->getAttribute('identity');
        $user = $identity ? $identity->getOriginalData() : null;
        if (!$user) {
            throw new UnauthorizedException();
        }
        $request->allowMethod('ajax');
        $SavedFilters = $this->fetchTable('SavedFilters');

        $viewBuild = $this->viewBuilder();
        $viewBuild->setClassName('Json');

        $save = $SavedFilters->find()
            ->where(
                [
                    'SavedFilters.id' => $id,
                    'user_id' => Hash::get($user, 'id'),
                ]
            )
            ->contain(['Filters'])
            ->first();
        if (empty($save)) {
            throw new Exception(__("La sauvegarde du filtre n'a pas été trouvée"));
        }
        $success = $SavedFilters->delete($save);
        /** @var ModalComponent $Modal */
        $Modal = $this->loadComponent('AsalaeCore.Modal');
        if ($success) {
            $Modal->success();
        } else {
            $Modal->fail();
            FormatError::logEntityErrors($save);
        }

        $this->set('success', $success);
        $this->set('_serialize', ['success']);
    }

    /**
     * @param string $id
     * @throws Exception
     */
    public function ajaxGetFilters(string $id)
    {
        $request = $this->getRequest();
        $identity = $request->getAttribute('identity');
        $user = $identity ? $identity->getOriginalData() : null;
        if (!$user) {
            throw new UnauthorizedException();
        }
        $request->allowMethod('ajax');
        $SavedFilters = $this->fetchTable('SavedFilters');
        $save = $SavedFilters->find()
            ->where(
                [
                    'SavedFilters.id' => $id,
                    'user_id' => Hash::get($user, 'id'),
                ]
            )
            ->contain(['Filters'])
            ->first();

        if (!$save) {
            throw new Exception("Valeur non trouvée");
        }

        $this->set('savedFilters', $save);
        $this->set('helperConfig', json_decode($request->getData('helper') ?: '', true));
    }

    /**
     * Enregistre un jeu de filtres
     */
    public function ajaxNewSave()
    {
        $request = $this->getRequest();
        $request->allowMethod('ajax');
        $request->allowMethod('post');
        $SavedFilters = $this->fetchTable('SavedFilters');

        /** @var ModalComponent $Modal */
        $Modal = $this->loadComponent('AsalaeCore.Modal');
        /** @var FilterComponent $Filter */
        $Filter = $this->loadComponent('AsalaeCore.Filter');
        $data = $request->getData();
        $name = $data['savename'];
        $controller = $data['controller'];
        $action = $data['action'];
        $filters = $Filter->dataToFilters(array_filter($data, 'is_array'));
        if (empty($filters)) {
            $Modal->fail();
            $report = "Empty filters";
        } else {
            $identity = $request->getAttribute('identity');
            $user = $identity ? $identity->getOriginalData() : null;
            if (!$user) {
                throw new UnauthorizedException();
            }
            $entity = $SavedFilters->newEntity(
                compact('name', 'controller', 'action') + [
                    'user_id' => Hash::get($user, 'id'),
                    'filters' => $filters,
                ],
                ['associated' => 'Filters']
            );
            if ($SavedFilters->save($entity)) {
                $Modal->success();
                $report = $entity->toArray();
            } else {
                $Modal->fail();
                $report = $entity->getErrors();
                FormatError::logEntityErrors($entity);
            }
        }
        return AbstractRenderData::staticRenderData($this, $report);
    }

    /**
     * Remplace les filtres d'une sauvegarde
     * @param string $id
     * @return Response
     * @throws Exception
     */
    public function ajaxOverwrite(string $id)
    {
        $request = $this->getRequest();
        $identity = $request->getAttribute('identity');
        $user = $identity ? $identity->getOriginalData() : null;
        if (!$user) {
            throw new UnauthorizedException();
        }
        $request->allowMethod('ajax');
        $request->allowMethod('post');
        $SavedFilters = $this->fetchTable('SavedFilters');
        $Filters = $this->fetchTable('Filters');

        /** @var ModalComponent $Modal */
        $Modal = $this->loadComponent('AsalaeCore.Modal');
        /** @var FilterComponent $Filter */
        $Filter = $this->loadComponent('AsalaeCore.Filter');
        $data = $request->getData();
        $filters = $Filter->dataToFilters(array_filter($data, 'is_array'));

        if (empty($filters)) {
            $Modal->fail();
            $report = "Empty filters";
        } else {
            $entity = $SavedFilters->find()
                ->where(['id' => $id, 'user_id' => Hash::get($user, 'id')])
                ->contain(['Filters'])
                ->firstOrFail();
            $Filters->deleteAll(
                [
                    'id IN' => Hash::extract($entity->get('filters'), '{n}.id'),
                ]
            );
            $SavedFilters->patchEntity($entity, compact('filters'));
            if ($SavedFilters->save($entity, ['associated' => 'Filters'])) {
                $Modal->success();
                $report = $entity->toArray();
            } else {
                $Modal->fail();
                $report = $entity->getErrors();
                FormatError::logEntityErrors($entity);
            }
        }
        return AbstractRenderData::staticRenderData($this, $report);
    }
}
