<?php

/**
 * AsalaeCore\Database\Type\DateTimeType
 */

namespace AsalaeCore\Database\Type;

use AsalaeCore\I18n\Date;
use Cake\Chronos\ChronosDate;
use Cake\Database\Driver;
use Cake\Database\Type\BatchCastingInterface;
use Cake\Database\Type\DateType as CakeDateType;
use Cake\Database\TypeInterface;
use Cake\I18n\Date as CakeDate;
use DateTimeInterface;
use Exception;
use InvalidArgumentException;

/**
 * Permet la triple compatibilité de dates : format DATE_RFC3339, format local et objet
 *
 * @category Database
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2020, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class DateType extends CakeDateType implements TypeInterface, BatchCastingInterface
{
    /**
     * Constructor
     * @param string|null $name The name identifying this type
     */
    public function __construct(?string $name = null)
    {
        parent::__construct($name);
        $this->_className = Date::class;
    }

    /**
     * Sets whether to parse strings passed to `marshal()` using
     * the locale-aware format set by `setLocaleFormat()`.
     *
     * @param bool $enable Whether to enable
     * @return $this
     */
    public function useLocaleParser(bool $enable = true)
    {
        $this->_useLocaleMarshal = $enable;
        return $this;
    }

    /**
     * Surcharge pour renvoyer un ChronosDate (avec toChronosDate())
     *
     * @param mixed  $value  Value to be converted to PHP equivalent
     * @param Driver $driver Object from which database preferences and configuration will be extracted
     * @return ChronosDate|null
     */
    public function toPHP(mixed $value, Driver $driver): ?ChronosDate
    {
        if ($value === null) {
            return null;
        }

        $class = $this->_className;
        if (is_int($value)) {
            $instance = new $class('@' . $value);
        } elseif (str_starts_with($value, '0000-00-00')) {
            return null;
        } else {
            $instance = new $class($value);
        }

        return $this->toChronosDate($instance);
    }

    /**
     * Donne la version cake
     * @param mixed $instance
     * @return CakeDate
     */
    private function toChronosDate(mixed $instance): CakeDate
    {
        if ($instance instanceof Date) {
            return $instance->getCakeDate();
        }
        return $instance;
    }

    /**
     * Surcharge pour renvoyer un ChronosDate (avec toChronosDate())
     *
     * @param mixed $value Request data
     * @return ChronosDate|null
     */
    public function marshal(mixed $value): ?ChronosDate
    {
        if ($value instanceof $this->_className) {
            return $this->toChronosDate($value);
        }

        /** @phpstan-ignore-next-line */
        if ($value instanceof DateTimeInterface || $value instanceof ChronosDate) {
            return $this->toChronosDate(new $this->_className($value->format($this->_format)));
        }

        $class = $this->_className;
        try {
            if (is_int($value) || (is_string($value) && ctype_digit($value))) {
                return new $class('@' . $value);
            }

            if (is_string($value)) {
                if ($this->_useLocaleMarshal) {
                    return $this->_parseLocaleValue($value);
                }

                return $this->_parseValue($value);
            }
        } catch (Exception) {
            return null;
        }

        if (
            !is_array($value) ||
            !isset($value['year'], $value['month'], $value['day']) ||
            !is_numeric($value['year']) || !is_numeric($value['month']) || !is_numeric($value['day'])
        ) {
            return null;
        }

        $format = sprintf('%d-%02d-%02d', $value['year'], $value['month'], $value['day']);

        return $this->toChronosDate(new $class($format));
    }

    /**
     * @param string $value
     * @return CakeDate|null
     */
    protected function _parseLocaleValue(string $value): ?CakeDate
    {
        if (!$value) {
            return null;
        }
        /** @var class-string<CakeDate> $class */
        $class = $this->_className;

        return $this->toChronosDate($class::parseDate($value, $this->_localeMarshalFormat));
    }

    /**
     * Surcharge pour renvoyer un ChronosDate (avec toChronosDate())
     *
     * @param string $value The value to parse and convert to an object.
     * @return ChronosDate|null
     */
    protected function _parseValue(string $value): ?ChronosDate
    {
        $class = $this->_className;
        foreach ($this->_marshalFormats as $format) {
            try {
                return $this->toChronosDate($class::createFromFormat($format, $value));
            } catch (InvalidArgumentException) {
                continue;
            }
        }

        return null;
    }
}
