<?php

/**
 * AsalaeCore\Error\ErrorLogger
 */

namespace AsalaeCore\Error;

use AsalaeCore\Http\AppServerRequest;
use Authorization\Exception\ForbiddenException;
use Cake\Controller\Exception\MissingActionException;
use Cake\Error\ErrorLogger as CakeErrorLogger;
use Cake\Error\PhpError;
use Cake\Http\Exception\MissingControllerException;
use Cake\Http\Exception\UnauthorizedException;
use Cake\Log\Log;
use Psr\Http\Message\ServerRequestInterface;
use Throwable;

/**
 * Log les erreurs de l'application
 *
 * @category    Error
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2024, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class ErrorLogger extends CakeErrorLogger
{
    /**
     * Log an error to Cake's Log subsystem
     *
     * @param PhpError                $error        The error to log
     * @param ?ServerRequestInterface $request      The request if in an HTTP context.
     * @param bool                    $includeTrace Should the log message include a stacktrace
     * @return void
     */
    public function logError(PhpError $error, ?ServerRequestInterface $request = null, bool $includeTrace = false): void
    {
        $message = $error->getMessage();
        if (str_starts_with($message, 'DbAcl::check() - Failed ACO node lookup in permissions check')) {
            $aro = '';
            if (preg_match('/\[username] => (\S+)/', $message, $m)) {
                $aro = "user:$m[1] ";
            }
            if (preg_match('/\[role] => .*\n *\(\n *\[id].*\n *\[name] => (\S+)/', $message, $m)) {
                $aro .= "role:$m[1]";
            }
            if ($aro && preg_match('/Aco: (\S+)/', $message, $m)) {
                $message = "This permission does not exists - Aro: $aro ; Aco: $m[1]";
                Log::write($error->getLabel(), $message);
                return;
            }
        }

        parent::logError($error, $request, $includeTrace);
    }

    /**
     * Log an exception to Cake's Log subsystem
     *
     * @param \Throwable                  $exception    The exception to log a message for.
     * @param ServerRequestInterface|null $request      The current request if available.
     * @param bool                        $includeTrace Whether or not a stack trace should be logged.
     * @return void
     */
    public function logException(
        Throwable $exception,
        ?ServerRequestInterface $request = null,
        bool $includeTrace = false
    ): void {
        $message = $exception->getMessage();
        if ($exception instanceof MissingControllerException || $exception instanceof MissingActionException) {
            if ($request) {
                $message = sprintf(
                    '%d - %s could not be found.',
                    $exception->getCode() ?: 404,
                    $request->getRequestTarget()
                );
                $message = $this->appendUserInfo($message, $request);
                $message = $this->appendClientIp($message, $request);
            }
            Log::error($message);
            return;
        }
        if ($exception instanceof ForbiddenException) {
            if ($request) {
                $message = sprintf(
                    '%d - %s access denied.',
                    $exception->getCode() ?: 403,
                    $request->getRequestTarget()
                );
                $message = $this->appendUserInfo($message, $request);
                $message = $this->appendClientIp($message, $request);
            }
            Log::error($message);
            return;
        }
        if ($exception instanceof UnauthorizedException) {
            if ($request) {
                $message = sprintf(
                    '%d - %s requires authentication.',
                    $exception->getCode() ?: 401,
                    $request->getRequestTarget()
                );
                $message = $this->appendClientIp($message, $request);
            }
            Log::error($message);
            return;
        }
        parent::logException($exception, $request, $includeTrace);
    }

    /**
     * Ajoute l'ip et l'url referer
     * @param string                 $message
     * @param ServerRequestInterface $request
     * @return string
     */
    private function appendClientIp(string $message, ServerRequestInterface $request): string
    {
        $referer = $request->getHeaderLine('Referer');
        if ($referer) {
            $message .= "\n    Referer URL: " . $referer;
        }

        if (method_exists($request, 'clientIp')) {
            $clientIp = $request->clientIp();
            if ($clientIp && $clientIp !== '::1') {
                $message .= "\n    Client IP: " . $clientIp;
            }
        }

        return $message;
    }

    /**
     * Ajoute le username de l'utilisateur connecté
     * @param string                 $message
     * @param ServerRequestInterface $request
     * @return string
     */
    private function appendUserInfo(string $message, ServerRequestInterface $request)
    {
        if (!$request instanceof AppServerRequest) {
            return $message;
        }
        $session = $request->getSession();
        $adminUsername = $session->read('Admin.data.username');
        if ($adminUsername) {
            return $message . ' admin user: ' . $adminUsername;
        }
        $username = $session->read('Auth.username');
        if ($username) {
            return $message . ' user: ' . $username;
        }
        return $message;
    }
}
