<?php

/**
 * AsalaeCore\Error\ErrorTrap
 */

namespace AsalaeCore\Error;

use AsalaeCore\Error\Renderer\ConsoleErrorRenderer;
use AsalaeCore\Error\Renderer\HtmlErrorRenderer;
use Cake\Core\Configure;
use Cake\Error\Debugger;
use Cake\Error\ErrorLogger;
use Cake\Error\ErrorTrap as CakeErrorTrap;
use Cake\Error\FatalErrorException;
use Cake\Error\PhpError;
use Cake\Log\Log;

/**
 * Gestions d'erreurs de l'application
 *
 * @category    Error
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2022, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class ErrorTrap extends CakeErrorTrap
{
    /**
     * @var bool Affiche l'erreur si le debug > 2
     */
    private static bool $errorOutput = true;

    /**
     * @var array stock les erreurs si l'affichage est désactivé
     */
    private static $errors = [];

    /**
     * Configuration options. Generally these are defined in config/app.php
     *
     * - `errorLevel` - int - The level of errors you are interested in capturing.
     * - `errorRenderer` - string - The class name of render errors with. Defaults
     *   to choosing between Html and Console based on the SAPI.
     * - `log` - boolean - Whether or not you want errors logged.
     * - `logger` - string - The class name of the error logger to use.
     * - `trace` - boolean - Whether or not backtraces should be included in
     *   logged errors.
     *
     * @var array<string, mixed>
     */
    protected array $_defaultConfig = [
        'errorLevel' => E_ALL,
        'errorRenderer' => null,
        'log' => true,
        'logger' => ErrorLogger::class,
        'trace' => false,
    ];

    /**
     * Choose an error renderer based on config or the SAPI
     *
     * @return class-string<ConsoleErrorRenderer|HtmlErrorRenderer>
     */
    protected function chooseErrorRenderer(): string
    {
        $config = $this->getConfig('errorRenderer');
        if ($config !== null) {
            return $config;
        }

        /** @var class-string<ConsoleErrorRenderer|HtmlErrorRenderer> */
        return PHP_SAPI === 'cli' ? ConsoleErrorRenderer::class : HtmlErrorRenderer::class;
    }

    /**
     * Logging helper method.
     *
     * @param PhpError $error The error object to log.
     * @return void
     */
    protected function logError(PhpError $error): void
    {
        $stdout = Log::getConfig('stdout');
        $stderr = Log::getConfig('stderr');
        Log::drop('stdout');
        Log::drop('stderr');
        parent::logError($error);
        if (PHP_SAPI === 'cli' && $stdout && $stderr) {
            Log::setConfig('stdout', $stdout);
            Log::setConfig('stderr', $stderr);
        }
    }

    /**
     * Désactive l'affichage des erreurs (elles sont stockés dans self::$errors)
     */
    public static function disableErrorOutput()
    {
        static::$errorOutput = false;
    }

    /**
     * Active l'affichage des erreurs
     */
    public static function enableErrorOutput()
    {
        static::$errorOutput = true;
    }

    /**
     * Permet de capturer les erreurs d'une fonction en particulier
     *
     * @param callable $callback la fonction a executer sans erreurs
     * @param mixed    ...$args  arguments à envoyer
     *                           au callback
     * @return array ['result' => 'résultat du callback', 'errors' => [errors]]
     */
    public static function captureErrors(callable $callback, ...$args): array
    {
        // hack pour éviter un exit dans le ExceptionTrap
        $oldEnv = getenv('FIXTURE_SCHEMA_METADATA');
        putenv('FIXTURE_SCHEMA_METADATA=true');

        $instance = new self(Configure::read('Error'));
        $level = $instance->getConfig('errorLevel') ?: -1;
        error_reporting($level); // NOSONAR
        set_error_handler([$instance, 'handleError'], $level);
        set_exception_handler([ExceptionTrap::instance(), 'handleException']);
        static::disableErrorOutput();
        $oldErrors = static::$errors;
        static::$errors = [];
        $result = call_user_func_array($callback, $args);
        $errors = static::$errors;
        static::$errors = $oldErrors;
        static::enableErrorOutput();
        restore_error_handler();
        restore_exception_handler();

        if ($oldEnv) {
            putenv('FIXTURE_SCHEMA_METADATA=' . $oldEnv);
        } else {
            putenv('FIXTURE_SCHEMA_METADATA');
        }
        return compact('result', 'errors');
    }

    /**
     * Permet de savoir si les erreurs ont été désactivés/capturés
     * @return bool
     */
    public static function isErrorDisabled(): bool
    {
        return static::$errorOutput === false;
    }

    /**
     * Donne les erreurs capturés
     * @return array
     */
    public static function getErrors(): array
    {
        $errors = static::$errors;
        static::$errors = [];
        return $errors;
    }

    /**
     * Permet de gréffer une erreur dans la liste
     * @param array $error
     */
    public static function addError(array $error)
    {
        static::$errors[] = $error;
    }

    /**
     * Handle an error from PHP set_error_handler
     *
     * Will use the configured renderer to generate output
     * and output it.
     *
     * This method will dispatch the `Error.beforeRender` event which can be listened
     * to on the global event manager.
     *
     * @param int         $code        Code of error
     * @param string      $description Error description
     * @param string|null $file        File on which error occurred
     * @param int|null    $line        Line that triggered the error
     * @return bool True if error was handled
     */
    public function handleError(
        int $code,
        string $description,
        ?string $file = null,
        ?int $line = null
    ): bool {
        if (!(error_reporting() & $code)) {
            return false;
        }
        if ($code === E_USER_ERROR || $code === E_ERROR || $code === E_PARSE) {
            throw new FatalErrorException($description, $code, $file, $line);
        }

        $trace = (array)Debugger::trace(['start' => 1, 'format' => 'points']);

        $ignoredPaths = (array)Configure::read('Error.ignoredDeprecationPaths');
        if ($code === E_USER_DEPRECATED && $ignoredPaths) {
            $relativePath = str_replace(DIRECTORY_SEPARATOR, '/', substr((string)$file, strlen(ROOT) + 1));
            foreach ($ignoredPaths as $pattern) {
                $pattern = str_replace(DIRECTORY_SEPARATOR, '/', $pattern);
                if (fnmatch($pattern, $relativePath)) {
                    return true;
                }
            }
        }

        if (self::isErrorDisabled()) {
            self::addError(compact('code', 'description', 'file', 'line', 'trace'));
            return true;
        }

        return parent::handleError($code, $description, $file, $line);
    }
}
