<?php

/**
 * AsalaeCore\Error\ExceptionTrap
 */

namespace AsalaeCore\Error;

use AsalaeCore\Error\Renderer\ConsoleExceptionRenderer;
use AsalaeCore\Error\Renderer\JsonExceptionRenderer;
use Cake\Console\CommandInterface;
use Cake\Console\Exception\ConsoleException;
use Cake\Error\ExceptionRendererInterface;
use Cake\Error\ExceptionTrap as CakeExceptionTrap;
use Cake\Http\ServerRequest;
use Cake\Log\Log;
use Cake\Routing\Router;
use Exception;
use Psr\Http\Message\ServerRequestInterface;
use Throwable;

use function Cake\Core\env;

/**
 * Gestions d'exceptions de l'application
 *
 * @category    Error
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2022, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class ExceptionTrap extends CakeExceptionTrap
{
    /**
     * Configuration options. Generally these will be defined in your config/app.php
     *
     * - `exceptionRenderer` - string - The class responsible for rendering uncaught exceptions.
     *   The chosen class will be used for for both CLI and web environments. If  you want different
     *   classes used in CLI and web environments you'll need to write that conditional logic as well.
     *   The conventional location for custom renderers is in `src/Error`. Your exception renderer needs to
     *   implement the `render()` method and return either a string or Http\Response.
     * - `log` Set to false to disable logging.
     * - `logger` - string - The class name of the error logger to use.
     * - `trace` - boolean - Whether or not backtraces should be included in
     *   logged exceptions.
     * - `skipLog` - array - List of exceptions to skip for logging. Exceptions that
     *   extend one of the listed exceptions will also not be logged. E.g.:
     *   ```
     *   'skipLog' => ['Cake\Http\Exception\NotFoundException', 'Cake\Http\Exception\UnauthorizedException']
     *   ```
     *   This option is forwarded to the configured `logger`
     * - `extraFatalErrorMemory` - int - The number of megabytes to increase the memory limit by when a fatal error is
     *   encountered. This allows breathing room to complete logging or error handling.
     * - `stderr` Used in console environments so that renderers have access to the current console output stream.
     *
     * @var array<string, mixed>
     */
    protected array $_defaultConfig = [
        'exceptionRenderer' => null,
        'logger' => ErrorLogger::class,
        'stderr' => null,
        'log' => true,
        'skipLog' => [],
        'trace' => false,
        'extraFatalErrorMemory' => 4,
    ];

    /**
     * @var ServerRequest|null
     */
    private $request;

    /**
     * Get an instance of the renderer.
     *
     * @param Throwable                   $exception Exception to render
     * @param ServerRequestInterface|null $request   The request if possible.
     * @return ExceptionRendererInterface
     */
    public function renderer(Throwable $exception, $request = null): ExceptionRendererInterface
    {
        $this->request = $request ?? Router::getRequest();

        if (str_contains($this->request?->getHeaderLine('Accept') ?? '', 'application/json')) {
            $this->setConfig('exceptionRenderer', JsonExceptionRenderer::class);
        }

        return parent::renderer($exception, $request);
    }

    /**
     * Choose an exception renderer based on config or the SAPI
     *
     * @return class-string<ConsoleExceptionRenderer|AppExceptionRenderer>
     */
    protected function chooseRenderer(): string
    {
        $config = $this->getConfig('exceptionRenderer');
        if ($config !== null) {
            return $config;
        }

        /** @var class-string<ConsoleExceptionRenderer|AppExceptionRenderer> */
        return PHP_SAPI === 'cli' && empty($this->request)
            ? ConsoleExceptionRenderer::class
            : AppExceptionRenderer::class;
    }

    /**
     * On retire stdout et stderr des logs pour éviter d'avoir un doublon dans
     * les sorties, l'entrée est bien inscrite dans error.log mais l'affichage
     * est géré par ExceptionRenderer
     *
     * @param Throwable                   $exception The exception to log
     * @param ServerRequestInterface|null $request   The optional request
     * @return void
     */
    public function logException(Throwable $exception, ?ServerRequestInterface $request = null): void
    {
        $stdout = Log::getConfig('stdout');
        $stderr = Log::getConfig('stderr');
        Log::drop('stdout');
        Log::drop('stderr');
        parent::logException($exception, $request);
        if (PHP_SAPI === 'cli' && $stdout && $stderr) {
            Log::setConfig('stdout', $stdout);
            Log::setConfig('stderr', $stderr);
        }
    }

    /**
     * Handle uncaught exceptions.
     *
     * Uses a template method provided by subclasses to display errors in an
     * environment appropriate way.
     *
     * @param \Throwable $exception Exception instance.
     * @return void
     * @throws Exception When renderer class not found
     * @see https://secure.php.net/manual/en/function.set-exception-handler.php
     */
    public function handleException(Throwable $exception): void
    {
        parent::handleException($exception);
        if (PHP_SAPI === 'cli' && empty($this->request) && !env('FIXTURE_SCHEMA_METADATA')) {
            $exitCode = CommandInterface::CODE_ERROR;
            if ($exception instanceof ConsoleException) {
                $exitCode = $exception->getCode();
            }
            exit($exitCode);
        }
    }
}
