<?php

/**
 * AsalaeCore\Form\AbstractGenericMessageForm
 */

namespace AsalaeCore\Form;

use AsalaeCore\Form\MessageSchema\MessageSchema;
use Cake\Cache\Cache;
use Cake\Core\Configure;
use Cake\Database\Expression\IdentifierExpression;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use DOMDocument;
use DOMElement;
use DOMNode;
use DOMXPath;
use Exception;

/**
 * Génère un formulaire à partir d'un schéma de type MessageFormSchemaInterface
 *
 * @category Form
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2024, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 * @mixin MessageFormInterface
 */
abstract class AbstractGenericMessageForm
{
    /**
     * @var array liste les schemas (lorsque Cache::disabled)
     */
    public static $schemas;

    /**
     * Donne l'uri du dossier "data" d'un transfert
     * @param string|int $transfer_id
     * @return string
     */
    public static function getDataDirectory($transfer_id): string
    {
        return Configure::read('App.paths.data') . DS . 'transfers' . DS . $transfer_id;
    }

    /**
     * Permet d'appeler plus facilement appendNode() en calculant certains attributs
     * @param DOMElement    $node
     * @param string        $append
     * @param null|bool|int $orgEntityId si valeur à true, la valeur est définie automatiquement
     *                                   par rapport à l'archive identifier dans le dom
     * @return DOMElement
     * @throws Exception
     */
    public static function smartAppendNode(DOMElement $node, string $append, $orgEntityId = null): DOMElement
    {
        if (is_bool($orgEntityId) && $orgEntityId) {
            $orgEntityId = static::searchOrgEntityId($node);
        }
        $path = [$node->tagName];
        $parent = $node->parentNode;
        while ($parent instanceof DOMElement) {
            array_unshift($path, $parent->tagName);
            $parent = $parent->parentNode;
        }
        return static::appendNode(
            $orgEntityId,
            $append,
            implode('.', $path),
            $node->ownerDocument,
            $node
        );
    }

    /**
     * Donne le org_entity_id à partir d'un DOMElement
     * @param DOMElement $node
     * @return int|null
     */
    private static function searchOrgEntityId(DOMElement $node)
    {
        $dom = $node->ownerDocument;
        $xpath = new DOMXPath($dom);
        $namespace = $dom->documentElement->getAttributeNode('xmlns')->nodeValue;
        $xpath->registerNamespace('ns', $namespace);
        $query = $xpath->query('ns:ArchivalAgency/ns:Identification | ns:ArchivalAgency/ns:Identifier')
            ->item(0);
        $identifier = $query?->nodeValue;
        $OrgEntities = TableRegistry::getTableLocator()->get('OrgEntities');
        $orgEntity = $OrgEntities->find()
            ->select(['id'])
            ->innerJoinWith('ArchivalAgencies')
            ->where(
                [
                    'OrgEntities.identifier' => $identifier,
                    'ArchivalAgencies.id' => new IdentifierExpression('OrgEntities.id'),
                ]
            )
            ->first();
        return $orgEntity?->id;
    }

    /**
     * Ajoute un element en respectant l'ordre dans le schema
     * @param string|int  $orgEntityId
     * @param string      $append      tagName de l'element
     *                                 à ajouter
     * @param string      $path
     * @param DOMDocument $dom
     * @param DOMElement  $parentNode  element sur lequel ajouter le nouveau noeud
     * @return DOMElement
     * @throws Exception
     */
    public static function appendNode(
        $orgEntityId,
        string $append,
        string $path,
        DOMDocument $dom,
        DOMElement $parentNode
    ): DOMElement {
        // récupère "Archive" dans "Archive.ArchiveObject.Document"
        $base = ($pos = strpos($path, '.')) ? substr($path, 0, $pos) : $path;
        $schema = MessageSchema::getCachedSchema($base, static::getVersion($dom), $orgEntityId);
        $ref =& $schema;
        $paths = $path ? explode('.', str_replace('.', '.children.', $path)) : [];
        foreach ($paths as $k) {
            $ref =& $ref[$k];
        }
        if (empty($ref)) {
            throw new Exception(
                __("Schema non trouvé pour le chemin {0}", $path)
            );
        }
        $appendBefore = [];
        $found = false;
        foreach (array_keys((array)$ref['children']) as $child) {
            if ($child === $append) {
                $found = true;
            } elseif ($found) {
                $appendBefore[] = $child;
            }
        }
        $newNode = $dom->createElementNS($parentNode->namespaceURI, $append);
        $inserted = false;
        if (!empty($ref['complex'])) {
            return static::insertComplex($newNode, $parentNode, $ref['complex']);
        }
        foreach ($parentNode->childNodes as $node) {
            if ($node instanceof DOMElement && in_array($node->tagName, $appendBefore)) {
                $newNode = $parentNode->insertBefore($newNode, $node);
                $inserted = true;
                break;
            }
        }
        if (!$inserted) {
            $newNode = $parentNode->appendChild($newNode);
        }
        return $newNode;
    }

    /**
     * Insertion d'un element avec structure complexe
     * @param DOMElement $newNode
     * @param DOMElement $parentNode
     * @param array      $complex
     * @return DOMElement|DOMNode
     */
    public static function insertComplex(DOMElement $newNode, DOMElement $parentNode, array $complex)
    {
        $flatten = Hash::flatten($complex);

        $key = preg_replace('/\.\d+$/', '', array_search($newNode->tagName, $flatten));
        $block = Hash::extract($complex, $key);
        $type = substr($key, ($pos = strrpos($key, '.')) ? $pos + 1 : 0);
        if (is_numeric($type)) {
            $type = 'sequence';
        }

        $beginAfterNode = false;
        $beginAfterBlock = false;
        $afterNode = [];
        $afterBlock = [];
        $target = end($block);
        foreach ($flatten as $value) {
            if ($beginAfterBlock) {
                $afterBlock[] = $value;
            } elseif ($value === $target) {
                $beginAfterBlock = true;
            }
            if ($beginAfterNode) {
                $afterNode[] = $value;
            } elseif ($value === $newNode->tagName) {
                $beginAfterNode = true;
            }
        }

        if ($type === 'sequence') {
            $sequences = [[]];
            $index = 0;
            // on liste les répétitions de la séquences
            foreach ($parentNode->childNodes as $node) {
                if (!$node instanceof DOMElement || !in_array($node->tagName, $block)) {
                    continue;
                }
                if (isset($sequences[$index][$node->tagName])) {
                    $index++;
                }
                $sequences[$index][$node->tagName] = $node;
            }
            // on trouve dans quelle séquence il faut insérer $newNode
            $targetIndex = null;
            foreach ($sequences as $index => $nodes) {
                if (!isset($nodes[$newNode->tagName])) {
                    $targetIndex = $index;
                    break;
                }
            }
            // on ajoute l'élement dans la sequence choisie
            if ($targetIndex !== null && isset($sequences[$targetIndex])) {
                foreach ($sequences[$targetIndex] as $tagName => $node) {
                    if (in_array($tagName, $afterNode)) {
                        return $parentNode->insertBefore($newNode, $node);
                    }
                }
                // si on arrive là, c'est qu'il faut insérer à la fin de la séquence
                // donc avant le 1er element de la sequence suivante
                if (isset($sequences[$targetIndex + 1])) {
                    return $parentNode->insertBefore($newNode, current($sequences[$targetIndex + 1]));
                }
                // l'élement choisi est en derniere position de la derniere sequence
            }
            // sinon c'est comme pour un choice (insertBefore le prochain element après le bloc)
        }

        // Si on arrive ici, c'est qu'il faut insérer sans tenir compte de la séquence
        foreach ($parentNode->childNodes as $node) {
            if ($node instanceof DOMElement && in_array($node->tagName, $afterBlock)) {
                return $parentNode->insertBefore($newNode, $node);
            }
        }

        // Si on arrive ici c'est que le bloc à insérer est position après ceux présent
        return $parentNode->appendChild($newNode);
    }

    /**
     * Supprime le caché lié à un org_entity et à ses parents
     * @param string     $key
     * @param string|int $orgEntityId
     */
    public static function clearCache(string $key, $orgEntityId)
    {
        $OrgEntities = TableRegistry::getTableLocator()->get('OrgEntities');
        try {
            $entity = $OrgEntities->get($orgEntityId);
        } catch (RecordNotFoundException) {
            return;
        }
        $query = $OrgEntities->find()
            ->select(['id'])
            ->where(
                [
                    'lft <=' => $entity->get('lft'),
                    'rght >=' => $entity->get('rght'),
                ]
            );
        /** @var EntityInterface $entity */
        foreach ($query as $entity) {
            Cache::delete($key . '_' . $entity->get('id'));
        }
    }

    /**
     * Donne un chemin xpath à partir d'un numéro de ligne
     * @param int   $line
     * @param array $lines
     * @return string chemin/vers/element
     */
    public static function lineToPath($line, array $lines): string
    {
        $output = $lines[$line]['element'];
        $indent = $lines[$line]['indent'];
        for ($i = (int)$line; $i > 0; $i--) {
            if (isset($lines[$i]['indent']) && $lines[$i]['indent'] < $indent) {
                $indent = $lines[$i]['indent'];
                $output = $lines[$i]['element'] . ' / ' . $output;
            }
        }
        return '[ ' . $output . ' ]';
    }

    /**
     * Donne la version du SEDA au format numérique (ex: 1.0)
     * @param DOMDocument|null $dom
     * @return string
     */
    public static function getVersion(?DOMDocument $dom = null): string
    {
        if (!$dom) {
            return 'unknown';
        }
        $xmlns = $dom->documentElement->getAttributeNode('xmlns');
        $namespace = $xmlns?->nodeValue;
        if ($namespace === NAMESPACE_SEDA_02) {
            return '0.2';
        } elseif (preg_match('/fr:gouv:culture:archivesdefrance:seda:v(\d+\.\d+)/', $namespace, $m)) {
            return $m[1];
        } else {
            return 'unknown';
        }
    }

    /**
     * @param Validator $validator
     * @param string    $key
     * @param string    $field
     * @param string    $max
     * @return void
     */
    public static function addValidationTooMuch(Validator $validator, string $key, string $field, string $max): void
    {
        $validator->add(
            $key,
            'toomuch',
            [
                'rule' => function ($value, $context) use ($field, $max) {
                    $count = 0;
                    foreach (array_keys($context['data']) as $dataField) {
                        if (preg_match("/^{$field}_#\d+$/", $dataField)) {
                            $count++;
                        }
                    }
                    return $count <= $max;
                },
                'message' => __("Le nombre de {0} est limité à {1} dans ce contexte", $field, $max),
            ]
        );
    }
}
