<?php

/**
 * AsalaeCore\Form\MessageSchema\Seda23Schema
 */

namespace AsalaeCore\Form\MessageSchema;

/**
 * Schema d'un message seda v2.3
 *
 * @category Form\MessageSchema
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2024, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class Seda23Schema extends Seda22Schema implements MessageFormSchemaInterface
{
    use SchemaTrait;

    /**
     * ref MinimalDataObjectType
     * @return array
     */
    protected function extMinimalDataObjectType()
    {
        return $this->list(
            'DataObjectProfile',
            'DataObjectSystemId',
            'DataObjectGroupSystemId',
            'Relationship',
            'DataObjectGroupReferenceId',
            'DataObjectGroupId',
            'DataObjectVersion',
            'PersistentIdentifier',
            'DataObjectUse',
            'DataObjectNumber',
        );
    }

    /**
     * ArchiveTransfer.DataObjectPackage.BinaryDataObject.DataObjectUse
     * @return array
     */
    protected function getDataObjectUse(): array
    {
        return [
            'label' => "DataObjectUse",
            'info' => __(
                "Usage d’un objet-données (ex. original numérique, vignette, etc.)."
            ),
            'cardinality' => '0..1',
            'disabled' => false,
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.BinaryDataObject.DataObjectNumber
     * @return array
     */
    protected function getDataObjectNumber(): array
    {
        return [
            'label' => "DataObjectNumber",
            'info' => __(
                "Numéro de version d’un objet-données (ex. 1, 18, 300, etc.)."
            ),
            'cardinality' => '0..1',
            'data-type' => 'integer',
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.BinaryDataObject
     * @return array
     */
    protected function getBinaryDataObject(): array
    {
        $parentData = parent::getBinaryDataObject();
        $parentData['complex'] = [
            'DataObjectSystemId',
            'DataObjectGroupSystemId',
            'Relationship',
            'DataObjectGroupReferenceId',
            'DataObjectGroupId',
            'DataObjectVersion',
            'PersistentIdentifier',
            'DataObjectUse',
            'DataObjectNumber',
            [
                'choice' => [
                    'Attachment',
                    'Uri',
                ],
            ],
            'MessageDigest',
            'Size',
            'Compressed',
            'FormatIdentification',
            'FileInfo',
            'Metadata',
            'OtherMetadata',
        ];

        return $parentData;
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content
     * @return array
     */
    protected function getContent(): array
    {
        return [
            'text' => 'Content',
            'cardinality' => '1..1',
            'li_attr' => [
                'title' => __("Métadonnées de description associées à un ArchiveUnit."),
            ],
            'children' => $this->list(
                'DescriptionLevel',
                'Title',
                'FilePlanPosition',
                'SystemId',
                'OriginatingSystemId',
                'ArchivalAgencyArchiveUnitIdentifier',
                'OriginatingAgencyArchiveUnitIdentifier',
                'TransferringAgencyArchiveUnitIdentifier',
                'Description',
                'CustodialHistory',
                'Type',
                'DocumentType',
                'Language',
                'DescriptionLanguage',
                'Status',
                'Version',
                'Tag',
                'Keyword',
                'Coverage',
                'OriginatingAgency',
                'SubmissionAgency',
                'Agent',
                'AuthorizedAgent',
                'Writer',
                'Addressee',
                'Recipient',
                'Transmitter',
                'Sender',
                'Source',
                'RelatedObjectReference',
                'CreatedDate',
                'TransactedDate',
                'AcquiredDate',
                'SentDate',
                'ReceivedDate',
                'RegisteredDate',
                [
                    'StartDate' => [
                        'label' => "StartDate : " . __("Date d'ouverture / date de début."),
                        'cardinality' => '0..1',
                        'help' => __("Optionnel"),
                        'data-type' => 'datetime',
                        'validations' => $this->validations('DateTimeType'),
                        'class' => 'datepicker',
                        'placeholder' => $this->datetimePlaceholder(),
                    ],
                ],
                'EndDate',
                'DateLitteral',
                'Event',
                'SigningInformation',
                'Gps',
                'OriginatingSystemIdReplyTo',
                'TextContent',
                'PersistentIdentifier'
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation
     * @return array
     */
    protected function getSigningInformation(): array
    {
        return [
            'text' => 'SigningInformation',
            'cardinality' => '0..1',
            'li_attr' => [
                'title' => __("Contient toutes les informations relatives à la signature."),
            ],
            'children' => $this->list(
                'SigningRole',
                'DetachedSigningRole',
                'SignedDocumentReferenceId',
                'SignatureDescription',
                'TimestampingInformation',
                'AdditionalProof',
                // 'Extended', non géré
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.SigningRole
     * @return array
     */
    protected function getSigningRole(): array
    {
        return [
            'text' => 'SigningRole',
            'cardinality' => '1..n',
            'options' => [
                ['value' => 'SignedDocument', 'text' => 'SignedDocument'],
                ['value' => 'Timestamp', 'text' => 'Timestamp'],
                ['value' => 'Signature', 'text' => 'Signature'],
                ['value' => 'AdditionalProof', 'text' => 'AdditionalProof'],
            ],
            'data-placeholder' => __("-- Choisir un SigningRole --"),
            'empty' => false,
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.DetachedSigningRole
     * @return array
     */
    protected function getDetachedSigningRole(): array
    {
        return [
            'text' => 'SigningRole',
            'cardinality' => '0..n',
            'options' => [
                ['value' => 'Timestamp', 'text' => 'Timestamp'],
                ['value' => 'Signature', 'text' => 'Signature'],
                ['value' => 'AdditionalProof', 'text' => 'AdditionalProof'],
            ],
            'data-placeholder' => __("-- Choisir un SigningRole --"),
            'empty' => false,
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.SignedDocumentReferenceId
     * @return array
     */
    protected function getSignedDocumentReferenceId(): array
    {
        return [
            'text' => 'SignedDocumentReferenceId',
            'cardinality' => '0..n',
            'info' => __(
                "Relation technique à l'unité d'archives racine décrivant le document signé à partir des unités "
                    . "d'archives encapsulées contenant les autres rôles d'un contexte de signature."
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.SignatureDescription
     * @return array
     */
    protected function getSignatureDescription(): array
    {
        return [
            'text' => 'SignatureDescription',
            'cardinality' => '0..n',
            'li_attr' => [
                'title' => __("Bloc permettant de décrire la signature dans un contexte de signature."),
            ],
            'children' => $this->list(
                [
                    'Signer' => [
                        'text' => 'Signer',
                        'cardinality' => '0..1',
                        'li_attr' => [
                            'title' => __("Signataire(s) de la transaction ou de l'objet."),
                        ],
                        'children' => $this->list(
                            $this->refPersonOrEntityGroup(),
                            'SigningTime',
                            $this->refBusinessGroup()
                        ),
                    ]
                ],
                'Validator',
                'SigningType',
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.SignatureDescription.Validator
     * @return array
     */
    protected function getValidator(): array
    {
        return [
            'text' => 'Validator',
            'cardinality' => '0..1',
            'true_cardinality' => '1..1',
            'li_attr' => [
                'title' => __("Validateur de la signature."),
            ],
            'children' => $this->list(
                $this->refPersonOrEntityGroup(),
                'ValidationTime',
                $this->refBusinessGroup()
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.SignatureDescription.SigningType
     * @return array
     */
    protected function getSigningType(): array
    {
        return [
            'text' => 'SigningType',
            'cardinality' => '0..1',
            'info' => __("Type de signature, au sens juridique du terme. Par exemple, simple, avancée, qualifiée."),
            'empty' => false,
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.TimestampingInformation
     * @return array
     */
    protected function getTimestampingInformation(): array
    {
        return [
            'text' => 'TimestampingInformation',
            'cardinality' => '0..1',
            'li_attr' => [
                'title' => __(
                    "Bloc permettant de conserver les informations d'horodatage dans un contexte de signature."
                ),
            ],
            'children' => $this->list(
                'TimeStamp',
                'AdditionalTimestampingInformation',
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.TimestampingInformation.TimeStamp
     * @return array
     */
    protected function getTimestamp(): array
    {
        return [
            'label' => "TimeStamp : " . __("Horodatage de la signature."),
            'cardinality' => '0..1',
            'help' => __("Optionnel"),
            'data-type' => 'datetime',
            'validations' => $this->validations('DateTimeType'),
            'class' => 'datepicker',
            'placeholder' => $this->datetimePlaceholder(),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.TimestampingInformation.AdditionalTimestampingInformation
     * @return array
     */
    protected function getAdditionalTimestampingInformation(): array
    {
        return [
            'label' => "AdditionalTimestampingInformation",
            'cardinality' => '0..1',
            'info' => __("Informations complémentaires sur l'horodatage d'une signature."),
            'help' => __("Optionnel"),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.SigningInformation.AdditionalProof
     * @return array
     */
    protected function getAdditionalProof(): array
    {
        return [
            'text' => "AdditionalProof",
            'cardinality' => '0..n',
            'info' => __("Bloc permettant de conserver les preuves complémentaires dans un contexte de signature."),
            'children' => $this->list(
                [
                    'AdditionalProofInformation' => [
                        'label' => "AdditionalProofInformation",
                        'cardinality' => '0..n',
                        'li_attr' => [
                            'title' => __(
                                "Informations relatives aux preuves complémentaires archivées "
                                    . "dans un contexte de signature."
                            ),
                        ],
                        'help' => __("Optionnel"),
                    ]
                ]
            ),
        ];
    }

    /**
     * ArchiveTransfer.DataObjectPackage.DescriptiveMetadata.ArchiveUnit.Content.PersistentIdentifier
     * @return array
     */
    protected function getPersistentIdentifier(): array
    {
        return [
            'text' => 'PersistentIdentifier',
            'cardinality' => '0..n',
            'li_attr' => [
                'title' => __("Identifiants pérennes."),
            ],
            'children' => $this->list(
                [
                    'PersistentIdentifierType' => [
                        'label' => "PersistentIdentifierType : " . __("Type d'identifiant pérenne."),
                        'cardinality' => '1..1',
                    ],
                ],
                [
                    'PersistentIdentifierOrigin' => [
                        'label' => "PersistentIdentifierOrigin : " . __("Origine d'identifiant pérenne."),
                        'cardinality' => '0..1',
                        'help' => __("Optionnel"),
                    ],
                ],
                [
                    'PersistentIdentifierReference' => [
                        'label' => "PersistentIdentifierReference : " . __("Référence d'identifiant pérenne."),
                        'cardinality' => '0..1',
                        'help' => __("Optionnel"),
                    ],
                ],
                [
                    'PersistentIdentifierContent' => [
                        'label' => "PersistentIdentifierContent : " . __("Valeur de l'identifiant pérenne."),
                        'cardinality' => '1..1',
                    ],
                ],
            ),
        ];
    }
}
