<?php

namespace AsalaeCore\I18n;

use AsalaeCore\Exception\GenericException;
use BadMethodCallException;
use Cake\I18n\Date as CakeDate;
use Cake\I18n\DateTime as CakeDateTime;
use DateInterval;
use DateMalformedStringException;
use DateTimeImmutable;
use DateTimeInterface;
use DateTimeZone;
use JsonSerializable;
use Stringable;

/**
 * Surcharge du Cake\I18n\Date afin d'être un DateTimeInterface et de pouvoir
 * le comparer avec un autre DateTimeInterface
 *
 * @category I18n
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2024, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class Date extends DateTimeImmutable implements DateTimeInterface, JsonSerializable, Stringable
{
    /**
     * @var CakeDate Objet Date encapsulé
     */
    private CakeDate $cakeDate;

    /**
     * Surcharge de constructeur
     * @param string            $datetime
     * @param DateTimeZone|null $timezone
     * @throws DateMalformedStringException
     */
    public function __construct(string $datetime = 'now', ?DateTimeZone $timezone = null)
    {
        $tempDateTime = new DateTimeImmutable($datetime, $timezone);
        parent::__construct($tempDateTime->format('Y-m-d 00:00:00'), $timezone);
        $this->cakeDate = new CakeDate($this, $timezone);
    }

    /**
     * Méthode magique pour appeler des méthodes statiques de Cake\I18n\Date
     * @param string $name
     * @param array  $arguments
     * @return mixed
     */
    public static function __callStatic(string $name, array $arguments): mixed
    {
        if (method_exists(CakeDate::class, $name)) {
            return forward_static_call_array([CakeDate::class, $name], $arguments);
        }
        throw new BadMethodCallException("Méthode statique $name non trouvée dans Cake\I18n\Date.");
    }

    /**
     * string ou objet Date/DateTime, de cakephp ou pas vers un DateTimeImmutable
     * @param mixed $date
     * @return DateTimeImmutable|null
     * @throws DateMalformedStringException
     */
    public static function anyToDateTime(mixed $date): ?DateTimeImmutable
    {
        if (is_string($date)) {
            return self::stringToDateTime($date);
        } elseif ($date instanceof DateTimeInterface) {
            return new DateTimeImmutable($date->format(DateTimeInterface::ATOM));
        } elseif ($date instanceof CakeDate) {
            return new DateTimeImmutable($date->format('Y-m-d'));
        }
        return null;
    }

    /**
     * Détecte le format date ISO ou I18n et renvoi un objet DateTimeImmutable
     * @param string $string
     * @return DateTimeImmutable|null
     * @throws DateMalformedStringException
     */
    public static function stringToDateTime(string $string): ?DateTimeImmutable
    {
        // Format YYYY-MM-DD
        if (preg_match('/^\d{4}-\d{2}-\d{2}$/', $string)) {
            return new DateTimeImmutable($string);
        }
        // Format YYYY-MM-DD HH:MM:SS
        if (preg_match('/^\d{4}-\d{2}-\d{2}[T ]\d{2}:\d{2}(:\d{2})?/', $string)) {
            return new DateTimeImmutable($string);
        }
        // Format ~= DD/MM/YYYY HH:MM:SS
        if (preg_match('/\d{2}:\d{2}(:\d{2})?/', $string)) {
            $datetime = CakeDateTime::parseDateTime($string);
            if (!$datetime) {
                throw new GenericException('Unable to convert string to DateTime.');
            }
            return new DateTimeImmutable($datetime->format(DateTimeInterface::ATOM));
        }
        // Format ~= DD/MM/YYYY
        $datetime = CakeDateTime::parseDate($string);
        if (!$datetime) {
            throw new GenericException('Unable to convert string to DateTime.');
        }
        return new DateTimeImmutable($datetime->format('Y-m-d'));
    }

    /**
     * Surcharge du diff
     * @param DateTimeInterface $targetObject
     * @param bool              $absolute
     * @return DateInterval
     */
    public function diff(
        DateTimeInterface $targetObject,
        bool $absolute = false
    ): DateInterval {
        if ($targetObject instanceof CakeDate) {
            return $this->cakeDate->diff($targetObject, $absolute);
        }
        return parent::diff($targetObject, $absolute);
    }

    /**
     * Méthode magique pour accéder aux méthodes dynamiques de Cake\I18n\Date
     * @param string $name
     * @param array  $arguments
     * @return mixed
     */
    public function __call(string $name, array $arguments): mixed
    {
        if (method_exists($this->cakeDate, $name)) {
            return call_user_func_array([$this->cakeDate, $name], $arguments);
        }
        throw new BadMethodCallException("Méthode $name non trouvée dans Cake\I18n\Date.");
    }

    /**
     * Méthode magique pour accéder aux propriétés dynamiques (lecture)
     * @param string $name
     * @return string|int|bool|float
     */
    public function __get(string $name): string|int|bool|float
    {
        if (property_exists($this->cakeDate, $name)) {
            return $this->cakeDate->$name;
        }
        throw new GenericException("Propriété $name non trouvée dans Cake\I18n\Date.");
    }

    /**
     * Méthode magique pour écrire dans des propriétés dynamiques
     * @param string $name
     * @param mixed  $value
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        if (property_exists($this->cakeDate, $name)) {
            $this->cakeDate->$name = $value;
        } else {
            throw new GenericException("Propriété $name non trouvée dans Cake\I18n\Date.");
        }
    }

    /**
     * Méthode pour récupérer l'objet CakeDate encapsulé si nécessaire
     * @return CakeDate
     */
    public function getCakeDate(): CakeDate
    {
        if (!isset($this->cakeDate)) {
            $this->cakeDate = new CakeDate($this, $this->getTimezone());
        }
        return $this->cakeDate;
    }

    /**
     * Renvoi le cakeDate en string
     * @return string
     */
    public function __toString(): string
    {
        return (string)$this->cakeDate;
    }

    /**
     * Returns a string that should be serialized when converting this object to JSON
     *
     * @return string|int
     */
    public function jsonSerialize(): mixed
    {
        return $this->cakeDate->jsonSerialize();
    }

    /**
     * @param string $date
     * @return CakeDate
     * @throws DateMalformedStringException
     */
    public static function parseDate(string $date): CakeDate
    {
        $instance = new static(static::stringToDateTime($date)->format('Y-m-d'));
        return $instance->cakeDate;
    }
}
