<?php

/**
 * AsalaeCore\ORM\Marshaller
 */

namespace AsalaeCore\ORM;

use AsalaeCore\DataType\CommaArrayString;
use AsalaeCore\DataType\EncryptedString;
use AsalaeCore\DataType\JsonString;
use AsalaeCore\DataType\SerializedObject;
use AsalaeCore\I18n\Date as CoreDate;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\InvalidPropertyInterface;
use Cake\I18n\Date as CakeDate;
use Cake\ORM\Marshaller as CakeMarshaller;
use ReflectionClass;
use ReflectionException;

/**
 * Surcharge de Marshaller
 *
 * @category Entity
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2023, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class Marshaller extends CakeMarshaller
{
    /**
     *
     * @param array $data
     * @param array $options
     * @return EntityInterface
     */
    public function one(array $data, array $options = []): EntityInterface
    {
        $entity = parent::one($data, $options);
        if ($entity->isNew()) {
            return $entity;
        }
        foreach ($entity->toArray() as $field => $value) {
            if (!isset($data[$field])) {
                continue;
            }
            if ($value instanceof CommaArrayString) {
                $class = CommaArrayString::class;
            } elseif ($value instanceof EncryptedString) {
                $class = EncryptedString::class;
            } elseif ($value instanceof JsonString) {
                $class = JsonString::class;
            } elseif ($value instanceof SerializedObject) {
                $class = SerializedObject::class;
            } else {
                continue;
            }
            $original = (string)$entity->getOriginal($field);
            if (is_string($data[$field])) {
                $dataValue = $class::createFromString($data[$field]);
            } else {
                $dataValue = new $class($data[$field]);
            }
            $entity->setDirty($field, $original !== (string)$dataValue);
        }

        return $entity;
    }

    /**
     * Returns data and options prepared to validate and marshall.
     *
     * @param array                $data    The data to prepare.
     * @param array<string, mixed> $options The options passed to this marshaller.
     * @return array An array containing prepared data and options.
     * @throws ReflectionException
     */
    protected function _prepareDataAndOptions(array $data, array $options): array
    {
        $preparedData = parent::_prepareDataAndOptions($data, $options);

        $relf = new ReflectionClass($this->_table->getEntityClass());
        $entity = $relf->newInstanceWithoutConstructor();
        foreach ($entity->jsonFields ?? ['app_meta'] as $jsonField) {
            if (isset($data[$jsonField])) {
                $preparedData[0][$jsonField] = (string)JsonString::create($data[$jsonField]);
            }
        }
        return $preparedData;
    }

    /**
     * Surcharge de merge pour comparer les dates CakeDate avec CoreDate
     * @param EntityInterface $entity
     * @param array           $data
     * @param array           $options
     * @return EntityInterface
     * @throws ReflectionException
     */
    public function merge(EntityInterface $entity, array $data, array $options = []): EntityInterface
    {
        [$data, $options] = $this->_prepareDataAndOptions($data, $options);

        $isNew = $entity->isNew();
        $keys = [];

        if (!$isNew) {
            $keys = $entity->extract((array)$this->_table->getPrimaryKey());
        }

        if (isset($options['accessibleFields'])) {
            foreach ((array)$options['accessibleFields'] as $key => $value) {
                $entity->setAccess($key, $value);
            }
        }

        $errors = $this->_validate($data + $keys, $options['validate'], $isNew);
        $options['isMerge'] = true;
        $propertyMap = $this->_buildPropertyMap($data, $options);
        $properties = [];
        /**
         * @var string $key
         */
        foreach ($data as $key => $value) {
            if (!empty($errors[$key])) {
                if ($entity instanceof InvalidPropertyInterface) {
                    $entity->setInvalidField($key, $value);
                }
                continue;
            }
            $original = $entity->get($key);

            if (isset($propertyMap[$key])) {
                $value = $propertyMap[$key]($value, $entity);

                // Don't dirty scalar values and objects that didn't
                // change. Arrays will always be marked as dirty because
                // the original/updated list could contain references to the
                // same objects, even though those objects may have changed internally.
                if (
                    (
                    is_scalar($value)
                    && $original === $value
                    )
                    || (
                    $value === null
                    && $original === $value
                    )
                    || (
                    is_object($value)
                    && !($value instanceof EntityInterface)
                    && $original == $value
                    )
                    // --- début de la surcharge: ---
                    || (
                    $value instanceof CakeDate
                    && $original instanceof CoreDate
                    && (string)$value === (string)$original->getCakeDate()
                    )
                    // --- fin de la surcharge ---
                ) {
                    continue;
                }
            }
            $properties[$key] = $value;
        }

        $entity->setErrors($errors);
        if (!isset($options['fields'])) {
            $entity->set($properties);

            foreach ($properties as $field => $value) {
                if ($value instanceof EntityInterface) {
                    $entity->setDirty($field, $value->isDirty());
                }
            }
            $this->dispatchAfterMarshal($entity, $data, $options);

            return $entity;
        }

        foreach ((array)$options['fields'] as $field) {
            assert(is_string($field));
            if (!array_key_exists($field, $properties)) {
                continue;
            }
            $entity->set($field, $properties[$field]);
            if ($properties[$field] instanceof EntityInterface) {
                $entity->setDirty($field, $properties[$field]->isDirty());
            }
        }
        $this->dispatchAfterMarshal($entity, $data, $options);

        return $entity;
    }
}
