<?php

/**
 * AsalaeCore\TestSuite\TestCase
 */

namespace AsalaeCore\TestSuite;

use AsalaeCore\Exception\GenericException;
use AsalaeCore\TestSuite\Fixture\TruncateStrategy;
use Cake\Core\App;
use Cake\ORM\TableRegistry;
use Cake\TestSuite\Fixture\FixtureStrategyInterface;
use Cake\TestSuite\TestCase as CakeTestSuite;

/**
 * TestCase avec fixtures settable
 *
 * @category TestSuite
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2020, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class TestCase extends CakeTestSuite
{
    /**
     * @var array
     */
    public $genericSessionData = [];

    /**
     * @var array Fixtures déjà chargée lors du setUp
     */
    public array $setUpLoadedFixtures = [];

    /**
     * Constructor
     * @param string|null $name
     * @noinspection RedundantSuppression
     */
    public function __construct(?string $name = null)
    {
        parent::__construct($name);
        if (defined('GENERIC_SESSION_DATA')) {
            /** @noinspection PhpUndefinedConstantInspection */
            $this->genericSessionData = unserialize(GENERIC_SESSION_DATA);
        }
    }

    /**
     * Ajout des Fixtures manquantes
     * @return void
     */
    public function setUp(): void
    {
        // Si une table existe encore suite à un autre test (ou async),
        // on ajoute la fixture au test
        // (évite de planter le tearDown a cause d'un autre test)
        $this->setUpLoadedFixtures = [];
        $loc = TableRegistry::getTableLocator();
        $paths = App::classPath('Model/Table');
        $models = glob($paths[0] . '*Table.php');
        foreach ($models as $file) {
            $name = basename($file, 'Table.php');
            $model = $loc->get($name);
            if ($model->find()->count()) {
                $this->setUpLoadedFixtures[] = $model->getAlias();
                $fixtureName = 'app.' . $model->getAlias();
                if (!in_array($fixtureName, $this->fixtures)) {
                    $this->fixtures[] = $fixtureName;
                }
            }
        }
        $loc->clear();
        parent::setUp();
    }

    /**
     * Destruction du test
     * @return void
     */
    public function tearDown(): void
    {
        parent::tearDown();
        $loc = TableRegistry::getTableLocator();
        $paths = App::classPath('Model/Table');
        $models = glob($paths[0] . '*Table.php');
        $missingFixtures = [];
        foreach ($models as $file) {
            $name = basename($file, 'Table.php');
            $model = $loc->get($name);
            if ($model->find()->count()) {
                $missingFixtures[] = $model->getAlias();
            }
        }
        if ($missingFixtures) {
            sort($missingFixtures);
            throw new GenericException(
                __("Il manque {0} dans les fixtures", implode(', ', $missingFixtures))
            );
        }
        $loc->clear();
    }

    /**
     * Setter de fixtures
     * @param array $value
     */
    public function setFixture(array $value): void
    {
        $this->fixtures = $value;
    }

    /**
     * Returns fixture strategy used by these tests.
     *
     * @return \Cake\TestSuite\Fixture\FixtureStrategyInterface
     */
    protected function getFixtureStrategy(): FixtureStrategyInterface
    {
        $instance = new TruncateStrategy();
        $instance->emptyFixtures = $this->emptyFixtures ?? false;

        return $instance;
    }

    /**
     * Récupère la sortie standard et la sorti d'erreur d'une commande pour debug
     * @return void
     */
    protected function debugConsole(): void
    {
        debuglPrev(
            [
                'Console out' => implode("\n", $this->_out->messages()),
                'Console err' => implode("\n", $this->_err->messages()),
            ],
            2
        );
    }
}
