<?php

/**
 * AsalaeCore\Utility\SessionFiles
 */

namespace AsalaeCore\Utility;

use Cake\Core\Configure;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Exception\NotFoundException;
use Cake\Http\ServerRequest;
use Cake\Http\Session;
use Cake\ORM\TableRegistry;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Importé de Refae
 *
 * @category Utility
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2019, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 */
class SessionFiles
{
    public const string SESSION_KEY = 'tmpfiles';
    public const string SESSION_FILENAME = 'session.ser';

    private array $tmpFiles = [];
    private ?Session $session = null;
    private bool $lastChunk = false;
    private ?int $completedChunksCount = 0;
    private ?string $uri = null;
    private ?string $chunkDirectory = null;

    /**
     * @param ServerRequest $request
     */
    public function __construct(
        private readonly ServerRequest $request,
    ) {
    }

    /**
     * @param string $uuid
     * @return string|null
     */
    public function getUri(string $uuid): ?string
    {
        $session = $this->getSession();
        if (!$session) {
            return null;
        }
        if (isset($this->tmpFiles[$uuid])) {
            return $this->tmpFiles[$uuid];
        }
        throw new NotFoundException("Not Found");
    }

    /**
     * @param string|null $uuid
     * @return bool
     */
    public function exists(?string $uuid): bool
    {
        $session = $this->getSession();
        if (!$session || !$uuid) {
            return false;
        }
        return isset($this->tmpFiles[$uuid]);
    }

    /**
     * @param string $uuid
     * @return string|null
     */
    public function delete(string $uuid): ?string
    {
        $this->getSession();
        if (isset($this->tmpFiles[$uuid])) {
            $targetUri = $this->getUri($uuid);
            $this->sessionRemoveFile($uuid);
        } else {
            $targetUri = sprintf(
                "%s/%s/%s/chunks/",
                sys_get_temp_dir(),
                session_id(),
                $uuid,
            );
        }
        $filesystem = new Filesystem();
        $filesystem->remove($targetUri);

        // supprime le dossier vide
        self::removeEmptyParentDirs($targetUri);

        return session_id() . '/' . $uuid;
    }

    /**
     * @return void
     */
    public function checkValidChunk(): void
    {
        $request = $this->getRequest();
        $uploadedFile = $request->getUploadedFile('file');
        $chunk = $request->getData('chunk');
        $chunkSize = (int)$chunk['size'];
        $stream = $uploadedFile->getStream();
        $chunkContent = $stream->getContents();
        $stream->rewind();
        if (strlen($chunkContent) !== $chunkSize) {
            throw new BadRequestException("chunk size does not match chunk content size");
        }
        $chunkHash = hexdec((string) $chunk['hash']);
        if ($chunkHash !== hexdec(hash('crc32b', $chunkContent))) {
            throw new BadRequestException("chunk hash does not match chunk content hash");
        }
        if (preg_match('/[^\w-]/', (string) $request->getData('uuid'))) {
            throw new BadRequestException("invalid uuid");
        }
    }

    /**
     * @param string|null $tmpDir
     * @return void
     */
    public function writeChunk(?string $tmpDir = null): void
    {
        $request = $this->getRequest();
        $uploadedFile = $request->getUploadedFile('file');
        $chunk = $request->getData('chunk');
        $stream = $uploadedFile->getStream();
        $chunkContent = $stream->getContents();
        $chunkSize = (int)$chunk['size'];
        $currentChunk = (int)$chunk['i'];
        $totalChunk = (int)$chunk['total'];
        $chunkHash = hexdec((string) $chunk['hash']);
        $len = strlen((string)$totalChunk);
        $targetUri = sprintf(
            "%s/%s/%s/chunks/%0{$len}d",
            $tmpDir ?: sys_get_temp_dir(),
            session_id(),
            $request->getData('uuid'),
            $currentChunk,
        );
        $targetBaseDirectory = dirname($targetUri);
        if (!is_dir($targetBaseDirectory)) {
            @mkdir($targetBaseDirectory, 0777, true);
        }
        $written = file_put_contents($targetUri, $chunkContent);
        if ($written !== $chunkSize || hexdec(hash_file('crc32b', $targetUri)) !== $chunkHash) {
            if (is_file($targetUri)) {
                @unlink($targetUri);
            }
            self::removeEmptyParentDirs($targetUri);
            throw new BadRequestException("file write error");
        }

        $this->chunkDirectory = dirname($targetUri);
        $chunk = $request->getData('chunk');
        $totalChunk = (int)$chunk['total'];

        $existingChunks = $this->completedChunksCount();
        $this->lastChunk = $existingChunks === $totalChunk;
        if ($this->lastChunk) {
            $lockFile = $targetBaseDirectory . '/.lock';
            if (is_file($lockFile)) {
                $this->lastChunk = false;
            } else {
                touch($lockFile);
            }
        }
    }

    /**
     * @return void
     */
    public function mergeChunks()
    {
        $request = $this->getRequest();
        $chunks = glob($this->chunkDirectory . '/*');
        $neededTime = (int)(300 / 1000000000 * (int)$request->getData('size'));// 1Go ~= 300s
        if (PHP_SAPI !== 'cli' && $neededTime > 30) {
            set_time_limit($neededTime);
        }
        $safeFilename = $this->filterFilename($request->getData('filename'));
        $uri = dirname((string) $this->chunkDirectory) . '/' . $safeFilename;
        $file = fopen($uri, 'w');
        foreach ($chunks as $chunkFile) {
            fwrite($file, file_get_contents($chunkFile));
        }
        fclose($file);

        if (is_file($this->chunkDirectory . '/.lock')) {
            @unlink($this->chunkDirectory . '/.lock');
        }

        if (is_dir($this->chunkDirectory)) {
            $filesystem = new Filesystem();
            $filesystem->remove($this->chunkDirectory);
        }
        self::removeEmptyParentDirs($this->chunkDirectory);

        $this->sessionAddFile($request->getData('uuid'), $uri);
        $this->uri = $uri;
    }

    /**
     * @param string|null $tmpDir
     * @return string
     */
    public function getRelativeUri(?string $tmpDir = null): string
    {
        return substr((string) $this->uri, strlen($tmpDir ?: sys_get_temp_dir()) + 1);
    }

    /**
     * @return bool
     */
    public function isLastChunk(): bool
    {
        return $this->lastChunk;
    }

    /**
     * @return ServerRequest|null
     */
    private function getRequest(): ?ServerRequest
    {
        return $this->request;
    }

    /**
     * @return Session|null
     */
    private function getSession(): ?Session
    {
        if (!$this->session) {
            $this->session = $this->getRequest()?->getSession();
            $this->tmpFiles = $this->session->read(self::SESSION_KEY, []);
        }
        return $this->session;
    }

    /**
     * @return int
     */
    public function getCompletedChunksCount(): int
    {
        return $this->completedChunksCount;
    }

    /**
     * @return int
     */
    public function getTotalChunkCount(): int
    {
        $chunk = $this->getRequest()?->getData('chunk');
        return (int)$chunk['total'] ?? 0;
    }

    /**
     * Déplace le fichier dans son emplacement définitif
     * @param string $uuid
     * @param string $path
     * @return bool
     */
    public function moveUploadedFile(string $uuid, string $path): bool
    {
        $uri = $this->getUri($uuid);
        if ($uri === $path) {
            $this->sessionRemoveFile($uuid);
            return true;
        }
        if (!is_dir(dirname($path))) {
            mkdir(dirname($path), 0777, true);
        }
        $success = rename($uri, $path);
        if ($success) {
            $this->sessionRemoveFile($uuid);
        }
        return $success;
    }

    /**
     * Remet les fichiers dans la session suite à une reconnexion
     * @return void
     */
    public function reload(): void
    {
        $file = $this->getPersistantSessionPath();
        if (file_exists($file)) {
            $this->setTmpFiles(unserialize(file_get_contents($file)));
        }
    }

    /**
     * Fichier de sauvegarde de la session (permet la reconnexion)
     * @return string
     */
    private function getPersistantSessionPath(): string
    {
        $user_id = $this->getSession()?->read('Auth.id') ?: 'unknown';
        $tmpDir = rtrim(Configure::read('Chunky.defaultTmpDir', TMP), '/');
        return sprintf(
            "%s/upload/user_%d/%s/%s",
            $tmpDir,
            $user_id,
            session_id(),
            self::SESSION_FILENAME,
        );
    }

    /**
     * @param string $uuid
     * @param string $uri
     * @return void
     */
    private function sessionAddFile(string $uuid, string $uri): void
    {
        $this->getSession();
        $this->tmpFiles[$uuid] = $uri;
        $this->setTmpFiles($this->tmpFiles);
    }

    /**
     * @param string $uuid
     * @return void
     */
    private function sessionRemoveFile(string $uuid): void
    {
        $this->getSession();
        unset($this->tmpFiles[$uuid]);
        $this->setTmpFiles($this->tmpFiles);
    }

    /**
     * @return int
     */
    private function completedChunksCount(): int
    {
        if (!$this->completedChunksCount) {
            $request = $this->getRequest();
            $chunk = $request->getData('chunk');
            $totalChunk = (int)$chunk['total'];
            $chunks = glob($this->chunkDirectory . '/*');
            $count = count($chunks);
            if (count($chunks) === $totalChunk) {
                $size = 0;
                foreach ($chunks as $chunkFile) {
                    $size += filesize($chunkFile);
                }
                if ($size !== (int)$request->getData('size')) {
                    $count--; // évite de compter un chunk incomplet
                }
            }
            $this->completedChunksCount = $count;
        }
        return $this->completedChunksCount;
    }

    /**
     * @param $name
     * @return string
     */
    private function filterFilename($name)
    {
        $name = str_replace(
            array_merge(
                array_map('chr', range(0, 31)),
                ['<', '>', ':', '"', '/', '\\', '|', '?', '*']
            ),
            '',
            (string) $name
        );
        $ext = pathinfo($name, PATHINFO_EXTENSION);
        return mb_strcut(
            pathinfo($name, PATHINFO_FILENAME),
            0,
            255 - ($ext ? strlen($ext) + 1 : 0),
            mb_detect_encoding($name)
        ) . ($ext ? '.' . $ext : '');
    }

    /**
     * Saves the persistent session data to a file.
     * @return void
     */
    private function savePersistantSession()
    {
        $path = $this->getPersistantSessionPath();
        if (empty($this->tmpFiles) && is_file($path)) {
            unlink($path);
            self::removeEmptyParentDirs($path);
            return;
        }
        if (!is_dir(dirname($path))) {
            mkdir(dirname($path), 0777, true);
        }
        file_put_contents($path, serialize($this->tmpFiles));
        chmod($path, 0600);
    }

    /**
     * Recursively removes empty parent directories for a given path.
     *
     * @param string $path The file path for which empty parent directories will be removed.
     * @return void
     */
    private static function removeEmptyParentDirs(string $path): void
    {
        $parentDir = dirname($path);
        if (is_dir($parentDir) && count(scandir($parentDir)) === 2) {
            if (!rmdir($parentDir)) {
                return;
            }
            self::removeEmptyParentDirs($parentDir);
        }
    }

    /**
     * Met à jour $this->tmpFiles et sauvegarde le tout en session et dans un
     * fichier serialized (pour renouveller la session)
     * @param array $tmpFiles
     * @return void
     */
    private function setTmpFiles(array $tmpFiles): void
    {
        $session = $this->getSession();
        $this->tmpFiles = [];
        foreach ($tmpFiles as $uuid => $uri) {
            if (is_file($uri)) {
                $this->tmpFiles[$uuid] = $uri;
            }
        }
        $session->write(self::SESSION_KEY, $this->tmpFiles);
        $this->savePersistantSession();
    }

    /**
     * Performs garbage collection by removing temporary upload directories
     * associated with sessions that no longer exist in the database.
     *
     * @return void
     */
    public static function gc(): void
    {
        $Sessions = TableRegistry::getTableLocator()->get('Sessions');
        $tmpDir = rtrim(Configure::read('Chunky.defaultTmpDir', TMP), '/');
        $tmpSessions = glob(
            sprintf(
                '%s/upload/user_*/*/%s',
                $tmpDir,
                self::SESSION_FILENAME
            )
        );
        foreach ($tmpSessions as $tmpSession) {
            $session_id = basename(dirname($tmpSession));
            if ($Sessions->exists(['id' => $session_id])) {
                continue;
            }
            $tmpFiles = unserialize(file_get_contents($tmpSession));
            foreach ($tmpFiles as $uri) {
                unlink($uri);
                self::removeEmptyParentDirs($uri);
            }
            unlink($tmpSession);
            self::removeEmptyParentDirs($tmpSession);
        }
    }

    /**
     * Donne le data-file pour l'input
     * @param string $tmpDir
     * @return array
     */
    public function allUploadDataByTmpdir(string $tmpDir): array
    {
        $session = $this->getSession();
        $tmpFiles = array_filter(
            $session->read(self::SESSION_KEY, []),
            fn ($v) => str_starts_with($v, $tmpDir)
        );
        $uploadData = [];
        foreach ($tmpFiles as $uuid => $uri) {
            if (!file_exists($uri)) {
                $this->delete($uuid);
                continue;
            }
            $uploadData[] = [
                'id' => $uuid,
                'name' => basename($uri),
                'size' => filesize($uri),
            ];
        }
        return $uploadData;
    }
}
