<?php

/**
 * AsalaeCore\Utility\Translit
 */

namespace AsalaeCore\Utility;

use Transliterator;

/**
 * Permet de manipuler des chaines de caractères
 *
 * @category    Utility
 *
 * @author      Libriciel SCOP <contact@libriciel.coop>
 * @copyright   (c) 2020, Libriciel
 * @license     https://www.gnu.org/licenses/agpl-3.0.txt
 */
class Translit
{
    /**
     * @var Transliterator
     */
    private static $transliterator;
    /**
     * @var Transliterator
     */
    private static $softTransliterator;

    /**
     * @var string[]
     */
    private static $chars = [
        "\xe2\x80\x9a" => ',', // ‚
        "\xc6\x92" => 'f', // ƒ
        "\xe2\x80\x9e" => ',,', // „
        "\xe2\x80\xa6" => '...', // …
        "\xc5\xa0" => 'S', // Š
        "\xe2\x80\xb9" => '<', // ‹
        "\xc5\x92" => 'OE', // Œ
        "\xc5\xbd" => 'Z', // Ž
        "\xe2\x80\x98" => "'", // ‘
        "\xe2\x80\x99" => "'", // ’
        "\xe2\x80\x9c" => '"', // “
        "\xe2\x80\x9d" => '"', // ”
        "\xe2\x80\x93" => '-', // –
        "\xe2\x80\x94" => '-', // —
        "\xc5\xa1" => 's', // š
        "\xe2\x80\xba" => '>', // ›
        "\xc5\x93" => 'oe', // œ
        "\xc5\xbe" => 'z', // ž
        "\xc5\xb8" => 'Y', // Ÿ
        "\xc2\xa9" => '(C)', // ©
        "\xc2\xab" => '<<', // «
        "\xc2\xae" => '(R)', // ®
        "\xc2\xbb" => '>>', // »
        "\xc2\xbc" => ' 1/4', // ¼
        "\xc2\xbd" => ' 1/2', // ½
        "\xc2\xbe" => ' 3/4', // ¾
        "\xc3\x80" => 'A', // À
        "\xc3\x81" => 'A', // Á
        "\xc3\x82" => 'A', // Â
        "\xc3\x83" => 'A', // Ã
        "\xc3\x84" => 'A', // Ä
        "\xc3\x85" => 'A', // Å
        "\xc3\x86" => 'AE', // Æ
        "\xc3\x87" => 'C', // Ç
        "\xc3\x88" => 'E', // È
        "\xc3\x89" => 'E', // É
        "\xc3\x8a" => 'E', // Ê
        "\xc3\x8b" => 'E', // Ë
        "\xc3\x8c" => 'I', // Ì
        "\xc3\x8d" => 'I', // Í
        "\xc3\x8e" => 'I', // Î
        "\xc3\x8f" => 'I', // Ï
        "\xc3\x90" => 'ETH', // Ð
        "\xc3\x91" => 'N', // Ñ
        "\xc3\x92" => 'O', // Ò
        "\xc3\x93" => 'O', // Ó
        "\xc3\x94" => 'O', // Ô
        "\xc3\x95" => 'O', // Õ
        "\xc3\x96" => 'O', // Ö
        "\xc3\x97" => '*', // ×
        "\xc3\x98" => 'O', // Ø
        "\xc3\x99" => 'U', // Ù
        "\xc3\x9a" => 'U', // Ú
        "\xc3\x9b" => 'U', // Û
        "\xc3\x9c" => 'U', // Ü
        "\xc3\x9d" => 'Y', // Ý
        "\xc3\x9e" => 'TH', // Þ
        "\xc3\x9f" => 'ss', // ß
        "\xc3\xa0" => 'a', // à
        "\xc3\xa1" => 'a', // á
        "\xc3\xa2" => 'a', // â
        "\xc3\xa3" => 'a', // ã
        "\xc3\xa4" => 'a', // ä
        "\xc3\xa5" => 'a', // å
        "\xc3\xa6" => 'ae', // æ
        "\xc3\xa7" => 'c', // ç
        "\xc3\xa8" => 'e', // è
        "\xc3\xa9" => 'e', // é
        "\xc3\xaa" => 'e', // ê
        "\xc3\xab" => 'e', // ë
        "\xc3\xac" => 'i', // ì
        "\xc3\xad" => 'i', // í
        "\xc3\xae" => 'i', // î
        "\xc3\xaf" => 'i', // ï
        "\xc3\xb0" => 'eth', // ð
        "\xc3\xb1" => 'n', // ñ
        "\xc3\xb2" => 'o', // ò
        "\xc3\xb3" => 'o', // ó
        "\xc3\xb4" => 'o', // ô
        "\xc3\xb5" => 'o', // õ
        "\xc3\xb6" => 'o', // ö
        "\xc3\xb7" => '/', // ÷
        "\xc3\xb8" => 'o', // ø
        "\xc3\xb9" => 'u', // ù
        "\xc3\xba" => 'u', // ú
        "\xc3\xbb" => 'u', // û
        "\xc3\xbc" => 'u', // ü
        "\xc3\xbd" => 'y', // ý
        "\xc3\xbe" => 'th', // þ
        "\xc3\xbf" => 'y', // ÿ
        "\xc4\x80" => 'A', // Ā
        "\xc4\x81" => 'a', // ā
        "\xc4\x82" => 'A', // Ă
        "\xc4\x83" => 'a', // ă
        "\xc4\x84" => 'A', // Ą
        "\xc4\x85" => 'a', // ą
        "\xc4\x86" => 'C', // Ć
        "\xc4\x87" => 'c', // ć
        "\xc4\x88" => 'C', // Ĉ
        "\xc4\x89" => 'c', // ĉ
        "\xc4\x8a" => 'C', // Ċ
        "\xc4\x8b" => 'c', // ċ
        "\xc4\x8c" => 'C', // Č
        "\xc4\x8d" => 'c', // č
        "\xc4\x8e" => 'D', // Ď
        "\xc4\x8f" => 'd', // ď
        "\xc4\x90" => 'D', // Đ
        "\xc4\x91" => 'd', // đ
        "\xc4\x92" => 'E', // Ē
        "\xc4\x93" => 'e', // ē
        "\xc4\x94" => 'E', // Ĕ
        "\xc4\x95" => 'e', // ĕ
        "\xc4\x96" => 'E', // Ė
        "\xc4\x97" => 'e', // ė
        "\xc4\x98" => 'E', // Ę
        "\xc4\x99" => 'e', // ę
        "\xc4\x9a" => 'E', // Ě
        "\xc4\x9b" => 'e', // ě
        "\xc4\x9c" => 'G', // Ĝ
        "\xc4\x9d" => 'g', // ĝ
        "\xc4\x9e" => 'G', // Ğ
        "\xc4\x9f" => 'g', // ğ
        "\xc4\xa0" => 'G', // Ġ
        "\xc4\xa1" => 'g', // ġ
        "\xc4\xa2" => 'G', // Ģ
        "\xc4\xa3" => 'g', // ģ
        "\xc4\xa4" => 'H', // Ĥ
        "\xc4\xa5" => 'h', // ĥ
        "\xc4\xa6" => 'H', // Ħ
        "\xc4\xa7" => 'h', // ħ
        "\xc4\xa8" => 'I', // Ĩ
        "\xc4\xa9" => 'i', // ĩ
        "\xc4\xaa" => 'I', // Ī
        "\xc4\xab" => 'i', // ī
        "\xc4\xac" => 'I', // Ĭ
        "\xc4\xad" => 'i', // ĭ
        "\xc4\xae" => 'I', // Į
        "\xc4\xaf" => 'i', // į
        "\xc4\xb0" => 'I', // İ
        "\xc4\xb1" => 'i', // ı
        "\xc4\xb2" => 'IJ', // Ĳ
        "\xc4\xb3" => 'ij', // ĳ
        "\xc4\xb4" => 'J', // Ĵ
        "\xc4\xb5" => 'j', // ĵ
        "\xc4\xb6" => 'K', // Ķ
        "\xc4\xb7" => 'k', // ķ
        "\xc4\xb8" => 'q', // ĸ
        "\xc4\xb9" => 'L', // Ĺ
        "\xc4\xba" => 'l', // ĺ
        "\xc4\xbb" => 'L', // Ļ
        "\xc4\xbc" => 'l', // ļ
        "\xc4\xbd" => 'L', // Ľ
        "\xc4\xbe" => 'l', // ľ
        "\xc4\xbf" => 'L', // Ŀ
        "\xc5\x80" => 'l', // ŀ
        "\xc5\x81" => 'L', // Ł
        "\xc5\x82" => 'l', // ł
        "\xc5\x83" => 'N', // Ń
        "\xc5\x84" => 'n', // ń
        "\xc5\x85" => 'N', // Ņ
        "\xc5\x86" => 'n', // ņ
        "\xc5\x87" => 'N', // Ň
        "\xc5\x88" => 'n', // ň
        "\xc5\x89" => "'n", // ŉ
        "\xc5\x8a" => 'N', // Ŋ
        "\xc5\x8b" => 'n', // ŋ
        "\xc5\x8c" => 'O', // Ō
        "\xc5\x8d" => 'o', // ō
        "\xc5\x8e" => 'O', // Ŏ
        "\xc5\x8f" => 'o', // ŏ
        "\xc5\x90" => 'O', // Ő
        "\xc5\x91" => 'o', // ő
        "\xc5\x94" => 'R', // Ŕ
        "\xc5\x95" => 'r', // ŕ
        "\xc5\x96" => 'R', // Ŗ
        "\xc5\x97" => 'r', // ŗ
        "\xc5\x98" => 'R', // Ř
        "\xc5\x99" => 'r', // ř
        "\xc5\x9a" => 'S', // Ś
        "\xc5\x9b" => 's', // ś
        "\xc5\x9c" => 'S', // Ŝ
        "\xc5\x9d" => 's', // ŝ
        "\xc5\x9e" => 'S', // Ş
        "\xc5\x9f" => 's', // ş
        "\xc5\xa2" => 'T', // Ţ
        "\xc5\xa3" => 't', // ţ
        "\xc5\xa4" => 'T', // Ť
        "\xc5\xa5" => 't', // ť
        "\xc5\xa6" => 'T', // Ŧ
        "\xc5\xa7" => 't', // ŧ
        "\xc5\xa8" => 'U', // Ũ
        "\xc5\xa9" => 'u', // ũ
        "\xc5\xaa" => 'U', // Ū
        "\xc5\xab" => 'u', // ū
        "\xc5\xac" => 'U', // Ŭ
        "\xc5\xad" => 'u', // ŭ
        "\xc5\xae" => 'U', // Ů
        "\xc5\xaf" => 'u', // ů
        "\xc5\xb0" => 'U', // Ű
        "\xc5\xb1" => 'u', // ű
        "\xc5\xb2" => 'U', // Ų
        "\xc5\xb3" => 'u', // ų
        "\xc5\xb4" => 'W', // Ŵ
        "\xc5\xb5" => 'w', // ŵ
        "\xc5\xb6" => 'Y', // Ŷ
        "\xc5\xb7" => 'y', // ŷ
        "\xc5\xb9" => 'Z', // Ź
        "\xc5\xba" => 'z', // ź
        "\xc5\xbb" => 'Z', // Ż
        "\xc5\xbc" => 'z', // ż
        "\xc5\xbf" => 's', // ſ
        "\xc6\x80" => 'b', // ƀ
        "\xc6\x81" => 'B', // Ɓ
        "\xc6\x82" => 'B', // Ƃ
        "\xc6\x83" => 'b', // ƃ
        "\xc6\x87" => 'C', // Ƈ
        "\xc6\x88" => 'c', // ƈ
        "\xc6\x89" => 'D', // Ɖ
        "\xc6\x8a" => 'D', // Ɗ
        "\xc6\x8b" => 'D', // Ƌ
        "\xc6\x8c" => 'd', // ƌ
        "\xc6\x90" => 'E', // Ɛ
        "\xc6\x91" => 'F', // Ƒ
        "\xc6\x93" => 'G', // Ɠ
        "\xc6\x95" => 'hv', // ƕ
        "\xc6\x96" => 'I', // Ɩ
        "\xc6\x97" => 'I', // Ɨ
        "\xc6\x98" => 'K', // Ƙ
        "\xc6\x99" => 'k', // ƙ
        "\xc6\x9a" => 'l', // ƚ
        "\xc6\x9d" => 'N', // Ɲ
        "\xc6\x9e" => 'n', // ƞ
        "\xc6\xa0" => 'O', // Ơ
        "\xc6\xa1" => 'o', // ơ
        "\xc6\xa2" => 'OI', // Ƣ
        "\xc6\xa3" => 'oi', // ƣ
        "\xc6\xa4" => 'P', // Ƥ
        "\xc6\xa5" => 'p', // ƥ
        "\xc6\xab" => 't', // ƫ
        "\xc6\xac" => 'T', // Ƭ
        "\xc6\xad" => 't', // ƭ
        "\xc6\xae" => 'T', // Ʈ
        "\xc6\xaf" => 'U', // Ư
        "\xc6\xb0" => 'u', // ư
        "\xc6\xb2" => 'V', // Ʋ
        "\xc6\xb3" => 'Y', // Ƴ
        "\xc6\xb4" => 'y', // ƴ
        "\xc6\xb5" => 'Z', // Ƶ
        "\xc6\xb6" => 'z', // ƶ
        "\xc7\x84" => 'DZ', // Ǆ
        "\xc7\x85" => 'Dz', // ǅ
        "\xc7\x86" => 'dz', // ǆ
        "\xc7\x87" => 'LJ', // Ǉ
        "\xc7\x88" => 'Lj', // ǈ
        "\xc7\x89" => 'lj', // ǉ
        "\xc7\x8a" => 'NJ', // Ǌ
        "\xc7\x8b" => 'Nj', // ǋ
        "\xc7\x8c" => 'nj', // ǌ
        "\xc7\x8d" => 'A', // Ǎ
        "\xc7\x8e" => 'a', // ǎ
        "\xc7\x8f" => 'I', // Ǐ
        "\xc7\x90" => 'i', // ǐ
        "\xc7\x91" => 'O', // Ǒ
        "\xc7\x92" => 'o', // ǒ
        "\xc7\x93" => 'U', // Ǔ
        "\xc7\x94" => 'u', // ǔ
        "\xc7\x95" => 'U', // Ǖ
        "\xc7\x96" => 'u', // ǖ
        "\xc7\x97" => 'U', // Ǘ
        "\xc7\x98" => 'u', // ǘ
        "\xc7\x99" => 'U', // Ǚ
        "\xc7\x9a" => 'u', // ǚ
        "\xc7\x9b" => 'U', // Ǜ
        "\xc7\x9c" => 'u', // ǜ
        "\xc7\x9e" => 'A', // Ǟ
        "\xc7\x9f" => 'a', // ǟ
        "\xc7\xa0" => 'A', // Ǡ
        "\xc7\xa1" => 'a', // ǡ
        "\xc7\xa2" => 'AE', // Ǣ
        "\xc7\xa3" => 'ae', // ǣ
        "\xc7\xa4" => 'G', // Ǥ
        "\xc7\xa5" => 'g', // ǥ
        "\xc7\xa6" => 'G', // Ǧ
        "\xc7\xa7" => 'g', // ǧ
        "\xc7\xa8" => 'K', // Ǩ
        "\xc7\xa9" => 'k', // ǩ
        "\xc7\xaa" => 'O', // Ǫ
        "\xc7\xab" => 'o', // ǫ
        "\xc7\xac" => 'O', // Ǭ
        "\xc7\xad" => 'o', // ǭ
        "\xc7\xae" => 'Ʒ', // Ǯ
        "\xc7\xaf" => 'ʒ', // ǯ
        "\xc7\xb0" => 'j', // ǰ
        "\xc7\xb1" => 'DZ', // Ǳ
        "\xc7\xb2" => 'Dz', // ǲ
        "\xc7\xb3" => 'dz', // ǳ
        "\xc7\xb4" => 'G', // Ǵ
        "\xc7\xb5" => 'g', // ǵ
        "\xc7\xb8" => 'N', // Ǹ
        "\xc7\xb9" => 'n', // ǹ
        "\xc7\xba" => 'A', // Ǻ
        "\xc7\xbb" => 'a', // ǻ
        "\xc7\xbc" => 'AE', // Ǽ
        "\xc7\xbd" => 'ae', // ǽ
        "\xc7\xbe" => 'O', // Ǿ
        "\xc7\xbf" => 'o', // ǿ
        "\xc8\x80" => 'A', // Ȁ
        "\xc8\x81" => 'a', // ȁ
        "\xc8\x82" => 'A', // Ȃ
        "\xc8\x83" => 'a', // ȃ
        "\xc8\x84" => 'E', // Ȅ
        "\xc8\x85" => 'e', // ȅ
        "\xc8\x86" => 'E', // Ȇ
        "\xc8\x87" => 'e', // ȇ
        "\xc8\x88" => 'I', // Ȉ
        "\xc8\x89" => 'i', // ȉ
        "\xc8\x8a" => 'I', // Ȋ
        "\xc8\x8b" => 'i', // ȋ
        "\xc8\x8c" => 'O', // Ȍ
        "\xc8\x8d" => 'o', // ȍ
        "\xc8\x8e" => 'O', // Ȏ
        "\xc8\x8f" => 'o', // ȏ
        "\xc8\x90" => 'R', // Ȑ
        "\xc8\x91" => 'r', // ȑ
        "\xc8\x92" => 'R', // Ȓ
        "\xc8\x93" => 'r', // ȓ
        "\xc8\x94" => 'U', // Ȕ
        "\xc8\x95" => 'u', // ȕ
        "\xc8\x96" => 'U', // Ȗ
        "\xc8\x97" => 'u', // ȗ
        "\xc8\x98" => 'S', // Ș
        "\xc8\x99" => 's', // ș
        "\xc8\x9a" => 'T', // Ț
        "\xc8\x9b" => 't', // ț
        "\xc8\x9e" => 'H', // Ȟ
        "\xc8\x9f" => 'h', // ȟ
        "\xc8\xa1" => 'd', // ȡ
        "\xc8\xa4" => 'Z', // Ȥ
        "\xc8\xa5" => 'z', // ȥ
        "\xc8\xa6" => 'A', // Ȧ
        "\xc8\xa7" => 'a', // ȧ
        "\xc8\xa8" => 'E', // Ȩ
        "\xc8\xa9" => 'e', // ȩ
        "\xc8\xaa" => 'O', // Ȫ
        "\xc8\xab" => 'o', // ȫ
        "\xc8\xac" => 'O', // Ȭ
        "\xc8\xad" => 'o', // ȭ
        "\xc8\xae" => 'O', // Ȯ
        "\xc8\xaf" => 'o', // ȯ
        "\xc8\xb0" => 'O', // Ȱ
        "\xc8\xb1" => 'o', // ȱ
        "\xc8\xb2" => 'Y', // Ȳ
        "\xc8\xb3" => 'y', // ȳ
        "\xc8\xb4" => 'l', // ȴ
        "\xc8\xb5" => 'n', // ȵ
        "\xc8\xb6" => 't', // ȶ
        "\xc8\xb7" => 'j', // ȷ
        "\xc8\xb8" => 'db', // ȸ
        "\xc8\xb9" => 'qp', // ȹ
        "\xc8\xba" => 'A', // Ⱥ
        "\xc8\xbb" => 'C', // Ȼ
        "\xc8\xbc" => 'c', // ȼ
        "\xc8\xbd" => 'L', // Ƚ
        "\xc8\xbe" => 'T', // Ⱦ
        "\xc8\xbf" => 's', // ȿ
        "\xc9\x80" => 'z', // ɀ
        "\xc9\x83" => 'B', // Ƀ
        "\xc9\x84" => 'U', // Ʉ
        "\xc9\x86" => 'E', // Ɇ
        "\xc9\x87" => 'e', // ɇ
        "\xc9\x88" => 'J', // Ɉ
        "\xc9\x89" => 'j', // ɉ
        "\xc9\x8c" => 'R', // Ɍ
        "\xc9\x8d" => 'r', // ɍ
        "\xc9\x8e" => 'Y', // Ɏ
        "\xc9\x8f" => 'y', // ɏ
        "\xc9\x93" => 'b', // ɓ
        "\xc9\x95" => 'c', // ɕ
        "\xc9\x96" => 'd', // ɖ
        "\xc9\x97" => 'd', // ɗ
        "\xc9\x9b" => 'e', // ɛ
        "\xc9\x9f" => 'j', // ɟ
        "\xc9\xa0" => 'g', // ɠ
        "\xc9\xa1" => 'g', // ɡ
        "\xc9\xa2" => 'G', // ɢ
        "\xc9\xa6" => 'h', // ɦ
        "\xc9\xa7" => 'h', // ɧ
        "\xc9\xa8" => 'i', // ɨ
        "\xc9\xaa" => 'I', // ɪ
        "\xc9\xab" => 'l', // ɫ
        "\xc9\xac" => 'l', // ɬ
        "\xc9\xad" => 'l', // ɭ
        "\xc9\xb1" => 'm', // ɱ
        "\xc9\xb2" => 'n', // ɲ
        "\xc9\xb3" => 'n', // ɳ
        "\xc9\xb4" => 'N', // ɴ
        "\xc9\xb6" => 'OE', // ɶ
        "\xc9\xbc" => 'r', // ɼ
        "\xc9\xbd" => 'r', // ɽ
        "\xc9\xbe" => 'r', // ɾ
        "\xca\x80" => 'R', // ʀ
        "\xca\x82" => 's', // ʂ
        "\xca\x88" => 't', // ʈ
        "\xca\x89" => 'u', // ʉ
        "\xca\x8b" => 'v', // ʋ
        "\xca\x8f" => 'Y', // ʏ
        "\xca\x90" => 'z', // ʐ
        "\xca\x91" => 'z', // ʑ
        "\xca\x99" => 'B', // ʙ
        "\xca\x9b" => 'G', // ʛ
        "\xca\x9c" => 'H', // ʜ
        "\xca\x9d" => 'j', // ʝ
        "\xca\x9f" => 'L', // ʟ
        "\xca\xa0" => 'q', // ʠ
        "\xca\xa3" => 'dz', // ʣ
        "\xca\xa5" => 'dz', // ʥ
        "\xca\xa6" => 'ts', // ʦ
        "\xca\xaa" => 'ls', // ʪ
        "\xca\xab" => 'lz', // ʫ
        "\xcd\x80" => '̀', // ̀
        "\xcd\x81" => '́', // ́
        "\xcd\x83" => '̓', // ̓
        "\xcd\x84" => '̈́', // ̈́
        "\xcd\xb4" => 'ʹ', // ʹ
        "\xcd\xba" => 'i', // ͺ
        "\xcd\xbe" => ';', // ;
        "\xce\x86" => 'A', // Ά
        "\xce\x87" => '·', // ·
        "\xce\x88" => 'E', // Έ
        "\xce\x89" => 'E', // Ή
        "\xce\x8a" => 'I', // Ί
        "\xce\x8c" => 'O', // Ό
        "\xce\x8e" => 'Y', // Ύ
        "\xce\x8f" => 'O', // Ώ
        "\xce\x90" => 'i', // ΐ
        "\xce\x91" => 'A', // Α
        "\xce\x92" => 'B', // Β
        "\xce\x93" => 'G', // Γ
        "\xce\x94" => 'D', // Δ
        "\xce\x95" => 'E', // Ε
        "\xce\x96" => 'Z', // Ζ
        "\xce\x97" => 'E', // Η
        "\xce\x98" => 'TH', // Θ
        "\xce\x99" => 'I', // Ι
        "\xce\x9a" => 'K', // Κ
        "\xce\x9b" => 'L', // Λ
        "\xce\x9c" => 'M', // Μ
        "\xce\x9d" => 'N', // Ν
        "\xce\x9e" => 'X', // Ξ
        "\xce\x9f" => 'O', // Ο
        "\xce\xa0" => 'P', // Π
        "\xce\xa1" => 'R', // Ρ
        "\xce\xa3" => 'S', // Σ
        "\xce\xa4" => 'T', // Τ
        "\xce\xa5" => 'Y', // Υ
        "\xce\xa6" => 'PH', // Φ
        "\xce\xa7" => 'CH', // Χ
        "\xce\xa8" => 'PS', // Ψ
        "\xce\xa9" => 'O', // Ω
        "\xce\xaa" => 'I', // Ϊ
        "\xce\xab" => 'Y', // Ϋ
        "\xce\xac" => 'a', // ά
        "\xce\xad" => 'e', // έ
        "\xce\xae" => 'e', // ή
        "\xce\xaf" => 'i', // ί
        "\xce\xb0" => 'y', // ΰ
        "\xce\xb1" => 'a', // α
        "\xce\xb2" => 'b', // β
        "\xce\xb3" => 'g', // γ
        "\xce\xb4" => 'd', // δ
        "\xce\xb5" => 'e', // ε
        "\xce\xb6" => 'z', // ζ
        "\xce\xb7" => 'e', // η
        "\xce\xb8" => 'th', // θ
        "\xce\xb9" => 'i', // ι
        "\xce\xba" => 'k', // κ
        "\xce\xbb" => 'l', // λ
        "\xce\xbc" => 'm', // μ
        "\xce\xbd" => 'n', // ν
        "\xce\xbe" => 'x', // ξ
        "\xce\xbf" => 'o', // ο
        "\xcf\x80" => 'p', // π
        "\xcf\x81" => 'r', // ρ
        "\xcf\x82" => 's', // ς
        "\xcf\x83" => 's', // σ
        "\xcf\x84" => 't', // τ
        "\xcf\x85" => 'y', // υ
        "\xcf\x86" => 'ph', // φ
        "\xcf\x87" => 'ch', // χ
        "\xcf\x88" => 'ps', // ψ
        "\xcf\x89" => 'o', // ω
        "\xcf\x8a" => 'i', // ϊ
        "\xcf\x8b" => 'y', // ϋ
        "\xcf\x8c" => 'o', // ό
        "\xcf\x8d" => 'y', // ύ
        "\xcf\x8e" => 'o', // ώ
        "\xcf\x90" => 'b', // ϐ
        "\xcf\x91" => 'th', // ϑ
        "\xcf\x92" => 'Y', // ϒ
        "\xcf\x93" => 'Y', // ϓ
        "\xcf\x94" => 'Y', // ϔ
        "\xcf\x95" => 'ph', // ϕ
        "\xcf\x96" => 'p', // ϖ
        "\xcf\xb0" => 'k', // ϰ
        "\xcf\xb1" => 'r', // ϱ
        "\xcf\xb2" => 's', // ϲ
        "\xcf\xb3" => 'j', // ϳ
        "\xcf\xb4" => 'TH', // ϴ
        "\xcf\xb5" => 'e', // ϵ
        "\xcf\xb7" => 'S', // Ϸ
        "\xcf\xb8" => 's', // ϸ
        "\xcf\xb9" => 'S', // Ϲ
        "\xcf\xba" => 'S', // Ϻ
        "\xcf\xbb" => 's', // ϻ
        "\xd0\x80" => 'E', // Ѐ
        "\xd0\x81" => 'E', // Ё
        "\xd0\x82" => 'D', // Ђ
        "\xd0\x83" => 'G', // Ѓ
        "\xd0\x84" => 'E', // Є
        "\xd0\x85" => 'Z', // Ѕ
        "\xd0\x86" => 'I', // І
        "\xd0\x87" => 'I', // Ї
        "\xd0\x88" => 'J', // Ј
        "\xd0\x89" => 'L', // Љ
        "\xd0\x8a" => 'N', // Њ
        "\xd0\x8b" => 'C', // Ћ
        "\xd0\x8c" => 'K', // Ќ
        "\xd0\x8d" => 'I', // Ѝ
        "\xd0\x8e" => 'U', // Ў
        "\xd0\x8f" => 'D', // Џ
        "\xd0\x90" => 'A', // А
        "\xd0\x91" => 'B', // Б
        "\xd0\x92" => 'V', // В
        "\xd0\x93" => 'G', // Г
        "\xd0\x94" => 'D', // Д
        "\xd0\x95" => 'E', // Е
        "\xd0\x96" => 'Z', // Ж
        "\xd0\x97" => 'Z', // З
        "\xd0\x98" => 'I', // И
        "\xd0\x99" => 'J', // Й
        "\xd0\x9a" => 'K', // К
        "\xd0\x9b" => 'L', // Л
        "\xd0\x9c" => 'M', // М
        "\xd0\x9d" => 'N', // Н
        "\xd0\x9e" => 'O', // О
        "\xd0\x9f" => 'P', // П
        "\xd0\xa0" => 'R', // Р
        "\xd0\xa1" => 'S', // С
        "\xd0\xa2" => 'T', // Т
        "\xd0\xa3" => 'U', // У
        "\xd0\xa4" => 'F', // Ф
        "\xd0\xa5" => 'H', // Х
        "\xd0\xa6" => 'C', // Ц
        "\xd0\xa7" => 'C', // Ч
        "\xd0\xa8" => 'S', // Ш
        "\xd0\xa9" => 'S', // Щ
        "\xd0\xaa" => 'ʺ̱', // Ъ
        "\xd0\xab" => 'Y', // Ы
        "\xd0\xac" => 'ʹ̱', // Ь
        "\xd0\xad" => 'E', // Э
        "\xd0\xae" => 'U', // Ю
        "\xd0\xaf" => 'A', // Я
        "\xd0\xb0" => 'a', // а
        "\xd0\xb1" => 'b', // б
        "\xd0\xb2" => 'v', // в
        "\xd0\xb3" => 'g', // г
        "\xd0\xb4" => 'd', // д
        "\xd0\xb5" => 'e', // е
        "\xd0\xb6" => 'z', // ж
        "\xd0\xb7" => 'z', // з
        "\xd0\xb8" => 'i', // и
        "\xd0\xb9" => 'j', // й
        "\xd0\xba" => 'k', // к
        "\xd0\xbb" => 'l', // л
        "\xd0\xbc" => 'm', // м
        "\xd0\xbd" => 'n', // н
        "\xd0\xbe" => 'o', // о
        "\xd0\xbf" => 'p', // п
        "\xd1\x80" => 'r', // р
        "\xd1\x81" => 's', // с
        "\xd1\x82" => 't', // т
        "\xd1\x83" => 'u', // у
        "\xd1\x84" => 'f', // ф
        "\xd1\x85" => 'h', // х
        "\xd1\x86" => 'c', // ц
        "\xd1\x87" => 'c', // ч
        "\xd1\x88" => 's', // ш
        "\xd1\x89" => 's', // щ
        "\xd1\x8a" => 'ʺ', // ъ
        "\xd1\x8b" => 'y', // ы
        "\xd1\x8c" => 'ʹ', // ь
        "\xd1\x8d" => 'e', // э
        "\xd1\x8e" => 'u', // ю
        "\xd1\x8f" => 'a', // я
        "\xd1\x90" => 'e', // ѐ
        "\xd1\x91" => 'e', // ё
        "\xd1\x92" => 'd', // ђ
        "\xd1\x93" => 'g', // ѓ
        "\xd1\x94" => 'e', // є
        "\xd1\x95" => 'z', // ѕ
        "\xd1\x96" => 'i', // і
        "\xd1\x97" => 'i', // ї
        "\xd1\x98" => 'j', // ј
        "\xd1\x99" => 'l', // љ
        "\xd1\x9a" => 'n', // њ
        "\xd1\x9b" => 'c', // ћ
        "\xd1\x9c" => 'k', // ќ
        "\xd1\x9d" => 'i', // ѝ
        "\xd1\x9e" => 'u', // ў
        "\xd1\x9f" => 'd', // џ
        "\xd2\x90" => 'G', // Ґ
        "\xd2\x91" => 'g', // ґ
        "\xd2\x92" => 'G', // Ғ
        "\xd2\x93" => 'g', // ғ
        "\xd2\x94" => 'G', // Ҕ
        "\xd2\x95" => 'g', // ҕ
        "\xd2\x98" => 'Z', // Ҙ
        "\xd2\x99" => 'z', // ҙ
        "\xd2\x9a" => 'Kˌ', // Қ
        "\xd2\x9b" => 'kˌ', // қ
        "\xd3\x81" => 'Z', // Ӂ
        "\xd3\x82" => 'z', // ӂ
        "\xd3\x90" => 'A', // Ӑ
        "\xd3\x91" => 'a', // ӑ
        "\xd3\x92" => 'A', // Ӓ
        "\xd3\x93" => 'a', // ӓ
        "\xd3\x94" => 'AE', // Ӕ
        "\xd3\x95" => 'ae', // ӕ
        "\xd3\x96" => 'E', // Ӗ
        "\xd3\x97" => 'e', // ӗ
        "\xd3\x98" => 'Ə', // Ә
        "\xd3\x99" => 'ə', // ә
        "\xd3\x9a" => 'Ə', // Ӛ
        "\xd3\x9b" => 'ə', // ӛ
        "\xd3\x9c" => 'Z', // Ӝ
        "\xd3\x9d" => 'z', // ӝ
        "\xd3\x9e" => 'Z', // Ӟ
        "\xd3\x9f" => 'z', // ӟ
        "\xd3\xa2" => 'I', // Ӣ
        "\xd3\xa3" => 'i', // ӣ
        "\xd3\xa4" => 'I', // Ӥ
        "\xd3\xa5" => 'i', // ӥ
        "\xd3\xa6" => 'O', // Ӧ
        "\xd3\xa7" => 'o', // ӧ
        "\xd3\xac" => 'E', // Ӭ
        "\xd3\xad" => 'e', // ӭ
        "\xd3\xae" => 'U', // Ӯ
        "\xd3\xaf" => 'u', // ӯ
        "\xd3\xb0" => 'U', // Ӱ
        "\xd3\xb1" => 'u', // ӱ
        "\xd3\xb2" => 'U', // Ӳ
        "\xd3\xb3" => 'u', // ӳ
        "\xd3\xb4" => 'C', // Ӵ
        "\xd3\xb5" => 'c', // ӵ
        "\xd3\xb8" => 'Y', // Ӹ
        "\xd3\xb9" => 'y', // ӹ
        "\xd4\xb1" => 'A', // Ա
        "\xd4\xb2" => 'B', // Բ
        "\xd4\xb3" => 'G', // Գ
        "\xd4\xb4" => 'D', // Դ
        "\xd4\xb5" => 'E', // Ե
        "\xd4\xb6" => 'Z', // Զ
        "\xd4\xb7" => 'E', // Է
        "\xd4\xb8" => 'Ə', // Ը
        "\xd4\xb9" => 'Tʻ', // Թ
        "\xd4\xba" => 'Z', // Ժ
        "\xd4\xbb" => 'I', // Ի
        "\xd4\xbc" => 'L', // Լ
        "\xd4\xbd" => 'X', // Խ
        "\xd4\xbe" => 'C', // Ծ
        "\xd4\xbf" => 'K', // Կ
        "\xd5\x80" => 'H', // Հ
        "\xd5\x81" => 'J', // Ձ
        "\xd5\x82" => 'G', // Ղ
        "\xd5\x83" => 'C', // Ճ
        "\xd5\x84" => 'M', // Մ
        "\xd5\x85" => 'Y', // Յ
        "\xd5\x86" => 'N', // Ն
        "\xd5\x87" => 'S', // Շ
        "\xd5\x88" => 'O', // Ո
        "\xd5\x89" => 'Cʻ', // Չ
        "\xd5\x8a" => 'P', // Պ
        "\xd5\x8b" => 'J', // Ջ
        "\xd5\x8c" => 'R', // Ռ
        "\xd5\x8d" => 'S', // Ս
        "\xd5\x8e" => 'V', // Վ
        "\xd5\x8f" => 'T', // Տ
        "\xd5\x90" => 'R', // Ր
        "\xd5\x91" => 'Cʻ', // Ց
        "\xd5\x92" => 'W', // Ւ
        "\xd5\x93" => 'Pʻ', // Փ
        "\xd5\x94" => 'Kʻ', // Ք
        "\xd5\x95" => 'O', // Օ
        "\xd5\x96" => 'F', // Ֆ
        "\xd5\xa1" => 'a', // ա
        "\xd5\xa2" => 'b', // բ
        "\xd5\xa3" => 'g', // գ
        "\xd5\xa4" => 'd', // դ
        "\xd5\xa5" => 'e', // ե
        "\xd5\xa6" => 'z', // զ
        "\xd5\xa7" => 'e', // է
        "\xd5\xa8" => 'ə', // ը
        "\xd5\xa9" => 'tʻ', // թ
        "\xd5\xaa" => 'z', // ժ
        "\xd5\xab" => 'i', // ի
        "\xd5\xac" => 'l', // լ
        "\xd5\xad" => 'x', // խ
        "\xd5\xae" => 'c', // ծ
        "\xd5\xaf" => 'k', // կ
        "\xd5\xb0" => 'h', // հ
        "\xd5\xb1" => 'j', // ձ
        "\xd5\xb2" => 'g', // ղ
        "\xd5\xb3" => 'c', // ճ
        "\xd5\xb4" => 'm', // մ
        "\xd5\xb5" => 'y', // յ
        "\xd5\xb6" => 'n', // ն
        "\xd5\xb7" => 's', // շ
        "\xd5\xb8" => 'o', // ո
        "\xd5\xb9" => 'cʻ', // չ
        "\xd5\xba" => 'p', // պ
        "\xd5\xbb" => 'j', // ջ
        "\xd5\xbc" => 'r', // ռ
        "\xd5\xbd" => 's', // ս
        "\xd5\xbe" => 'v', // վ
        "\xd5\xbf" => 't', // տ
        "\xd6\x80" => 'r', // ր
        "\xd6\x81" => 'cʻ', // ց
        "\xd6\x82" => 'w', // ւ
        "\xd6\x83" => 'pʻ', // փ
        "\xd6\x84" => 'kʻ', // ք
        "\xd6\x85" => 'o', // օ
        "\xd6\x86" => 'f', // ֆ
        "\xd6\x87" => 'ev', // և
        "\xe1\x83\x90" => 'a', // ა
        "\xe1\x83\x91" => 'b', // ბ
        "\xe1\x83\x92" => 'g', // გ
        "\xe1\x83\x93" => 'd', // დ
        "\xe1\x83\x94" => 'e', // ე
        "\xe1\x83\x95" => 'v', // ვ
        "\xe1\x83\x96" => 'z', // ზ
        "\xe1\x83\x97" => 't', // თ
        "\xe1\x83\x98" => 'i', // ი
        "\xe1\x83\x99" => 'kʼ', // კ
        "\xe1\x83\x9a" => 'l', // ლ
        "\xe1\x83\x9b" => 'm', // მ
        "\xe1\x83\x9c" => 'n', // ნ
        "\xe1\x83\x9d" => 'o', // ო
        "\xe1\x83\x9e" => 'pʼ', // პ
        "\xe1\x83\x9f" => 'zh', // ჟ
        "\xe1\x83\xa0" => 'r', // რ
        "\xe1\x83\xa1" => 's', // ს
        "\xe1\x83\xa2" => 'tʼ', // ტ
        "\xe1\x83\xa3" => 'u', // უ
        "\xe1\x83\xa4" => 'p', // ფ
        "\xe1\x83\xa5" => 'k', // ქ
        "\xe1\x83\xa6" => 'gh', // ღ
        "\xe1\x83\xa7" => 'qʼ', // ყ
        "\xe1\x83\xa8" => 'sh', // შ
        "\xe1\x83\xa9" => 'ch', // ჩ
        "\xe1\x83\xaa" => 'ts', // ც
        "\xe1\x83\xab" => 'dz', // ძ
        "\xe1\x83\xac" => 'tsʼ', // წ
        "\xe1\x83\xad" => 'chʼ', // ჭ
        "\xe1\x83\xae" => 'kh', // ხ
        "\xe1\x83\xaf" => 'j', // ჯ
        "\xe1\x83\xb0" => 'h', // ჰ
        "\xe1\x83\xb3" => 'ui', // ჳ
        "\xe1\x83\xb4" => 'q', // ჴ
        "\xe1\xb8\x80" => 'A', // Ḁ
        "\xe1\xb8\x81" => 'a', // ḁ
        "\xe1\xb8\x82" => 'B', // Ḃ
        "\xe1\xb8\x83" => 'b', // ḃ
        "\xe1\xb8\x84" => 'B', // Ḅ
        "\xe1\xb8\x85" => 'b', // ḅ
        "\xe1\xb8\x86" => 'B', // Ḇ
        "\xe1\xb8\x87" => 'b', // ḇ
        "\xe1\xb8\x88" => 'C', // Ḉ
        "\xe1\xb8\x89" => 'c', // ḉ
        "\xe1\xb8\x8a" => 'D', // Ḋ
        "\xe1\xb8\x8b" => 'd', // ḋ
        "\xe1\xb8\x8c" => 'D', // Ḍ
        "\xe1\xb8\x8d" => 'd', // ḍ
        "\xe1\xb8\x8e" => 'D', // Ḏ
        "\xe1\xb8\x8f" => 'd', // ḏ
        "\xe1\xb8\x90" => 'D', // Ḑ
        "\xe1\xb8\x91" => 'd', // ḑ
        "\xe1\xb8\x92" => 'D', // Ḓ
        "\xe1\xb8\x93" => 'd', // ḓ
        "\xe1\xb8\x94" => 'E', // Ḕ
        "\xe1\xb8\x95" => 'e', // ḕ
        "\xe1\xb8\x96" => 'E', // Ḗ
        "\xe1\xb8\x97" => 'e', // ḗ
        "\xe1\xb8\x98" => 'E', // Ḙ
        "\xe1\xb8\x99" => 'e', // ḙ
        "\xe1\xb8\x9a" => 'E', // Ḛ
        "\xe1\xb8\x9b" => 'e', // ḛ
        "\xe1\xb8\x9c" => 'E', // Ḝ
        "\xe1\xb8\x9d" => 'e', // ḝ
        "\xe1\xb8\x9e" => 'F', // Ḟ
        "\xe1\xb8\x9f" => 'f', // ḟ
        "\xe1\xb8\xa0" => 'G', // Ḡ
        "\xe1\xb8\xa1" => 'g', // ḡ
        "\xe1\xb8\xa2" => 'H', // Ḣ
        "\xe1\xb8\xa3" => 'h', // ḣ
        "\xe1\xb8\xa4" => 'H', // Ḥ
        "\xe1\xb8\xa5" => 'h', // ḥ
        "\xe1\xb8\xa6" => 'H', // Ḧ
        "\xe1\xb8\xa7" => 'h', // ḧ
        "\xe1\xb8\xa8" => 'H', // Ḩ
        "\xe1\xb8\xa9" => 'h', // ḩ
        "\xe1\xb8\xaa" => 'H', // Ḫ
        "\xe1\xb8\xab" => 'h', // ḫ
        "\xe1\xb8\xac" => 'I', // Ḭ
        "\xe1\xb8\xad" => 'i', // ḭ
        "\xe1\xb8\xae" => 'I', // Ḯ
        "\xe1\xb8\xaf" => 'i', // ḯ
        "\xe1\xb8\xb0" => 'K', // Ḱ
        "\xe1\xb8\xb1" => 'k', // ḱ
        "\xe1\xb8\xb2" => 'K', // Ḳ
        "\xe1\xb8\xb3" => 'k', // ḳ
        "\xe1\xb8\xb4" => 'K', // Ḵ
        "\xe1\xb8\xb5" => 'k', // ḵ
        "\xe1\xb8\xb6" => 'L', // Ḷ
        "\xe1\xb8\xb7" => 'l', // ḷ
        "\xe1\xb8\xb8" => 'L', // Ḹ
        "\xe1\xb8\xb9" => 'l', // ḹ
        "\xe1\xb8\xba" => 'L', // Ḻ
        "\xe1\xb8\xbb" => 'l', // ḻ
        "\xe1\xb8\xbc" => 'L', // Ḽ
        "\xe1\xb8\xbd" => 'l', // ḽ
        "\xe1\xb8\xbe" => 'M', // Ḿ
        "\xe1\xb8\xbf" => 'm', // ḿ
        "\xe1\xb9\x80" => 'M', // Ṁ
        "\xe1\xb9\x81" => 'm', // ṁ
        "\xe1\xb9\x82" => 'M', // Ṃ
        "\xe1\xb9\x83" => 'm', // ṃ
        "\xe1\xb9\x84" => 'N', // Ṅ
        "\xe1\xb9\x85" => 'n', // ṅ
        "\xe1\xb9\x86" => 'N', // Ṇ
        "\xe1\xb9\x87" => 'n', // ṇ
        "\xe1\xb9\x88" => 'N', // Ṉ
        "\xe1\xb9\x89" => 'n', // ṉ
        "\xe1\xb9\x8a" => 'N', // Ṋ
        "\xe1\xb9\x8b" => 'n', // ṋ
        "\xe1\xb9\x8c" => 'O', // Ṍ
        "\xe1\xb9\x8d" => 'o', // ṍ
        "\xe1\xb9\x8e" => 'O', // Ṏ
        "\xe1\xb9\x8f" => 'o', // ṏ
        "\xe1\xb9\x90" => 'O', // Ṑ
        "\xe1\xb9\x91" => 'o', // ṑ
        "\xe1\xb9\x92" => 'O', // Ṓ
        "\xe1\xb9\x93" => 'o', // ṓ
        "\xe1\xb9\x94" => 'P', // Ṕ
        "\xe1\xb9\x95" => 'p', // ṕ
        "\xe1\xb9\x96" => 'P', // Ṗ
        "\xe1\xb9\x97" => 'p', // ṗ
        "\xe1\xb9\x98" => 'R', // Ṙ
        "\xe1\xb9\x99" => 'r', // ṙ
        "\xe1\xb9\x9a" => 'R', // Ṛ
        "\xe1\xb9\x9b" => 'r', // ṛ
        "\xe1\xb9\x9c" => 'R', // Ṝ
        "\xe1\xb9\x9d" => 'r', // ṝ
        "\xe1\xb9\x9e" => 'R', // Ṟ
        "\xe1\xb9\x9f" => 'r', // ṟ
        "\xe1\xb9\xa0" => 'S', // Ṡ
        "\xe1\xb9\xa1" => 's', // ṡ
        "\xe1\xb9\xa2" => 'S', // Ṣ
        "\xe1\xb9\xa3" => 's', // ṣ
        "\xe1\xb9\xa4" => 'S', // Ṥ
        "\xe1\xb9\xa5" => 's', // ṥ
        "\xe1\xb9\xa6" => 'S', // Ṧ
        "\xe1\xb9\xa7" => 's', // ṧ
        "\xe1\xb9\xa8" => 'S', // Ṩ
        "\xe1\xb9\xa9" => 's', // ṩ
        "\xe1\xb9\xaa" => 'T', // Ṫ
        "\xe1\xb9\xab" => 't', // ṫ
        "\xe1\xb9\xac" => 'T', // Ṭ
        "\xe1\xb9\xad" => 't', // ṭ
        "\xe1\xb9\xae" => 'T', // Ṯ
        "\xe1\xb9\xaf" => 't', // ṯ
        "\xe1\xb9\xb0" => 'T', // Ṱ
        "\xe1\xb9\xb1" => 't', // ṱ
        "\xe1\xb9\xb2" => 'U', // Ṳ
        "\xe1\xb9\xb3" => 'u', // ṳ
        "\xe1\xb9\xb4" => 'U', // Ṵ
        "\xe1\xb9\xb5" => 'u', // ṵ
        "\xe1\xb9\xb6" => 'U', // Ṷ
        "\xe1\xb9\xb7" => 'u', // ṷ
        "\xe1\xb9\xb8" => 'U', // Ṹ
        "\xe1\xb9\xb9" => 'u', // ṹ
        "\xe1\xb9\xba" => 'U', // Ṻ
        "\xe1\xb9\xbb" => 'u', // ṻ
        "\xe1\xb9\xbc" => 'V', // Ṽ
        "\xe1\xb9\xbd" => 'v', // ṽ
        "\xe1\xb9\xbe" => 'V', // Ṿ
        "\xe1\xb9\xbf" => 'v', // ṿ
        "\xe1\xba\x80" => 'W', // Ẁ
        "\xe1\xba\x81" => 'w', // ẁ
        "\xe1\xba\x82" => 'W', // Ẃ
        "\xe1\xba\x83" => 'w', // ẃ
        "\xe1\xba\x84" => 'W', // Ẅ
        "\xe1\xba\x85" => 'w', // ẅ
        "\xe1\xba\x86" => 'W', // Ẇ
        "\xe1\xba\x87" => 'w', // ẇ
        "\xe1\xba\x88" => 'W', // Ẉ
        "\xe1\xba\x89" => 'w', // ẉ
        "\xe1\xba\x8a" => 'X', // Ẋ
        "\xe1\xba\x8b" => 'x', // ẋ
        "\xe1\xba\x8c" => 'X', // Ẍ
        "\xe1\xba\x8d" => 'x', // ẍ
        "\xe1\xba\x8e" => 'Y', // Ẏ
        "\xe1\xba\x8f" => 'y', // ẏ
        "\xe1\xba\x90" => 'Z', // Ẑ
        "\xe1\xba\x91" => 'z', // ẑ
        "\xe1\xba\x92" => 'Z', // Ẓ
        "\xe1\xba\x93" => 'z', // ẓ
        "\xe1\xba\x94" => 'Z', // Ẕ
        "\xe1\xba\x95" => 'z', // ẕ
        "\xe1\xba\x96" => 'h', // ẖ
        "\xe1\xba\x97" => 't', // ẗ
        "\xe1\xba\x98" => 'w', // ẘ
        "\xe1\xba\x99" => 'y', // ẙ
        "\xe1\xba\x9a" => 'a', // ẚ
        "\xe1\xba\x9b" => 's', // ẛ
        "\xe1\xba\x9c" => 's', // ẜ
        "\xe1\xba\x9d" => 's', // ẝ
        "\xe1\xba\x9e" => 'SS', // ẞ
        "\xe1\xba\xa0" => 'A', // Ạ
        "\xe1\xba\xa1" => 'a', // ạ
        "\xe1\xba\xa2" => 'A', // Ả
        "\xe1\xba\xa3" => 'a', // ả
        "\xe1\xba\xa4" => 'A', // Ấ
        "\xe1\xba\xa5" => 'a', // ấ
        "\xe1\xba\xa6" => 'A', // Ầ
        "\xe1\xba\xa7" => 'a', // ầ
        "\xe1\xba\xa8" => 'A', // Ẩ
        "\xe1\xba\xa9" => 'a', // ẩ
        "\xe1\xba\xaa" => 'A', // Ẫ
        "\xe1\xba\xab" => 'a', // ẫ
        "\xe1\xba\xac" => 'A', // Ậ
        "\xe1\xba\xad" => 'a', // ậ
        "\xe1\xba\xae" => 'A', // Ắ
        "\xe1\xba\xaf" => 'a', // ắ
        "\xe1\xba\xb0" => 'A', // Ằ
        "\xe1\xba\xb1" => 'a', // ằ
        "\xe1\xba\xb2" => 'A', // Ẳ
        "\xe1\xba\xb3" => 'a', // ẳ
        "\xe1\xba\xb4" => 'A', // Ẵ
        "\xe1\xba\xb5" => 'a', // ẵ
        "\xe1\xba\xb6" => 'A', // Ặ
        "\xe1\xba\xb7" => 'a', // ặ
        "\xe1\xba\xb8" => 'E', // Ẹ
        "\xe1\xba\xb9" => 'e', // ẹ
        "\xe1\xba\xba" => 'E', // Ẻ
        "\xe1\xba\xbb" => 'e', // ẻ
        "\xe1\xba\xbc" => 'E', // Ẽ
        "\xe1\xba\xbd" => 'e', // ẽ
        "\xe1\xba\xbe" => 'E', // Ế
        "\xe1\xba\xbf" => 'e', // ế
        "\xe1\xbb\x80" => 'E', // Ề
        "\xe1\xbb\x81" => 'e', // ề
        "\xe1\xbb\x82" => 'E', // Ể
        "\xe1\xbb\x83" => 'e', // ể
        "\xe1\xbb\x84" => 'E', // Ễ
        "\xe1\xbb\x85" => 'e', // ễ
        "\xe1\xbb\x86" => 'E', // Ệ
        "\xe1\xbb\x87" => 'e', // ệ
        "\xe1\xbb\x88" => 'I', // Ỉ
        "\xe1\xbb\x89" => 'i', // ỉ
        "\xe1\xbb\x8a" => 'I', // Ị
        "\xe1\xbb\x8b" => 'i', // ị
        "\xe1\xbb\x8c" => 'O', // Ọ
        "\xe1\xbb\x8d" => 'o', // ọ
        "\xe1\xbb\x8e" => 'O', // Ỏ
        "\xe1\xbb\x8f" => 'o', // ỏ
        "\xe1\xbb\x90" => 'O', // Ố
        "\xe1\xbb\x91" => 'o', // ố
        "\xe1\xbb\x92" => 'O', // Ồ
        "\xe1\xbb\x93" => 'o', // ồ
        "\xe1\xbb\x94" => 'O', // Ổ
        "\xe1\xbb\x95" => 'o', // ổ
        "\xe1\xbb\x96" => 'O', // Ỗ
        "\xe1\xbb\x97" => 'o', // ỗ
        "\xe1\xbb\x98" => 'O', // Ộ
        "\xe1\xbb\x99" => 'o', // ộ
        "\xe1\xbb\x9a" => 'O', // Ớ
        "\xe1\xbb\x9b" => 'o', // ớ
        "\xe1\xbb\x9c" => 'O', // Ờ
        "\xe1\xbb\x9d" => 'o', // ờ
        "\xe1\xbb\x9e" => 'O', // Ở
        "\xe1\xbb\x9f" => 'o', // ở
        "\xe1\xbb\xa0" => 'O', // Ỡ
        "\xe1\xbb\xa1" => 'o', // ỡ
        "\xe1\xbb\xa2" => 'O', // Ợ
        "\xe1\xbb\xa3" => 'o', // ợ
        "\xe1\xbb\xa4" => 'U', // Ụ
        "\xe1\xbb\xa5" => 'u', // ụ
        "\xe1\xbb\xa6" => 'U', // Ủ
        "\xe1\xbb\xa7" => 'u', // ủ
        "\xe1\xbb\xa8" => 'U', // Ứ
        "\xe1\xbb\xa9" => 'u', // ứ
        "\xe1\xbb\xaa" => 'U', // Ừ
        "\xe1\xbb\xab" => 'u', // ừ
        "\xe1\xbb\xac" => 'U', // Ử
        "\xe1\xbb\xad" => 'u', // ử
        "\xe1\xbb\xae" => 'U', // Ữ
        "\xe1\xbb\xaf" => 'u', // ữ
        "\xe1\xbb\xb0" => 'U', // Ự
        "\xe1\xbb\xb1" => 'u', // ự
        "\xe1\xbb\xb2" => 'Y', // Ỳ
        "\xe1\xbb\xb3" => 'y', // ỳ
        "\xe1\xbb\xb4" => 'Y', // Ỵ
        "\xe1\xbb\xb5" => 'y', // ỵ
        "\xe1\xbb\xb6" => 'Y', // Ỷ
        "\xe1\xbb\xb7" => 'y', // ỷ
        "\xe1\xbb\xb8" => 'Y', // Ỹ
        "\xe1\xbb\xb9" => 'y', // ỹ
        "\xe1\xbb\xba" => 'LL', // Ỻ
        "\xe1\xbb\xbb" => 'll', // ỻ
        "\xe1\xbb\xbc" => 'V', // Ỽ
        "\xe1\xbb\xbd" => 'v', // ỽ
        "\xe1\xbb\xbe" => 'Y', // Ỿ
        "\xe1\xbb\xbf" => 'y', // ỿ
        "\xe1\xbc\x80" => 'a', // ἀ
        "\xe1\xbc\x81" => 'ha', // ἁ
        "\xe1\xbc\x82" => 'a', // ἂ
        "\xe1\xbc\x83" => 'ha', // ἃ
        "\xe1\xbc\x84" => 'a', // ἄ
        "\xe1\xbc\x85" => 'ha', // ἅ
        "\xe1\xbc\x86" => 'a', // ἆ
        "\xe1\xbc\x87" => 'ha', // ἇ
        "\xe1\xbc\x88" => 'A', // Ἀ
        "\xe1\xbc\x89" => 'HA', // Ἁ
        "\xe1\xbc\x8a" => 'A', // Ἂ
        "\xe1\xbc\x8b" => 'HA', // Ἃ
        "\xe1\xbc\x8c" => 'A', // Ἄ
        "\xe1\xbc\x8d" => 'HA', // Ἅ
        "\xe1\xbc\x8e" => 'A', // Ἆ
        "\xe1\xbc\x8f" => 'HA', // Ἇ
        "\xe1\xbc\x90" => 'e', // ἐ
        "\xe1\xbc\x91" => 'he', // ἑ
        "\xe1\xbc\x92" => 'e', // ἒ
        "\xe1\xbc\x93" => 'he', // ἓ
        "\xe1\xbc\x94" => 'e', // ἔ
        "\xe1\xbc\x95" => 'he', // ἕ
        "\xe1\xbc\x98" => 'E', // Ἐ
        "\xe1\xbc\x99" => 'HE', // Ἑ
        "\xe1\xbc\x9a" => 'E', // Ἒ
        "\xe1\xbc\x9b" => 'HE', // Ἓ
        "\xe1\xbc\x9c" => 'E', // Ἔ
        "\xe1\xbc\x9d" => 'HE', // Ἕ
        "\xe1\xbc\xa0" => 'e', // ἠ
        "\xe1\xbc\xa1" => 'he', // ἡ
        "\xe1\xbc\xa2" => 'e', // ἢ
        "\xe1\xbc\xa3" => 'he', // ἣ
        "\xe1\xbc\xa4" => 'e', // ἤ
        "\xe1\xbc\xa5" => 'he', // ἥ
        "\xe1\xbc\xa6" => 'e', // ἦ
        "\xe1\xbc\xa7" => 'he', // ἧ
        "\xe1\xbc\xa8" => 'E', // Ἠ
        "\xe1\xbc\xa9" => 'HE', // Ἡ
        "\xe1\xbc\xaa" => 'E', // Ἢ
        "\xe1\xbc\xab" => 'HE', // Ἣ
        "\xe1\xbc\xac" => 'E', // Ἤ
        "\xe1\xbc\xad" => 'HE', // Ἥ
        "\xe1\xbc\xae" => 'E', // Ἦ
        "\xe1\xbc\xaf" => 'HE', // Ἧ
        "\xe1\xbc\xb0" => 'i', // ἰ
        "\xe1\xbc\xb1" => 'hi', // ἱ
        "\xe1\xbc\xb2" => 'i', // ἲ
        "\xe1\xbc\xb3" => 'hi', // ἳ
        "\xe1\xbc\xb4" => 'i', // ἴ
        "\xe1\xbc\xb5" => 'hi', // ἵ
        "\xe1\xbc\xb6" => 'i', // ἶ
        "\xe1\xbc\xb7" => 'hi', // ἷ
        "\xe1\xbc\xb8" => 'I', // Ἰ
        "\xe1\xbc\xb9" => 'HI', // Ἱ
        "\xe1\xbc\xba" => 'I', // Ἲ
        "\xe1\xbc\xbb" => 'HI', // Ἳ
        "\xe1\xbc\xbc" => 'I', // Ἴ
        "\xe1\xbc\xbd" => 'HI', // Ἵ
        "\xe1\xbc\xbe" => 'I', // Ἶ
        "\xe1\xbc\xbf" => 'HI', // Ἷ
        "\xe1\xbd\x80" => 'o', // ὀ
        "\xe1\xbd\x81" => 'ho', // ὁ
        "\xe1\xbd\x82" => 'o', // ὂ
        "\xe1\xbd\x83" => 'ho', // ὃ
        "\xe1\xbd\x84" => 'o', // ὄ
        "\xe1\xbd\x85" => 'ho', // ὅ
        "\xe1\xbd\x88" => 'O', // Ὀ
        "\xe1\xbd\x89" => 'HO', // Ὁ
        "\xe1\xbd\x8a" => 'O', // Ὂ
        "\xe1\xbd\x8b" => 'HO', // Ὃ
        "\xe1\xbd\x8c" => 'O', // Ὄ
        "\xe1\xbd\x8d" => 'HO', // Ὅ
        "\xe1\xbd\x90" => 'y', // ὐ
        "\xe1\xbd\x91" => 'hy', // ὑ
        "\xe1\xbd\x92" => 'y', // ὒ
        "\xe1\xbd\x93" => 'hy', // ὓ
        "\xe1\xbd\x94" => 'y', // ὔ
        "\xe1\xbd\x95" => 'hy', // ὕ
        "\xe1\xbd\x96" => 'y', // ὖ
        "\xe1\xbd\x97" => 'hy', // ὗ
        "\xe1\xbd\x99" => 'HY', // Ὑ
        "\xe1\xbd\x9b" => 'HY', // Ὓ
        "\xe1\xbd\x9d" => 'HY', // Ὕ
        "\xe1\xbd\x9f" => 'HY', // Ὗ
        "\xe1\xbd\xa0" => 'o', // ὠ
        "\xe1\xbd\xa1" => 'ho', // ὡ
        "\xe1\xbd\xa2" => 'o', // ὢ
        "\xe1\xbd\xa3" => 'ho', // ὣ
        "\xe1\xbd\xa4" => 'o', // ὤ
        "\xe1\xbd\xa5" => 'ho', // ὥ
        "\xe1\xbd\xa6" => 'o', // ὦ
        "\xe1\xbd\xa7" => 'ho', // ὧ
        "\xe1\xbd\xa8" => 'O', // Ὠ
        "\xe1\xbd\xa9" => 'HO', // Ὡ
        "\xe1\xbd\xaa" => 'O', // Ὢ
        "\xe1\xbd\xab" => 'HO', // Ὣ
        "\xe1\xbd\xac" => 'O', // Ὤ
        "\xe1\xbd\xad" => 'HO', // Ὥ
        "\xe1\xbd\xae" => 'O', // Ὦ
        "\xe1\xbd\xaf" => 'HO', // Ὧ
        "\xe1\xbd\xb0" => 'a', // ὰ
        "\xe1\xbd\xb1" => 'a', // ά
        "\xe1\xbd\xb2" => 'e', // ὲ
        "\xe1\xbd\xb3" => 'e', // έ
        "\xe1\xbd\xb4" => 'e', // ὴ
        "\xe1\xbd\xb5" => 'e', // ή
        "\xe1\xbd\xb6" => 'i', // ὶ
        "\xe1\xbd\xb7" => 'i', // ί
        "\xe1\xbd\xb8" => 'o', // ὸ
        "\xe1\xbd\xb9" => 'o', // ό
        "\xe1\xbd\xba" => 'y', // ὺ
        "\xe1\xbd\xbb" => 'y', // ύ
        "\xe1\xbd\xbc" => 'o', // ὼ
        "\xe1\xbd\xbd" => 'o', // ώ
        "\xe1\xbe\x80" => 'ai', // ᾀ
        "\xe1\xbe\x81" => 'hai', // ᾁ
        "\xe1\xbe\x82" => 'ai', // ᾂ
        "\xe1\xbe\x83" => 'hai', // ᾃ
        "\xe1\xbe\x84" => 'ai', // ᾄ
        "\xe1\xbe\x85" => 'hai', // ᾅ
        "\xe1\xbe\x86" => 'ai', // ᾆ
        "\xe1\xbe\x87" => 'hai', // ᾇ
        "\xe1\xbe\x88" => 'AI', // ᾈ
        "\xe1\xbe\x89" => 'HAI', // ᾉ
        "\xe1\xbe\x8a" => 'AI', // ᾊ
        "\xe1\xbe\x8b" => 'HAI', // ᾋ
        "\xe1\xbe\x8c" => 'AI', // ᾌ
        "\xe1\xbe\x8d" => 'HAI', // ᾍ
        "\xe1\xbe\x8e" => 'AI', // ᾎ
        "\xe1\xbe\x8f" => 'HAI', // ᾏ
        "\xe1\xbe\x90" => 'ei', // ᾐ
        "\xe1\xbe\x91" => 'hei', // ᾑ
        "\xe1\xbe\x92" => 'ei', // ᾒ
        "\xe1\xbe\x93" => 'hei', // ᾓ
        "\xe1\xbe\x94" => 'ei', // ᾔ
        "\xe1\xbe\x95" => 'hei', // ᾕ
        "\xe1\xbe\x96" => 'ei', // ᾖ
        "\xe1\xbe\x97" => 'hei', // ᾗ
        "\xe1\xbe\x98" => 'EI', // ᾘ
        "\xe1\xbe\x99" => 'HEI', // ᾙ
        "\xe1\xbe\x9a" => 'EI', // ᾚ
        "\xe1\xbe\x9b" => 'HEI', // ᾛ
        "\xe1\xbe\x9c" => 'EI', // ᾜ
        "\xe1\xbe\x9d" => 'HEI', // ᾝ
        "\xe1\xbe\x9e" => 'EI', // ᾞ
        "\xe1\xbe\x9f" => 'HEI', // ᾟ
        "\xe1\xbe\xa0" => 'oi', // ᾠ
        "\xe1\xbe\xa1" => 'hoi', // ᾡ
        "\xe1\xbe\xa2" => 'oi', // ᾢ
        "\xe1\xbe\xa3" => 'hoi', // ᾣ
        "\xe1\xbe\xa4" => 'oi', // ᾤ
        "\xe1\xbe\xa5" => 'hoi', // ᾥ
        "\xe1\xbe\xa6" => 'oi', // ᾦ
        "\xe1\xbe\xa7" => 'hoi', // ᾧ
        "\xe1\xbe\xa8" => 'OI', // ᾨ
        "\xe1\xbe\xa9" => 'HOI', // ᾩ
        "\xe1\xbe\xaa" => 'OI', // ᾪ
        "\xe1\xbe\xab" => 'HOI', // ᾫ
        "\xe1\xbe\xac" => 'OI', // ᾬ
        "\xe1\xbe\xad" => 'HOI', // ᾭ
        "\xe1\xbe\xae" => 'OI', // ᾮ
        "\xe1\xbe\xaf" => 'HOI', // ᾯ
        "\xe1\xbe\xb0" => 'a', // ᾰ
        "\xe1\xbe\xb1" => 'a', // ᾱ
        "\xe1\xbe\xb2" => 'ai', // ᾲ
        "\xe1\xbe\xb3" => 'ai', // ᾳ
        "\xe1\xbe\xb4" => 'ai', // ᾴ
        "\xe1\xbe\xb6" => 'a', // ᾶ
        "\xe1\xbe\xb7" => 'ai', // ᾷ
        "\xe1\xbe\xb8" => 'A', // Ᾰ
        "\xe1\xbe\xb9" => 'A', // Ᾱ
        "\xe1\xbe\xba" => 'A', // Ὰ
        "\xe1\xbe\xbb" => 'A', // Ά
        "\xe1\xbe\xbc" => 'AI', // ᾼ
        "\xe1\xbe\xbe" => 'i', // ι
        "\xe1\xbf\x81" => '¨̂', // ῁
        "\xe1\xbf\x82" => 'ei', // ῂ
        "\xe1\xbf\x83" => 'ei', // ῃ
        "\xe1\xbf\x84" => 'ei', // ῄ
        "\xe1\xbf\x86" => 'e', // ῆ
        "\xe1\xbf\x87" => 'ei', // ῇ
        "\xe1\xbf\x88" => 'E', // Ὲ
        "\xe1\xbf\x89" => 'E', // Έ
        "\xe1\xbf\x8a" => 'E', // Ὴ
        "\xe1\xbf\x8b" => 'E', // Ή
        "\xe1\xbf\x8c" => 'EI', // ῌ
        "\xe1\xbf\x8f" => '᾿̂', // ῏
        "\xe1\xbf\x90" => 'i', // ῐ
        "\xe1\xbf\x91" => 'i', // ῑ
        "\xe1\xbf\x92" => 'i', // ῒ
        "\xe1\xbf\x93" => 'i', // ΐ
        "\xe1\xbf\x96" => 'i', // ῖ
        "\xe1\xbf\x97" => 'i', // ῗ
        "\xe1\xbf\x98" => 'I', // Ῐ
        "\xe1\xbf\x99" => 'I', // Ῑ
        "\xe1\xbf\x9a" => 'I', // Ὶ
        "\xe1\xbf\x9b" => 'I', // Ί
        "\xe1\xbf\x9f" => '῾̂', // ῟
        "\xe1\xbf\xa0" => 'y', // ῠ
        "\xe1\xbf\xa1" => 'y', // ῡ
        "\xe1\xbf\xa2" => 'y', // ῢ
        "\xe1\xbf\xa3" => 'y', // ΰ
        "\xe1\xbf\xa4" => 'r', // ῤ
        "\xe1\xbf\xa5" => 'rh', // ῥ
        "\xe1\xbf\xa6" => 'y', // ῦ
        "\xe1\xbf\xa7" => 'y', // ῧ
        "\xe1\xbf\xa8" => 'Y', // Ῠ
        "\xe1\xbf\xa9" => 'Y', // Ῡ
        "\xe1\xbf\xaa" => 'Y', // Ὺ
        "\xe1\xbf\xab" => 'Y', // Ύ
        "\xe1\xbf\xac" => 'RH', // Ῥ
        "\xe1\xbf\xb2" => 'oi', // ῲ
        "\xe1\xbf\xb3" => 'oi', // ῳ
        "\xe1\xbf\xb4" => 'oi', // ῴ
        "\xe1\xbf\xb6" => 'o', // ῶ
        "\xe1\xbf\xb7" => 'oi', // ῷ
        "\xe1\xbf\xb8" => 'O', // Ὸ
        "\xe1\xbf\xb9" => 'O', // Ό
        "\xe1\xbf\xba" => 'O', // Ὼ
        "\xe1\xbf\xbb" => 'O', // Ώ
        "\xe1\xbf\xbc" => 'OI', // ῼ
        "\xe2\x80\x81" => ' ', //  
        "\xe2\x80\x82" => ' ', //  
        "\xe2\x80\x83" => ' ', //  
        "\xe2\x80\x84" => ' ', //  
        "\xe2\x80\x85" => ' ', //  
        "\xe2\x80\x86" => ' ', //  
        "\xe2\x80\x87" => ' ', //  
        "\xe2\x80\x88" => ' ', //  
        "\xe2\x80\x89" => ' ', //  
        "\xe2\x80\x8a" => ' ', //  
        "\xe2\x80\x90" => '-', // ‐
        "\xe2\x80\x91" => '-', // ‑
        "\xe2\x80\x92" => '-', // ‒
        "\xe2\x80\x95" => '-', // ―
        "\xe2\x80\x96" => '||', // ‖
        "\xe2\x80\x9b" => "'", // ‛
        "\xe2\x80\x9f" => '"', // ‟
        "\xe2\x80\xa4" => '.', // ․
        "\xe2\x80\xa5" => '..', // ‥
        "\xe2\x80\xb2" => "'", // ′
        "\xe2\x80\xb3" => '"', // ″
        "\xe2\x80\xbc" => '!!', // ‼
        "\xe2\x81\x84" => '/', // ⁄
        "\xe2\x81\x85" => '[', // ⁅
        "\xe2\x81\x86" => ']', // ⁆
        "\xe2\x81\x87" => '??', // ⁇
        "\xe2\x81\x88" => '?!', // ⁈
        "\xe2\x81\x89" => '!?', // ⁉
        "\xe2\x81\x8e" => '*', // ⁎
        "\xe2\x81\x9f" => ' ', //  
    ];

    /**
     * @var string[]
     */
    private static $uriBlacklist = [
        '"' => '',
        '*' => '',
        ':' => '_',
        '<' => '',
        '>' => '',
        '?' => '',
        '\\' => '',
        '|' => '',
        '~' => '',
        "\00" => '',
        "\01" => '',
        "\02" => '',
        "\03" => '',
        "\04" => '',
        "\05" => '',
        "\06" => '',
        "\07" => '',
        "\10" => '',
        "\11" => '',
        "\12" => '',
        "\13" => '',
        "\14" => '',
        "\15" => '',
        "\16" => '',
        "\17" => '',
        "\20" => '',
        "\21" => '',
        "\22" => '',
        "\23" => '',
        "\24" => '',
        "\25" => '',
        "\26" => '',
        "\27" => '',
        "\30" => '',
        "\31" => '',
    ];

    /**
     * Donne une instance de Transliterator sans suppression des accents
     * @return Transliterator
     */
    public static function getSoftTransliterator(): Transliterator
    {
        if (empty(self::$softTransliterator)) {
            self::$softTransliterator = Transliterator::create(
                implode(
                    '; ',
                    [
                        'Any-Latin', // converti tous les alphabets en latin
                        'NFC', // réassemble les caractères composés restants
                        'UTF-8', // converti en UTF8
                    ]
                )
            );
        }
        return self::$softTransliterator;
    }

    /**
     * ascii + mb_strtolower
     * @param string $str
     * @param bool   $keepSize
     * @return string
     */
    public static function asciiToLower(string $str, bool $keepSize = false): string
    {
        return mb_strtolower(self::ascii($str, $keepSize));
    }

    /**
     * Retourne la chaine dépourvue de chars != ascii avec transliteration
     * @param string $str
     * @param bool   $keepSize
     * @return string
     */
    public static function ascii(string $str, bool $keepSize = false): string
    {
        $len = mb_strlen($str);
        $stripped = self::getTransliterator()->transliterate($str);
        $replaced = str_replace(array_keys(self::$chars), array_values(self::$chars), $stripped);
        if ($keepSize && mb_strlen($replaced) > $len) {
            $replaced = '';
            for ($i = 0; $i < $len; $i++) {
                $replaced .= mb_substr(self::ascii(mb_substr($str, $i, 1)), 0, 1);
            }
        }
        return self::iconv($replaced);
    }

    /**
     * Donne l'instance de Transliterator
     * @return Transliterator
     */
    public static function getTransliterator(): Transliterator
    {
        if (empty(self::$transliterator)) {
            self::$transliterator = Transliterator::create(
                implode(
                    '; ',
                    [
                        'Any-Latin', // converti tous les alphabets en latin
                        'NFD', // sépare les chars composés ex: "è" -> "e`"
                        '[:Nonspacing Mark:] Remove', // retire les accents seuls "e`" -> "e"
                        'NFC', // réassemble les caractères composés restants
                    ]
                )
            );
        }
        return self::$transliterator;
    }

    /**
     * Appel iconv
     * @param string $str
     * @param string $in
     * @param string $out
     * @return string
     */
    public static function iconv(string $str, string $in = 'UTF-8', string $out = 'ASCII//TRANSLIT//IGNORE'): string
    {
        return (string)iconv($in, $out, $str);
    }

    /**
     * Donne un URI valide
     * @param string $uri
     * @return string
     */
    public static function safeUri(string $uri): string
    {
        $fnames = explode('/', $uri);
        $newFnames = [];
        $keys = array_keys(self::$uriBlacklist);
        $values = array_values(self::$uriBlacklist);
        foreach ($fnames as $fname) {
            $newFnames[] = str_replace(
                $keys,
                $values,
                ltrim($fname, '-')
            );
        }
        return implode('/', $newFnames);
    }

    /**
     * Equivalent de AsalaeGlobal.safeUrlEncode() en php
     * (contourne la protection d'apache sur le / ou le ?)
     * @param string $str
     * @param string $escapeChar un caractère rare mais autorisé dans l'url
     * @return string
     */
    public static function safeUrlEncode(string $str, string $escapeChar = '~'): string
    {
        $encodedStr = urlencode($str);
        $escapeCharCode = strtoupper(dechex(ord($escapeChar)));
        $encodedStr = str_replace($escapeChar, '%' . $escapeCharCode, $encodedStr);
        return str_replace('%', $escapeChar, $encodedStr);
    }

    /**
     * Equivalent de AsalaeGlobal.safeUrlDecode() en php
     * @param string $str
     * @param string $escapeChar un caractère rare mais autorisé dans l'url
     * @return string
     */
    public static function safeUrlDecode(string $str, string $escapeChar = '~'): string
    {
        $encodedStr = str_replace($escapeChar, '%', $str);
        return urldecode($encodedStr);
    }
}
