<?php

/**
 * AsalaeCore\View\Helper\PaginatorHelper
 */

namespace AsalaeCore\View\Helper;

use Cake\I18n\I18n;
use Cake\Utility\Inflector;
use Cake\View\Helper\NumberHelper;
use Cake\View\Helper\PaginatorHelper as CakePaginatorHelper;
use Cake\View\Helper\UrlHelper;
use Cake\View\View;

use function Cake\Core\h;
use function Cake\I18n\__;

/**
 * Permet de générer le bloc de pagination
 *
 * @category Helper
 *
 * @author    Libriciel SCOP <contact@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   https://www.gnu.org/licenses/agpl-3.0.txt
 * @property UrlHelper                    $Url
 * @property NumberHelper                 $Number
 * @property \Cake\View\Helper\HtmlHelper $Html
 */
class PaginatorHelper extends CakePaginatorHelper
{
    /**
     * Default config for this class
     *
     * Options: Holds the default options for pagination links
     *
     * The values that may be specified are:
     *
     * - `url` Url of the action. See Router::url()
     * - `url['sort']` the key that the recordset is sorted.
     * - `url['direction']` Direction of the sorting (default: 'asc').
     * - `url['page']` Page number to use in links.
     * - `model` The name of the model.
     * - `escape` Defines if the title field for the link should be escaped (default: true).
     *
     * Templates: the templates used by this class
     *
     * @var array
     */
    protected array $_defaultConfig = [
        'params' => [],
        'options' => [],
        'numbers' => [
            'first' => 2,
            'last' => 2,
            'modulus' => 4,
        ],
    ];

    /**
     * Constructor. Overridden to merge passed args with URL options.
     *
     * @param \Cake\View\View $View   The View this helper is being attached to.
     * @param array           $config Configuration settings for the helper.
     */
    public function __construct(View $View, array $config = [])
    {
        $this->_defaultConfig['templates'] = [
            'nextActive' => '<li class="next"><a rel="next" href="{{url}}"'
                . ' data-mode="section-ajax" title="' . __('Next') . '">{{text}}</a></li>',
            'nextDisabled' => '<li class="next disabled"><a href="#" onclick="return false;">{{text}}</a></li>',
            'prevActive' => '<li class="prev"><a rel="prev" href="{{url}}"'
                . ' data-mode="section-ajax" title="' . __('Previous') . '">{{text}}</a></li>',
            'prevDisabled' => '<li class="prev disabled"><a href="#" onclick="return false;">{{text}}</a></li>',
            'counterRange' => '<i class="start">{{start}}</i> - {{end}} of {{count}}',
            'counterPages' => '{{page}} of {{pages}}',
            'first' => '<li class="first"><a href="{{url}}" data-mode="section-ajax"'
                . ' title="' . __('Page: {0}', '{{text}}') . '">{{text}}</a></li>',
            'last' => '<li class="last"><a href="{{url}}" data-mode="section-ajax"'
                . ' title="' . __('Page: {0}', '{{text}}') . '">{{text}}</a></li>',
            'number' => '<li><a href="{{url}}" data-mode="section-ajax" '
                . 'title="' . __('Page: {0}', '{{text}}') . '">{{text}}</a></li>',
            'current' => '<li class="active"><a href="#" title="' . __('Page: {0}', '{{text}}') . '">{{text}}</a></li>',
            'ellipsis' => '<li class="ellipsis"><button type="button" class="btn-link">&hellip;</button></li>',
            'sort' => '<a class="sortable" href="{{url}}" title="{{title}}" data-mode="section-ajax">{{text}}</a>',
            'sortAsc' => '<a class="sort-asc" href="{{url}}" title="{{title}}" data-mode="section-ajax">{{text}}</a>',
            'sortDesc' => '<a class="sort-desc" href="{{url}}" title="{{title}}" data-mode="section-ajax">{{text}}</a>',
            'sortAscLocked' => '<a class="sort-asc locked" href="{{url}}" '
                . 'title="{{title}}" data-mode="section-ajax">{{text}}</a>',
            'sortDescLocked' => '<a class="sort-desc locked" href="{{url}}" '
                . 'title="{{title}}" data-mode="section-ajax">{{text}}</a>',
        ];
        parent::__construct($View, $config);
        // Dans certains cas, le sort n'apparait pas dans paginated()->pagingParams()
        $request = $this->_View->getRequest();
        $params = $this->getConfig('params', []) + array_filter(
            [
                'sort' => $request->getQuery('sort'),
                'direction' => $request->getQuery('direction'),
            ]
        );
        $this->setConfig('params', $params);
    }

    /**
     * Permet de générer le bloc de pagination
     *
     * @param string $id      'select-'.$id pour le select et 'btn-'.$id pour le bouton
     * @param array  $options options du select
     * @return string
     */
    public function blocPagination(string $id = 'default-bloc-pagination', array $options = []): string
    {
        $url = $options['url'] ?? [];
        unset($options['url']);
        unset($options['model']);
        if (empty($options)) {
            $first = '';
            $second = '';
        } else {
            $first = $this->Html->tag('div', null, ['class' => 'form-group form-group-sm'])
                . $this->Html->tag(
                    'label',
                    __("Action sur les cases à cocher"),
                    ['class' => 'sr-only', 'for' => 'select-' . $id]
                )
                . $this->Html->tag('select', null, ['class' => 'form-control w-sm inline', 'id' => 'select-' . $id]);
            foreach ($options['actions'] ?? [] as $value => $message) {
                $first .= $this->Html->tag('option', $message, compact('value'));
            }
            $first .= $this->Html->tag('/select') . $this->Html->tag('/div');
            $second = $this->Html->tag(
                'button',
                '<i class="fa fa-cog" aria-hidden="true"></i> ' . __("Executer"),
                [
                    'type' => 'button',
                    'class' => 'btn btn-primary',
                    'id' => 'btn-' . $id,
                ]
            );
        }
        $third = $this->Html->tag(
            'small',
            $this->counter(
                __(
                    'Affichage des résultats de {0} à {1} sur les<br>{2} résultats',
                    '<i class="start">{{start}}</i>',
                    '<i class="end">{{end}}</i>',
                    '<i class="count">{{count}}</i>'
                ),
            ),
            ['class' => 'text-muted pagination-counters']
        );
        $fourth = $this->Html->tag(
            'ul',
            $this->prev(
                '<i class="fa fa-chevron-left" aria-hidden="true"></i><span class="sr-only">'
                . __("Previous")
                . '</span>',
                ['escape' => false, 'url' => $url]
            )
            . $this->numbers(['url' => $url] + $this->getConfig('numbers'))
            . $this->next(
                '<i class="fa fa-chevron-right" aria-hidden="true"></i><span class="sr-only">'
                . __("Next") . '</span>',
                ['escape' => false, 'url' => $url]
            ),
            ['class' => 'pagination pagination-sm m-0']
        );

        return $this->Html->tag('div', $first, ['class' => 'col-md-2', 'id' => $id . '-bloc1'])
            . $this->Html->tag('div', $second, ['class' => 'col-md-2', 'id' => $id . '-bloc2'])
            . $this->Html->tag('div', $third, ['class' => 'col-md-3 text-left', 'id' => $id . '-bloc3'])
            . $this->Html->tag('div', $fourth, ['class' => 'col-md-5 text-right', 'id' => $id . '-bloc4']);
    }

    /**
     * Generates a sorting link. Sets named parameters for the sort and direction. Handles
     * direction switching automatically.
     *
     * ### Options:
     *
     * - `escape` Whether you want the contents html entity encoded, defaults to true.
     * - `model` The model to use, defaults to PaginatorHelper::defaultModel().
     * - `direction` The default direction to use when this link isn't active.
     * - `lock` Lock direction. Will only use the default direction then, defaults to false.
     *
     * @param string            $key     The name of the key that the recordset should be sorted.
     * @param string|array|null $title   Title for the link. If $title is null $key will be used
     *                                   for the title and will be generated by inflection. It
     *                                   can also be an array with keys `asc` and `desc` for
     *                                   specifying separate titles based on the direction.
     * @param array             $options Options for sorting link. See above for list of keys.
     * @return string A link sorting default by 'asc'. If the resultset is sorted 'asc' by the specified
     *  key the returned link will sort by 'desc'.
     * @link https://book.cakephp.org/4/en/views/helpers/paginator.html#creating-sort-links
     */
    public function sort(string $key, $title = null, array $options = []): string
    {
        $options += ['url' => [], 'escape' => true];
        $url = $options['url'];
        unset($options['url']);

        if (!$title) {
            $title = $key;

            if (str_contains($title, '.')) {
                $title = str_replace('.', ' ', $title);
            }

            $title = I18n::getTranslator()->translate(
                Inflector::humanize(
                    preg_replace('/_id$/', '', $title)
                )
            );
        }
        $defaultDir = isset($options['direction']) ? strtolower($options['direction']) : 'asc';
        unset($options['direction']);

        $locked = $options['lock'] ?? false;
        unset($options['lock']);

        $sortKey = (string)$this->param('sort');
        $alias = $this->param('alias');
        [$table, $field] = explode('.', $key . '.');
        if (!$field) {
            $field = $table;
            $table = $alias;
        }
        $isSorted = (
            $sortKey === $table . '.' . $field ||
            $sortKey === $alias . '.' . $key ||
            $table . '.' . $field === $alias . '.' . $sortKey
        );

        $template = 'sort';
        $dir = $defaultDir;
        if ($isSorted) {
            if ($locked) {
                $template = $dir === 'asc' ? 'sortDescLocked' : 'sortAscLocked';
            } else {
                $dir = $this->sortDir() === 'asc' ? 'desc' : 'asc';
                $template = $dir === 'asc' ? 'sortDesc' : 'sortAsc';
            }
        }
        if (is_array($title) && array_key_exists($dir, $title)) {
            $title = $title[$dir];
        }

        $paging = ['sort' => $key, 'direction' => $dir, 'page' => 1];

        $vars = [
            'text' => $options['escape'] ? h($title) : $title,
            'url' => $this->generateUrl($paging, $url),
        ];

        // Partie surchargée
        foreach (['asc', 'desc'] as $oDir) {
            if (!empty($options[$oDir])) {
                $o = $options[$oDir];
                unset($options[$oDir]);
                if ($dir === $oDir) {
                    $options = array_merge($options, $o);
                }
            }
        }
        $vars = array_merge($vars, $options);
        // FIN

        return $this->templater()->format($template, $vars);
    }

    /**
     * Gets the current direction the recordset is sorted
     *
     * @return string The direction by which the recordset is being sorted, or
     *  null if the results are not currently sorted.
     */
    protected function sortDir(): string
    {
        $dir = parent::sortDir();
        if ($dir === 'asc' && $this->param('direction') === 'desc') {
            $dir = 'desc';
        }
        return $dir;
    }
}
