<?php

namespace AsalaeCore\Test\TestCase\Command;

use AsalaeCore\Command\CronCommand;
use AsalaeCore\Cron\CronInterface;
use AsalaeCore\Cron\LiberSign;
use AsalaeCore\Driver\Volume\VolumeFilesystem;
use AsalaeCore\Factory\Utility;
use AsalaeCore\Model\Table\CronExecutionsTable;
use AsalaeCore\Test\Mock\FakeCron;
use AsalaeCore\TestSuite\InvokePrivateTrait;
use AsalaeCore\TestSuite\TestCase;
use AsalaeCore\Utility\Notify;
use Cake\Console\ConsoleIo;
use Cake\Console\TestSuite\ConsoleIntegrationTestTrait;
use Cake\Core\Configure;
use Cake\Datasource\EntityInterface;
use Cake\Mailer\Mailer;
use DateTime;
use Libriciel\Filesystem\Utility\Filesystem;
use PHPUnit\Framework\MockObject\Exception;
use ReflectionClass;
use ReflectionException;
use ZMQSocket;

class CronCommandTest extends TestCase
{
    use ConsoleIntegrationTestTrait;
    use InvokePrivateTrait;

    public array $fixtures = [
        'app.Crons',
        'app.CronExecutions',
    ];

    /**
     * setup
     */
    public function setUp(): void
    {
        parent::setUp();
        if (is_dir(TMP_TESTDIR)) {
            Filesystem::remove(TMP_TESTDIR);
        }
        Filesystem::mkdir(TMP_TESTDIR);
        $socket = $this->createMock(ZMQSocket::class);
        $socket->method('send');
        $Notify = Notify::getInstance();
        $this->invokeProperty($Notify, 'socket', 'set', $socket);
        Configure::write('debug_mails', true);
        Configure::write('App.paths.administrators_json', TMP_TESTDIR . DS . 'admins.json');
        $adminData = [
            [
                'username' => 'admin',
                'password' => '*******',
                'name' => 'admin',
                'email' => 'admin@test.invalid',
                'notify' => '1',
                'cookies' => []
            ],
        ];
        file_put_contents(TMP_TESTDIR . DS . 'admins.json', json_encode($adminData));
        $email = $this->getMockBuilder(Mailer::class)
            ->onlyMethods(['send'])
            ->getMock();
        $email->method('send')->willReturn(['email content']);
        Utility::set(Mailer::class, $email);
    }

    /**
     * tearDown
     */
    public function tearDown(): void
    {
        parent::tearDown();
        if (is_dir(TMP_TESTDIR)) {
            Filesystem::remove(TMP_TESTDIR);
        }
    }

    /**
     * test run
     */
    public function testrun(): void
    {
        $uniqid = uniqid('testrun');
        $Crons = $this->fetchTable('Crons');
        $cron = $Crons->newEntity(
            [
                'name' => "test cron",
                'description' => "test cron",
                'classname' => FakeCron::class,
                'active' => true,
                'locked' => false,
                'frequency' => 'one_shot',
                'next' => (new DateTime())->format(DATE_RFC3339),
                'app_meta' => json_encode(['test' => $uniqid]),
                'parallelisable' => false,
            ]
        );
        $Crons->saveOrFail($cron);
        $this->exec('cron run ' . $cron->id);
        $this->assertOutputContains('success');

        // one_shot supprimé
        $this->assertCount(0, $Crons->find()->where(['id' => $cron->id]));

        // log écrit
        $this->assertFileExists($cron->get('log_filename'));
        $lastLine = exec('tail -n1 ' . escapeshellarg($cron->get('log_filename')));
        $this->assertStringEndsWith($uniqid, $lastLine);
    }

    /**
     * @return void
     */
    public function testGetOptionParser(): void
    {
        $this->exec('cron -h');
        $this->assertOutputContains('--no-interaction');
    }

    /**
     * @return void
     */
    public function testExecute(): void
    {
        // phpcs:disable
        $testClass = new class () implements CronInterface {
            public static string $willReturn = CronExecutionsTable::S_SUCCESS;
            public static bool $worked = false;
            public function __construct(array $params = [], $out = null, $err = null) { }
            public static function getVirtualFields(): array { return []; }
            public function work(EntityInterface $exec = null, EntityInterface $cron = null): string
            {
                self::$worked = true;
                return self::$willReturn;
            }
            public function getOutput(): string { return ''; }
            public function err($message = null, int $newlines = 1) { return 0; }
        };
        // phpcs:enable
        $Crons = $this->fetchTable('Crons');
        $cronData = [
            'name' => 'test cron',
            'description' => 'test cron',
            'classname' => $testClass::class,
            'active' => true,
            'locked' => false,
            'frequency' => 'one_shot',
            'next' => new DateTime('2000-01-01 00:00:00'),
            'app_meta' => '{}',
            'parallelisable' => 1,
            'max_duration' => 'PT1H',
            'last_email' => null,
        ];
        $cron = $Crons->newEntity($cronData);
        $Crons->saveOrFail($cron);

        $this->exec('cron');
        $this->assertTrue($testClass::$worked);
        $this->assertEquals(0, $Crons->find()->where(['classname' => $testClass::class])->count());

        $cron = $Crons->newEntity($cronData);
        $Crons->saveOrFail($cron);

        $testClass::$willReturn = CronExecutionsTable::S_ERROR;
        $this->exec('cron');
        $this->assertEquals(1, $Crons->find()->where(['classname' => $testClass::class])->count());
    }

    /**
     * @return void
     */
    public function testshutdown(): void
    {
        $CronExecutions = $this->fetchTable('CronExecutions');
        $cronData = [
            'name' => 'test cron',
            'description' => 'test cron',
            'classname' => FakeCron::class,
            'active' => true,
            'locked' => true,
            'frequency' => 'one_shot',
            'next' => new DateTime('2000-01-01 00:00:00'),
            'app_meta' => '{}',
            'parallelisable' => 1,
            'max_duration' => 'PT1H',
            'last_email' => null,
        ];
        $Crons = $this->fetchTable('Crons');
        $cron = $Crons->newEntity($cronData);
        $Crons->saveOrFail($cron);
        $class = new CronCommand();
        $class->cron = $cron;
        $class->shutdown();
        $this->assertEquals(0, $Crons->find()->where(['id' => $cron->id])->count());

        $cronData['frequency'] = 'P1D';
        $cron = $Crons->newEntity($cronData);
        $Crons->saveOrFail($cron);
        $class = new CronCommand();
        $class->cron = $cron;
        $class->exec = $CronExecutions->get(1);
        $prevNext = $cron->get('next')->format('Y-m-d H:i:s');
        $class->shutdown();
        $updatedEntity = $Crons->find()->where(['id' => $cron->id]);
        $this->assertEquals(1, $updatedEntity->count());
        $this->assertNotEquals($prevNext, $cron->get('next')->format('Y-m-d H:i:s'));
    }

    /**
     * @return void
     * @throws Exception
     * @throws ReflectionException
     * @see CronCommand::chooseCron()
     */
    public function testchooseCron(): void
    {
        $class = new CronCommand();
        $class->io = $this->createMock(ConsoleIo::class);
        $class->io->method('ask')->willReturn(LiberSign::class);
        $class->io->method('askChoice')->willReturn('y');
        $reflection = new ReflectionClass($class);
        $method = $reflection->getMethod('chooseCron');
        $entity = $method->invoke($class, 'some_random_string');
        $this->assertEquals(LiberSign::class, $entity->get('classname'));
    }
}
