<?php

namespace AsalaeCore\Test\TestCase\Controller\Component;

use AsalaeCore\Controller\Component\IndexComponent;
use AsalaeCore\Factory\Utility;
use AsalaeCore\Model\Table\ArosTable;
use AsalaeCore\Model\Table\FiltersTable;
use AsalaeCore\Model\Table\SavedFiltersTable;
use AsalaeCore\TestSuite\TestCase;
use Cake\Core\Configure;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Exception;
use PHPUnit\Framework\MockObject\Exception as MockException;
use ReflectionClass;
use ReflectionException;

class IndexComponentTest extends TestCase
{
    use ComponentTestTrait;

    /**
     * @var string|IndexComponent
     */
    public $componentClassname = IndexComponent::class;

    /**
     * @var IndexComponent $component
     */
    public $component = null;

    public array $fixtures = [
        'app.Acos',
        'app.Agreements',
        'app.Aros',
        'app.Crons',
        'app.Filters',
        'app.SavedFilters',
        'app.Users',
    ];

    /**
     * setup
     */
    public function setUp(): void
    {
        parent::setUp();
        Utility::set('Notify', 'AsalaeCore\Test\Mock\FakeClass');
        Configure::write('App.defaultLocale', 'fr_FR');
    }

    /**
     * test Init
     */
    public function testInit(): void
    {
        $this->setComponent(
            [
                'environment' => ['HTTP_X_REQUESTED_WITH' => 'XMLHttpRequest'],
                'session' => ['Auth' => ['id' => 1]],
            ]
        );
        $reflection = new ReflectionClass($this->controller);
        $property = $reflection->getProperty('paginate');
        $initialPaginate = $property->getValue($this->controller);
        $initialPaginate['limit'] = null;
        $property->setValue($this->controller, $initialPaginate);
        $this->component->initialize();
        $reflection = new ReflectionClass($this->controller);
        $property = $reflection->getProperty('paginate');
        $paginate = $property->getValue($this->controller);
        $this->assertNotEquals($initialPaginate, $paginate);
    }

    /**
     * test Filter
     */
    public function testFilter(): void
    {
        $this->setComponent(['query' => ['id' => 1, 'sort' => 'favorite']]);
        $loc = TableRegistry::getTableLocator();
        $Crons = $loc->get('Crons');
        $query = $Crons->find();
        $this->component->setQuery($query)
            ->filter('id');
        $sql = $query->sql();
        $this->assertTextContains('id =', $sql);

        $query = $Crons->find();
        $this->setComponent(
            [
                'query' => [
                    'dateoperator_next' => ['<'],
                    'next' => [
                        '30/01/2020',
                    ],
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('next', IndexComponent::FILTER_DATEOPERATOR, 'next');
        $sql = $query->sql();
        $this->assertTextContains('date(next) <', $sql);

        $query = $Crons->find();
        $this->setComponent(
            [
                'query' => ['favorite' => 'true'],
                'controller' => 'mock-controlle',
                'cookies' => [
                    'table-favorite-mock-controlle-test-table-123' => '123',
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('favorite', IndexComponent::FILTER_FAVORITES);
        $sql = $query->sql();
        $this->assertTextContains('id IN', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['id' => ['%1%']]]);
        $this->component->setQuery($query)
            ->filter('id', IndexComponent::FILTER_ILIKE);
        $sql = $query->sql();
        $this->assertTextContains('LIKE', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['id' => [1, 2, 3]]]);
        $this->component->setQuery($query)
            ->filter('id', IndexComponent::FILTER_IN);
        $sql = $query->sql();
        $this->assertTextContains('id IN', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['locked' => ['true']]]);
        $this->component->setQuery($query)
            ->filter('locked', IndexComponent::FILTER_IS, 'locked');
        $sql = $query->sql();
        $this->assertTextContains('locked', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['locked' => ['true']]]);
        $this->component->setQuery($query)
            ->filter('locked', IndexComponent::FILTER_IS_NOT, 'locked');
        $sql = $query->sql();
        $this->assertTextContains('locked', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['id' => [1, 2, 3]]]);
        $this->component->setQuery($query)
            ->filter('id', IndexComponent::FILTER_NOT_IN);
        $sql = $query->sql();
        $this->assertTextContains('id NOT IN', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['id' => [1]]]);
        $this->component->setQuery($query)
            ->filter('id', IndexComponent::FILTER_LESS_OR_EQUAL);
        $sql = $query->sql();
        $this->assertTextContains('id <=', $sql);

        $query = $Crons->find();
        $this->setComponent(['query' => ['id' => [1]]]);
        $this->component->setQuery($query)
            ->filter(
                'id',
                function ($id) {
                    return ['id = ' . ($id * 10)];
                }
            );
        $sql = $query->sql();
        $this->assertTextContains('id = 10', $sql);

        $e = null;
        try {
            $this->setComponent(['query' => ['id' => [1]]]);
            $this->component->filter('id');
        } catch (Exception $e) {
        }
        $this->assertInstanceOf(Exception::class, $e);

        $query = $Crons->find();
        $this->setComponent(
            [
                'query' => [
                    'next' => [
                        '<',
                    ],
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('next', IndexComponent::FILTER_DATECOMPARENOW, 'next');
        $sql = $query->sql();
        $this->assertTextContains('date(next) <', $sql);

        $Agreements = $loc->get('Agreements');
        $query = $Agreements->find();
        $this->setComponent(
            [
                'query' => [
                    'max_size_per_transfer' => [
                        [
                            'operator' => '>',
                            'mult' => 1024 * 1024,
                            'value' => 1000,
                        ],
                    ],
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('max_size_per_transfer', IndexComponent::FILTER_SIZEOPERATOR);
        $sql = $query->sql();
        $this->assertTextContains('max_size_per_transfer >', $sql);

        $query = $Agreements->find();
        $this->setComponent(
            [
                'query' => [
                    'max_size_per_transfer' => [
                        [
                            'operator' => '>',
                            'value' => 1000,
                        ],
                    ],
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('max_size_per_transfer', IndexComponent::FILTER_COUNTOPERATOR);
        $sql = $query->sql();
        $this->assertTextContains('max_size_per_transfer >', $sql);

        $query = $Agreements->find();
        $this->setComponent(
            [
                'query' => [
                    'identifier' => [
                        [
                            'agr-test01',
                        ],
                    ],
                ],
            ]
        );
        $this->component->setQuery($query)
            ->filter('identifier');
        $sql = $query->sql();
        $this->assertTextContains('identifier =', $sql);
    }

    /**
     * @return void
     */
    public function testExtractModels(): void
    {
        /** @var FiltersTable $Filters */
        $Filters = $this->fetchTable('Filters');
        $query = $Filters->find();
        $expected = ['Filters'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // belongsTo
        $query = $Filters->find()
            ->innerJoinWith('SavedFilters');
        $expected = ['Filters', 'SavedFilters'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // hasMany
        /** @var SavedFiltersTable $Filters */
        $SavedFilters = $this->fetchTable('SavedFilters');
        $query = $SavedFilters->find()
            ->innerJoinWith('Filters');
        $expected = ['SavedFilters', 'Filters'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // belongsToMany
        /** @var ArosTable $Filters */
        $Aros = $this->fetchTable('Aros');
        $query = $Aros->find()
            ->innerJoinWith('Acos');
        $expected = ['Aros', 'Acos'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // nested
        $query = $Filters->find()
            ->innerJoinWith('SavedFilters.Users');
        $expected = ['Filters', 'SavedFilters', 'Users'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // with subquery
        $subquery = $Aros->find()->select(['Aros.id'])->limit(1);
        $query = $Filters->find()
            ->where(['Filters.id IN' => $subquery]);
        $expected = ['Filters'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));

        // nested in contain
        $query = $Filters->find()
            ->contain(['SavedFilters' => ['Users']]);
        $expected = ['Filters', 'SavedFilters', 'Users'];
        $this->assertEquals($expected, IndexComponent::extractModels($query));
    }

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testExportCsv(): void
    {
        $this->setComponent(
            [
                'environment' => ['HTTP_X_REQUESTED_WITH' => 'XMLHttpRequest'],
                'session' => ['Auth' => ['id' => 1]],
            ]
        );
        $this->component->exportCsv();
        $url = Hash::get($this->controller->viewBuilder()->getVars(), 'downloadCsv.href');
        $this->assertStringContainsString('export_csv=true', $url);

        $this->setComponent(['query' => ['export_csv' => 'true']]);
        $response = $this->component->exportCsv();
        $this->assertNotEmpty($response);
        ob_start();
        ob_start(); // étrange, mais il est nécéssaire de doubler
        $response->getBody()->getContents(); // provoque le readfile du template
        $content = ob_get_clean() . ob_get_clean();
        $this->assertStringContainsString('foo,bar', $content);
    }
}
