<?php

namespace AsalaeCore\Test\TestCase\Form;

use AsalaeCore\Form\AbstractGenericMessageForm;
use AsalaeCore\Form\Seda21Form;
use AsalaeCore\TestSuite\TestCase;
use AsalaeCore\Utility\DOMUtility;
use Cake\Cache\Cache;
use Cake\Core\Configure;
use Cake\ORM\Entity;
use Cake\ORM\TableRegistry;
use DOMDocument;
use DOMElement;
use DOMXPath;
use Exception;
use Libriciel\Filesystem\Utility\Filesystem;

class Seda21FormTest extends TestCase
{
    public const string SEDA21_TEST_FILE = TESTS . 'Data' . DS . 'sample_seda21.xml';

    public string $testDir = TMP_TESTDIR . DS . 'seda_21_form_test';

    public array $fixtures = [
        'app.Agreements',
        'app.KeywordLists',
        'app.Keywords',
        'app.OrgEntities',
        'app.Profiles',
        'app.ServiceLevels',
        'app.TransferAttachments',
        'app.Transfers',
        'app.TypeEntities',
    ];

    /**
     * setup
     */
    public function setUp(): void
    {
        parent::setUp();
        Cache::disable();
        Filesystem::reset();
        if (is_dir($this->testDir)) {
            Filesystem::remove($this->testDir);
        }
        Filesystem::mkdir($this->testDir);
    }

    /**
     * tearDown
     */
    public function tearDown(): void
    {
        parent::tearDown();
        if (is_dir($this->testDir)) {
            Filesystem::remove($this->testDir);
        }
        Configure::write('App.paths.data', TMP_TESTDIR);
    }

    /**
     * test AutoRepair
     */
    public function testAutoRepair(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $this->assertFileExists($file);

        $TransferAttachments = TableRegistry::getTableLocator()->get('TransferAttachments');
        $TransferAttachments->deleteAll([]);

        /**
         * On créé un xml avec des erreurs réparable automatiquement
         */
        $util = DOMUtility::load($file);
        /** @var DOMElement $BinaryDataObject */
        foreach ($util->xpath->query('//ns:BinaryDataObject') as $BinaryDataObject) {
            $Filename = $BinaryDataObject->getElementsByTagName('Filename')->item(0);
            $TransferAttachments->save(
                new Entity(
                    [
                        'transfer_id' => 1,
                        'filename' => $Filename->nodeValue,
                        'size' => 1,
                        'hash' => 'foo',
                        'hash_algo' => 'bar',
                    ]
                )
            );
            $Filename->nodeValue = '/' . $Filename->nodeValue;
            $util->dom->save($file);
        }
        // Integrity manquant
        /** @var DOMElement $integrity */
        foreach ($util->xpath->query('//ns:MessageDigest') as $integrity) {
            $integrity->parentNode->removeChild($integrity);
        }
        $util->dom->save($file);
        $util = DOMUtility::load($file);
        $filename = $util->xpath->query('//ns:BinaryDataObject/ns:FileInfo/ns:Filename')->item(0)->nodeValue;
        $this->assertEquals('/sample.pdf', $filename);
        $this->assertEquals(0, $util->xpath->query('//ns:MessageDigest')->count());

        /**
         * Test
         */
        Seda21Form::autoRepair(new Entity(['id' => 1, 'archival_agency_id' => 2, 'xml' => $file]));
        $util = DOMUtility::load($file);
        $filename = $util->xpath->query('//ns:BinaryDataObject/ns:FileInfo/ns:Filename')->item(0)->nodeValue;
        $this->assertEquals('sample.pdf', $filename);
        $this->assertEquals(1, $util->xpath->query('//ns:MessageDigest')->count());
    }

    /**
     * test AppendNodeComplex
     */
    public function testAppendNodeComplex(): void
    {
        $dom = new DOMDocument();
        $dom->formatOutput = true;
        $dom->preserveWhiteSpace = false;
        $dom->loadXML(
            <<<EOT
<ArchiveTransfer xmlns="fr:gouv:culture:archivesdefrance:seda:v2.1">
    <DataObjectPackage xml:id="ID1">
        <ManagementMetadata>
            <StorageRule>
                <Rule>included</Rule>
                <StartDate>2020</StartDate>
            </StorageRule>
        </ManagementMetadata>
    </DataObjectPackage>
</ArchiveTransfer>
EOT
        );
        $xpath = new DOMXPath($dom);
        $namespace = $dom->documentElement->getAttributeNode('xmlns')->nodeValue;
        $xpath->registerNamespace('ns', $namespace);
        $parentNode = $xpath
            ->query('/ns:ArchiveTransfer/ns:DataObjectPackage/ns:ManagementMetadata/ns:StorageRule')
            ->item(0);

        $node = AbstractGenericMessageForm::appendNode(
            2,
            'Rule',
            'ArchiveTransfer.DataObjectPackage.ManagementMetadata.StorageRule',
            $dom,
            $parentNode
        );
        $this->assertInstanceOf(DOMElement::class, $node);
        $node->nodeValue = 'test';
        $expected = <<<EOT
<StorageRule>
  <Rule>included</Rule>
  <StartDate>2020</StartDate>
  <Rule>test</Rule>
</StorageRule>
EOT;
        $this->assertEquals($expected, $dom->saveXML($node->parentNode));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testListFiles(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $this->assertEquals(['sample.pdf'], Seda21Form::listFiles($file));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testaddMultipleKeywords(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda21Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $form->addMultipleKeywords([2]);
        $util = DOMUtility::loadXML($util->dom->saveXML());
        $this->assertEquals('testunit1', $util->nodeValue('//ns:KeywordReference'));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testaddMultipleCustomKeywords(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda21Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $form->addMultipleCustomKeywords([['name' => 'foo', 'code' => 'bar'], ['bad_keyword' => 'empty']]);
        $util = DOMUtility::loadXML($util->dom->saveXML());
        $this->assertEquals('bar', $util->nodeValue('//ns:KeywordReference'));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testremoveNode(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda21Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $form->removeNode();
        $util = DOMUtility::loadXML($util->dom->saveXML());
        $this->assertEmpty($util->node('//ns:Content'));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testExecute(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda21Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $this->assertTrue($form->execute([]));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testAddBinaryDataObject(): void
    {
        Filesystem::copy(self::SEDA21_TEST_FILE, $file = $this->testDir . 'seda21.xml');
        $util = DOMUtility::load($file);

        // On supprime les binary existant pour faciliter les assertions
        foreach ($util->xpath->query('//ns:BinaryDataObject') as $binary) {
            $binary->parentNode->removeChild($binary);
        }

        $element = $util->node('//ns:DataObjectPackage');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $node = AbstractGenericMessageForm::appendNode(
            2,
            'BinaryDataObject',
            $path,
            $util->dom,
            $element
        );
        $form = Seda21Form::buildForm(
            2,
            "$path.BinaryDataObject",
            $util->dom,
            $node
        );
        $custom = [
            'action' => "multiAction",
            'args' => [
                [
                    'action' => "unsetValue",
                    'args' => [
                        "Attachment",
                    ],
                ],
                [
                    'action' => "addBinaryDataObject",
                ],
            ],
        ];
        $entity = $this->fetchTable('Transfers')->get(1);
        $attachment = $this->fetchTable('TransferAttachments')->get(2);
        Configure::write('App.paths.data', $this->testDir);
        Filesystem::dumpFile($attachment->get('path'), 'test');
        $form->setEntity($entity);
        $form->execute(
            [
                'Attachment' => [
                    ['@' => '', 'filename' => 'sample.pdf'],
                ],
                'custom' => [
                    ['@' => json_encode($custom)]
                ]
            ]
        );
        $result = $util->dom->saveXML();
        $this->assertStringContainsString('filename="sample.pdf"', $result);
    }
}
