<?php

namespace AsalaeCore\Test\TestCase\Form;

use AsalaeCore\Form\AbstractGenericMessageForm;
use AsalaeCore\Form\MessageSchema\MessageSchema;
use AsalaeCore\Form\Seda22Form;
use AsalaeCore\Test\Fixture\OrgEntitiesFixture;
use AsalaeCore\TestSuite\TestCase;
use AsalaeCore\Utility\DOMUtility;
use Cake\Cache\Cache;
use Cake\Core\Configure;
use Cake\ORM\Entity;
use Cake\ORM\TableRegistry;
use DateTime;
use DOMDocument;
use DOMElement;
use DOMXPath;
use Exception;
use Libriciel\Filesystem\Utility\Filesystem;
use ReflectionClass;
use ReflectionException;

class Seda22FormTest extends TestCase
{
    public const string SEDA22_TEST_FILE = TESTS . 'Data' . DS . 'sample_seda22.xml';

    public string $testDir = TMP_TESTDIR . DS . 'seda_22_form_test';

    public array $fixtures = [
        'app.Agreements',
        'app.KeywordLists',
        'app.Keywords',
        'app.OrgEntities',
        'app.Profiles',
        'app.ServiceLevels',
        'app.TransferAttachments',
        'app.Transfers',
        'app.TypeEntities',
    ];

    /**
     * setup
     */
    public function setUp(): void
    {
        parent::setUp();
        Cache::disable();
        Filesystem::reset();
        if (is_dir($this->testDir)) {
            Filesystem::remove($this->testDir);
        }
        Filesystem::mkdir($this->testDir);
    }

    /**
     * tearDown
     */
    public function tearDown(): void
    {
        parent::tearDown();
        if (is_dir($this->testDir)) {
            Filesystem::remove($this->testDir);
        }
        Cache::disable();
    }

    /**
     * test AutoRepair
     */
    public function testAutoRepair(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);

        $TransferAttachments = TableRegistry::getTableLocator()->get('TransferAttachments');
        $TransferAttachments->deleteAll([]);

        /**
         * On créé un xml avec des erreurs réparable automatiquement
         */
        $util = DOMUtility::load($file);
        /** @var DOMElement $BinaryDataObject */
        foreach ($util->xpath->query('//ns:BinaryDataObject') as $BinaryDataObject) {
            $Filename = $BinaryDataObject->getElementsByTagName('Filename')->item(0);
            $TransferAttachments->save(
                new Entity(
                    [
                        'transfer_id' => 1,
                        'filename' => $Filename->nodeValue,
                        'size' => 1,
                        'hash' => 'foo',
                        'hash_algo' => 'bar',
                    ]
                )
            );
            $Filename->nodeValue = '/' . $Filename->nodeValue;
            $util->dom->save($file);
        }
        // Integrity manquant
        /** @var DOMElement $integrity */
        foreach ($util->xpath->query('//ns:MessageDigest') as $integrity) {
            $integrity->parentNode->removeChild($integrity);
        }
        $util->dom->save($file);
        $util = DOMUtility::load($file);
        $filename = $util->xpath->query('//ns:BinaryDataObject/ns:FileInfo/ns:Filename')->item(0)->nodeValue;
        $this->assertEquals('/sample.pdf', $filename);
        $this->assertEquals(0, $util->xpath->query('//ns:MessageDigest')->count());

        /**
         * Test
         */
        Seda22Form::autoRepair(new Entity(['id' => 1, 'archival_agency_id' => 2, 'xml' => $file]));
        $util = DOMUtility::load($file);
        $filename = $util->xpath->query('//ns:BinaryDataObject/ns:FileInfo/ns:Filename')->item(0)->nodeValue;
        $this->assertEquals('sample.pdf', $filename);
        $this->assertEquals(1, $util->xpath->query('//ns:MessageDigest')->count());
    }

    /**
     * test AppendNodeComplex
     */
    public function testAppendNodeComplex(): void
    {
        $dom = new \DOMDocument();
        $dom->formatOutput = true;
        $dom->preserveWhiteSpace = false;
        $dom->loadXML(
            <<<EOT
<ArchiveTransfer xmlns="fr:gouv:culture:archivesdefrance:seda:v2.1">
    <DataObjectPackage xml:id="ID1">
        <ManagementMetadata>
            <StorageRule>
                <Rule>included</Rule>
                <StartDate>2020</StartDate>
            </StorageRule>
        </ManagementMetadata>
    </DataObjectPackage>
</ArchiveTransfer>
EOT
        );
        $xpath = new DOMXPath($dom);
        $namespace = $dom->documentElement->getAttributeNode('xmlns')->nodeValue;
        $xpath->registerNamespace('ns', $namespace);
        $parentNode = $xpath->query(
            '/ns:ArchiveTransfer/ns:DataObjectPackage/ns:ManagementMetadata/ns:StorageRule'
        )->item(0);

        $node = AbstractGenericMessageForm::appendNode(
            2,
            'Rule',
            'ArchiveTransfer.DataObjectPackage.ManagementMetadata.StorageRule',
            $dom,
            $parentNode
        );
        $this->assertInstanceOf(DOMElement::class, $node);
        $node->nodeValue = 'test';
        $expected = <<<EOT
<StorageRule>
  <Rule>included</Rule>
  <StartDate>2020</StartDate>
  <Rule>test</Rule>
</StorageRule>
EOT;
        $this->assertEquals($expected, $dom->saveXML($node->parentNode));
    }

    /**
     * test _buildSchema
     */
    public function testbuildSchema(): void
    {
        MessageSchema::$schemas = [];
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);
        $dom = new DOMDocument();
        $dom->load($file);
        $form = Seda22Form::buildForm(2, 'ArchiveTransfer', $dom);
        $schema = $form->getSchema();
        $this->assertTrue(in_array('MessageIdentifier', $schema->fields()));

        $element = $dom->getElementsByTagName('BinaryDataObject')->item(0);
        $this->assertNotEmpty($element);
        $form = Seda22Form::buildForm(2, 'ArchiveTransfer.DataObjectPackage.BinaryDataObject', $dom, $element);
        $schema = $form->getSchema();
        $this->assertTrue(in_array('Attachment_#0_@filename', $schema->fields()));
    }

    /**
     * test validation
     */
    public function testvalidation(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);
        $dom = new DOMDocument();
        $dom->load($file);
        $element = $dom->getElementsByTagName('BinaryDataObject')->item(0);
        $this->assertNotEmpty($element);
        $form = Seda22Form::buildForm(2, 'ArchiveTransfer.DataObjectPackage.BinaryDataObject', $dom, $element);
        $this->assertTrue($form->validate(['Attachment_#0_@filename' => 'test']));
    }

    /**
     * test execute
     */
    public function testexecute(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);
        $dom = new DOMDocument();
        $dom->load($file);
        $element = $dom->getElementsByTagName('BinaryDataObject')->item(0);
        $this->assertNotEmpty($element);
        $form = Seda22Form::buildForm(2, 'ArchiveTransfer.DataObjectPackage.BinaryDataObject', $dom, $element);
        $uid = uniqid('uniq-', true);
        $this->assertTrue($form->execute(['Attachment' => [0 => ['filename' => $uid]]]));
        $this->assertStringContainsString($uid, $dom->saveXML());
    }

    /**
     * test getDataJsTree
     */
    public function testgetDataJsTree(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);
        $form = Seda22Form::buildForm(2);
        $json = json_decode($form->getDataJsTree($file), true);
        $this->assertIsArray($json);
        $this->assertArrayHasKey(0, $json);
        $this->assertArrayHasKey('path', $json[0]);
        $this->assertEquals('ArchiveTransfer', $json[0]['path']);
    }

    /**
     * @return void
     */
    public function testgetDataDirectory(): void
    {
        $this->assertNotEmpty(AbstractGenericMessageForm::getDataDirectory(1));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testsmartAppendNode(): void
    {
        $dom = new \DOMDocument();
        $dom->formatOutput = true;
        $dom->preserveWhiteSpace = false;
        $dom->loadXML(
            <<<EOT
<ArchiveTransfer xmlns="fr:gouv:culture:archivesdefrance:seda:v2.2">
    <DataObjectPackage xml:id="ID1">
        <ManagementMetadata>
            <StorageRule>
                <Rule>included</Rule>
                <StartDate>2020</StartDate>
            </StorageRule>
            <AccessRule>
                <Rule>AR038</Rule>
                <StartDate>2020</StartDate>
            </AccessRule>
        </ManagementMetadata>
    </DataObjectPackage>
    <ArchivalAgency>
        <Identifier>sa</Identifier>
    </ArchivalAgency>
</ArchiveTransfer>
EOT
        );
        $xpath = new DOMXPath($dom);
        $namespace = $dom->documentElement->getAttributeNode('xmlns')->nodeValue;
        $xpath->registerNamespace('ns', $namespace);
        $parentNode = $xpath->query(
            '/ns:ArchiveTransfer/ns:DataObjectPackage/ns:ManagementMetadata'
        )->item(0);

        // Va ajouter un AppraisalRule entre StorageRule et AccessRule en prenant en compte le SA automatiquement
        AbstractGenericMessageForm::smartAppendNode($parentNode, 'AppraisalRule', true);
        $nodesOrder = [];
        foreach ($parentNode->childNodes as $child) {
            if ($child instanceof \DOMElement) {
                $nodesOrder[] = $child->nodeName;
            }
        }
        $expected = [
            'StorageRule',
            'AppraisalRule',
            'AccessRule',
        ];
        $this->assertEquals($expected, $nodesOrder);
    }

    /**
     * @return void
     */
    public function testclearCache(): void
    {
        Cache::enable();
        $key = Seda22Form::SCHEMA_CACHE_KEY . '_2'; // sv (id=4) dans sa (id=2)
        Cache::write($key, 'test');
        AbstractGenericMessageForm::clearCache(Seda22Form::SCHEMA_CACHE_KEY, 4);
        $result = Cache::read($key);
        $this->assertEmpty($result);
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testvalidateFile(): void
    {
        $file = $this->testDir . 'seda22.xml';
        Filesystem::copy(self::SEDA22_TEST_FILE, $file, true);
        $form = Seda22Form::buildForm(2);
        $this->assertEmpty($form->validateFile($file));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testsearchForMissingNodes(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $util = DOMUtility::load($file);
        $node = $util->node('//ns:ArchivalAgency/ns:Identifier');
        $node->parentNode->removeChild($node);
        $util->dom->save($file);

        $form = Seda22Form::buildForm(
            2,
            '',
            $util->dom,
        );
        $form->getSchema();
        $result = $form->searchForMissingNodes($file);
        $expected = [
            __(
                "Element obligatoire manquant dans {0}: {1}",
                'ArchivalAgency',
                'Identifier',
            )
        ];
        $this->assertEquals($expected, $result);
    }

    /**
     * @return void
     */
    public function testsetEntity(): void
    {
        $form = Seda22Form::buildForm(2);
        $form->setEntity(new Entity(['id' => 1]));
        $this->assertTrue(true);
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testremoveNode(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertFileExists($file);
        $dom = new DOMDocument();
        $dom->load($file);
        $element = $dom->getElementsByTagName('BinaryDataObject')->item(0);
        $this->assertNotEmpty($element);
        $form = Seda22Form::buildForm(2, 'ArchiveTransfer.DataObjectPackage.BinaryDataObject', $dom, $element);
        $form->removeNode();
        $element = $dom->getElementsByTagName('BinaryDataObject')->item(0);
        $this->assertEmpty($element);
    }

    /**
     * @return void
     * @throws ReflectionException
     */
    public function testconvertValueDataType(): void
    {
        $form = Seda22Form::buildForm(2);
        $reflection = new ReflectionClass($form);
        $method = $reflection->getMethod('convertValueDataType');
        $value = null;
        $method->invokeArgs($form, [&$value, 'date']);
        $this->assertNull($value);
        $date = new DateTime('2020-01-01');
        $value = $date;
        $method->invokeArgs($form, [&$value, 'date']);
        $this->assertEquals('2020-01-01', $value);
        $value = '01/01/2020';
        $method->invokeArgs($form, [&$value, 'date']);
        $this->assertEquals('2020-01-01', $value);
        $value = '01/01/2020 15:20';
        $method->invokeArgs($form, [&$value, 'datetime']);
        $this->assertEquals('2020-01-01T15:20:00', $value);
    }

    /**
     * @return void
     * @throws ReflectionException
     */
    public function testinsertOrgEntityData(): void
    {
        $form = Seda22Form::buildForm(2);
        $reflection = new ReflectionClass($form);
        $method = $reflection->getMethod('insertOrgEntityData');
        $data = $reflection->getProperty('data');
        $data->setValue($form, ['Identifier' => [0 => ['@' => 'sa']]]);
        $initialField = 'Identifier';
        $replacement = ['name' => 'Identifier'];
        $method->invoke($form, $initialField, $replacement);
        $result = $data->getValue($form);
        $expected = ['Identifier' => [0 => ['@' => OrgEntitiesFixture::defaultValues()[1]['name']]]];
        $this->assertEquals($expected, $result);
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testListFiles(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $this->assertEquals(['sample.pdf'], Seda22Form::listFiles($file));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testaddMultipleKeywords(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda22Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $form->addMultipleKeywords([2]);
        $util = DOMUtility::loadXML($util->dom->saveXML());
        $this->assertEquals('testunit1', $util->nodeValue('//ns:KeywordReference'));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testaddMultipleCustomKeywords(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $util = DOMUtility::load($file);
        $element = $util->node('//ns:Content');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $form = Seda22Form::buildForm(
            2,
            $path,
            $util->dom,
            $element
        );
        $form->addMultipleCustomKeywords([['name' => 'foo', 'code' => 'bar'], ['bad_keyword' => 'empty']]);
        $util = DOMUtility::loadXML($util->dom->saveXML());
        $this->assertEquals('bar', $util->nodeValue('//ns:KeywordReference'));
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testAddBinaryDataObject(): void
    {
        Filesystem::copy(self::SEDA22_TEST_FILE, $file = $this->testDir . 'seda22.xml');
        $util = DOMUtility::load($file);

        // On supprime les binary existant pour faciliter les assertions
        foreach ($util->xpath->query('//ns:BinaryDataObject') as $binary) {
            $binary->parentNode->removeChild($binary);
        }

        $element = $util->node('//ns:DataObjectPackage');
        $path = str_replace('/', '.', ltrim($util->getElementXpath($element, ''), '/'));
        $node = AbstractGenericMessageForm::appendNode(
            2,
            'BinaryDataObject',
            $path,
            $util->dom,
            $element
        );
        $form = Seda22Form::buildForm(
            2,
            "$path.BinaryDataObject",
            $util->dom,
            $node
        );
        $custom = [
            'action' => "multiAction",
            'args' => [
                [
                    'action' => "unsetValue",
                    'args' => [
                        "Attachment",
                    ],
                ],
                [
                    'action' => "addBinaryDataObject",
                ],
            ],
        ];
        $entity = $this->fetchTable('Transfers')->get(1);
        $attachment = $this->fetchTable('TransferAttachments')->get(2);
        Configure::write('App.paths.data', $this->testDir);
        Filesystem::dumpFile($attachment->get('path'), 'test');
        $form->setEntity($entity);
        $form->execute(
            [
                'Attachment' => [
                    ['@' => '', 'filename' => 'sample.pdf'],
                ],
                'custom' => [
                    ['@' => json_encode($custom)]
                ]
            ]
        );
        $result = $util->dom->saveXML();
        $this->assertStringContainsString('filename="sample.pdf"', $result);
    }
}
