<?php

namespace AsalaeCore\Test\TestCase\Model\Entity;

use AsalaeCore\Factory\Utility;
use AsalaeCore\Http\Client;
use AsalaeCore\Model\Entity\ArchivingSystem;
use AsalaeCore\TestSuite\TestCase;
use Cake\Core\Configure;
use Cake\Http\Client as CakeClient;
use PHPUnit\Framework\MockObject\Exception as MockException;
use ReflectionException;

class ArchivingSystemTest extends TestCase
{
    public array $fixtures = [
        'app.ArchivingSystems',
        'app.AuthUrls',
        'app.OutgoingTransfers',
    ];

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testwhoami(): void
    {
        $json = json_encode(
            [
                'username' => 'testunit',
                'fullname' => 'testunit',
                'entity_name' => "Mon autre service d'archive",
                'entity_identifier' => 'sa2',
                'archival_agency_name' => "Mon autre service d'archive",
                'archival_agency_identifier' => 'sa2',
                'role' => 'ws versant',
                'role_root' => 'ws versant',
            ]
        );
        $response = $this->createMock(CakeClient\Response::class);
        $response->method('getJson')->willReturn($json);
        $client = new class extends Client {
            public static CakeClient\Response $response;
            // phpcs:ignore
            public function get(string $url, array|string $data = [], array $options = []): CakeClient\Response
            {
                return self::$response;
            }
        };
        $client::$response = $response;
        Utility::set(Client::class, $client);

        /** @var ArchivingSystem $archivingSystem */
        $archivingSystem = $this->fetchTable('ArchivingSystems')->get(1);
        $this->assertStringContainsString('testunit', $archivingSystem->whoami());
    }

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testrequestOptions(): void
    {
        $json = json_encode(
            [
                "archival_agency_identifier" => "sa2",
                "archival_agency_name" => "Mon autre service d'archive",
                "transferring_agencies" => [
                    "Mon autre service d'archive" => [
                        "sa2" => "sa2 - Mon autre service d'archive",
                    ],
                ],
                "originating_agencies" => [
                    "Mon autre service d'archive" => [
                        "sa2" => "sa2 - Mon autre service d'archive",
                    ],
                ],
                "agreements" => [
                    "auto" => "auto - Auto",
                ],
                "profiles" => [
                    "profil" => "profil - profil",
                ],
                "services" => [],
            ]
        );
        $response = $this->createMock(CakeClient\Response::class);
        $response->method('getJson')->willReturn($json);
        $response->method('getStatusCode')->willReturn(200);
        $client = new class extends Client {
            public static CakeClient\Response $response;
            // phpcs:ignore
            public function get(string $url, array|string $data = [], array $options = []): CakeClient\Response
            {
                return self::$response;
            }
        };
        $client::$response = $response;
        Utility::set(Client::class, $client);

        /** @var ArchivingSystem $archivingSystem */
        $archivingSystem = $this->fetchTable('ArchivingSystems')->get(1);
        $this->assertStringContainsString("Mon autre service d'archive", $archivingSystem->requestOptions());
    }

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testpostOutgoingTransfer(): void
    {
        Configure::write('App.paths.data', TMP_TESTDIR);
        $response = $this->createMock(CakeClient\Response::class);
        $response->method('getStringBody')->willReturn('OK');
        $client = new class extends Client {
            public static CakeClient\Response $response;
            // phpcs:ignore
            public function post(string $url, mixed $data = [], array $options = []): CakeClient\Response
            {
                return self::$response;
            }
        };
        $client::$response = $response;
        Utility::set(Client::class, $client);

        /** @var ArchivingSystem $archivingSystem */
        $archivingSystem = $this->fetchTable('ArchivingSystems')->get(1);
        $outgoingTransfer = $this->fetchTable('OutgoingTransfers')->get(1);
        $zip = $outgoingTransfer->get('zip');
        $dir = dirname($zip);
        if (!is_dir($dir)) {
            mkdir($dir, 077, true);
        }
        touch($zip);
        $this->assertStringContainsString("OK", $archivingSystem->postOutgoingTransfer($outgoingTransfer));
    }

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testgetAcknowledgement(): void
    {
        $response = $this->createMock(CakeClient\Response::class);
        $response->method('getStringBody')->willReturn('OK');
        $client = new class extends Client {
            public static CakeClient\Response $response;
            // phpcs:ignore
            public function get(string $url, array|string $data = [], array $options = []): CakeClient\Response
            {
                return self::$response;
            }
        };
        $client::$response = $response;
        Utility::set(Client::class, $client);

        /** @var ArchivingSystem $archivingSystem */
        $archivingSystem = $this->fetchTable('ArchivingSystems')->get(1);
        $outgoingTransfer = $this->fetchTable('OutgoingTransfers')->get(1);
        $this->assertStringContainsString(
            "OK",
            $archivingSystem->getAcknowledgement($outgoingTransfer)->getStringBody()
        );
    }

    /**
     * @return void
     * @throws MockException
     * @throws ReflectionException
     */
    public function testgetReply(): void
    {
        $response = $this->createMock(CakeClient\Response::class);
        $response->method('getStringBody')->willReturn('OK');
        $client = new class extends Client {
            public static CakeClient\Response $response;
            // phpcs:ignore
            public function get(string $url, array|string $data = [], array $options = []): CakeClient\Response
            {
                return self::$response;
            }
        };
        $client::$response = $response;
        Utility::set(Client::class, $client);

        /** @var ArchivingSystem $archivingSystem */
        $archivingSystem = $this->fetchTable('ArchivingSystems')->get(1);
        $outgoingTransfer = $this->fetchTable('OutgoingTransfers')->get(1);
        $this->assertStringContainsString(
            "OK",
            $archivingSystem->getReply($outgoingTransfer)->getStringBody()
        );
    }

    /**
     * @return void
     */
    public function testgetDeletable(): void
    {
        $ArchivingSystems = $this->fetchTable('ArchivingSystems');
        $entity = $ArchivingSystems->newEmptyEntity();
        $this->assertTrue($entity->get('deletable'));
        $entity->set('id', 1);
        $this->assertTrue($entity->get('deletable'));
        $entity->set('org_entity_id', 2);
        $this->assertFalse($entity->get('deletable'));
    }

    /**
     * @return void
     */
    public function testgetRemovable(): void
    {
        $ArchivingSystems = $this->fetchTable('ArchivingSystems');
        $entity = $ArchivingSystems->newEmptyEntity();
        $this->assertFalse($entity->get('removable'));
        $entity->set('id', 1);
        $this->assertTrue($entity->get('removable'));
    }
}
