<?php

namespace AsalaeCore\Test\TestCase\Utility;

use AsalaeCore\Factory\Utility;
use AsalaeCore\TestSuite\TestCase;
use AsalaeCore\Utility\Antivirus\Clamav;
use AsalaeCore\Utility\Check;
use Beanstalk\Utility\Beanstalk;
use Cake\Core\Configure;
use PHPUnit\Framework\MockObject\Exception;
use ReflectionException;

class CheckTest extends TestCase
{
    /**
     * @return void
     */
    public function tearDown(): void
    {
        parent::tearDown();
        Utility::reset();
    }

    /**
     * @return void
     */
    public function testdatabase(): void
    {
        $this->assertTrue(Check::database());
    }

    /**
     * @return void
     * @throws Exception
     */
    public function testbeanstalkd(): void
    {
        $mock = $this->createMock(Beanstalk::class);
        $mock->method('isConnected')->willReturn(true);
        Beanstalk::$instance = $mock;
        $this->assertTrue(Check::beanstalkd());
    }

    /**
     * @return void
     */
    public function testratchet(): void
    {
        Check::$getHeadersCallable = fn() => ['HTTP/1.1 200 OK'];
        $this->assertTrue(Check::ratchet());
    }

    /**
     * @return void
     * @throws ReflectionException
     */
    public function testclamav(): void
    {
        Configure::write('Antivirus.classname', Clamav::class);
        $clamav = new class extends Clamav {
            private static int $i = 0;
            // phpcs:ignore
            public static function ping(): bool
            {
                return true;
            }
            // phpcs:ignore
            public static function scan(string $directory): array
            {
                if (self::$i === 0) {
                    self::$i = 1;
                    return ['virus found'];
                }
                return [];
            }
        };
        Utility::set(Clamav::class, $clamav);
        $this->assertTrue(Check::clamav());
    }

    /**
     * @return void
     */
    public function testNeedInstall(): void
    {
        Check::$prerequis = ['php'];
        Check::$prerequisVersion = [
            'php' => [
                'command' => 'php -v',
                'regex' => '/PHP (\d+\.\d+(?:\.\d)?(?:\.\d)?)/',
                'constrain' => '^8.3',
            ],
        ];
        $this->assertFalse(Check::needInstall());
    }

    /**
     * @return void
     */
    public function testServiceExists(): void
    {
        exec('service --status-all', $o);
        foreach ($o as $service) {
            $service = preg_replace('/\s/', '', $service);
            if (preg_match('/\[([-+])](.*)$/', $service, $m)) {
                $service = $m[2];
                break;
            }
        }
        if (!isset($service)) {
            $this->fail('no service found');
        }
        $this->assertTrue(Check::serviceExists($service));
        $this->assertFalse(Check::serviceExists('this-service-does-not-exists'));
    }

    /**
     * @return void
     */
    public function testServiceIsRunning(): void
    {
        exec('service --status-all', $o);
        $running = null;
        $notRunning = null;
        foreach ($o as $service) {
            $service = preg_replace('/\s/', '', $service);
            if (preg_match('/\[([-+])](.*)$/', $service, $m)) {
                $service = $m[2];
                if ($m[1] === '-') {
                    $notRunning = $service;
                } else {
                    $running = $service;
                }
            }
        }
        if ($notRunning) {
            $this->assertFalse(Check::serviceIsRunning($notRunning));
        }
        if ($running) {
            $this->assertTrue(Check::serviceIsRunning($running));
        }
        if (!$notRunning && !$running) {
            $this->fail('service not found');
        }
        $this->assertNull(Check::serviceIsRunning('this-service-does-not-exists'));
    }

    /**
     * @return void
     */
    public function testPhpOk(): void
    {
        ini_set('upload_max_filesize', '2M');
        ini_set('post_max_size', '1G');
        Configure::write('Flow.chunkSize', 200000000); // chunk de 200Mo
        $this->assertFalse(Check::phpOk());

        Configure::write('Flow.chunkSize', 200); // chunk de 200 octets
        $this->assertTrue(Check::phpOk());
    }

    /**
     * @return void
     */
    public function testwritableOk(): void
    {
        $originalPathToData = Configure::read('App.paths.data');
        Configure::write('App.paths.data', $originalPathToData . DS . 'this_path_does_not_exists');
        $test1 = Check::writableOk();
        Configure::write('App.paths.data', $originalPathToData);
        $test2 = Check::writableOk();
        $this->assertFalse($test1);
        $this->assertTrue($test2);
    }
}
